/*****************************************************************************
 * COPYRIGHT (C) BERNECKER + RAINER, AUSTRIA, A-5142 EGGELSBERG, B&R STRASZE 1
 *****************************************************************************
 * Module:  ANSLArror
 * File:    ANSLERROR.h
 * Author:  A.T.
 * Created: 2008.05.07
 ******************************************************************************
 * Description:
 * public error codes for ANSL Online user interface
 *
 * History:
 ******************************************************************************/

#if !defined(AFX_ANSLERROR_H__545480F4_36C6_4AE8_9F08_9BC0169A9DA4__INCLUDED_)
#define AFX_ANSLERROR_H__545480F4_36C6_4AE8_9F08_9BC0169A9DA4__INCLUDED_


/******************************************************************************
*  Base Error Types:
******************************************************************************/

typedef unsigned int	ANSL_ERRCODE;	/* ANSL error code type				*/

/* base error code (format: uint32_t or uint16_t): */

#ifdef _VXWORKS_
/* VxWorks: reserved range: 31650 - 31949 */
#define ANSL_BASE_ERR					31650
#else
#define ANSL_BASE_ERR					10000
#endif

/******************************************************************************
*  ANSL Core Error Codes (reserved ANSL_BASE_ERR+0 - ANSL_BASE_ERR+99):
******************************************************************************/

#define ANSL_NO_ERR						0

#define ANSL_ERR_UNKNOWN				(ANSL_BASE_ERR+0)
#define ANSL_ERR_MEMORY_RESOURCE		(ANSL_BASE_ERR+1)
#define ANSL_ERR_EVENT_RESOURCE			(ANSL_BASE_ERR+2)
#define ANSL_ERR_THREAD_RESOURCE		(ANSL_BASE_ERR+3)
#define ANSL_ERR_SYNC_RESOURCE			(ANSL_BASE_ERR+4)

#define ANSL_ERR_UNSUPPORTED			(ANSL_BASE_ERR+10)
#define ANSL_ERR_UNUSABLE_SIZE			(ANSL_BASE_ERR+11)
#define ANSL_ERR_INVALID_PARAM			(ANSL_BASE_ERR+12)
#define ANSL_ERR_INVALID_NAME			(ANSL_BASE_ERR+13)
#define ANSL_ERR_INVALID_USAGE			(ANSL_BASE_ERR+14)
#define ANSL_ERR_DAMAGED				(ANSL_BASE_ERR+15)
#define ANSL_ERR_INTERNAL				(ANSL_BASE_ERR+16)
#define ANSL_ERR_STACK_CONFIG			(ANSL_BASE_ERR+17)
#define ANSL_ERR_VERSION_CONFLICT		(ANSL_BASE_ERR+18)
#define ANSL_ERR_OPTION_CONFLICT		(ANSL_BASE_ERR+19)
#define ANSL_ERR_APPL_NOT_EXIST			(ANSL_BASE_ERR+20)

#define ANSL_ERR_CONN_CLOSE				(ANSL_BASE_ERR+30)
#define ANSL_ERR_CONN_TIMEOUT			(ANSL_BASE_ERR+31)
#define ANSL_ERR_CONN_LOGIN				(ANSL_BASE_ERR+32)
#define ANSL_ERR_CONN_CLOSED			(ANSL_BASE_ERR+33)

#define ANSL_ERR_DATA_INCOMPLETE		(ANSL_BASE_ERR+35)
#define ANSL_ERR_DATA_INVALID			(ANSL_BASE_ERR+36)
#define ANSL_ERR_DATA_PROTOCOL			(ANSL_BASE_ERR+37)

#define ANSL_ERR_SOCK_VERSION			(ANSL_BASE_ERR+40)
#define ANSL_ERR_SOCK_NOTRDY			(ANSL_BASE_ERR+41)
#define ANSL_ERR_SOCK_CREATE			(ANSL_BASE_ERR+42)
#define ANSL_ERR_SOCK_LISTEN			(ANSL_BASE_ERR+44)
#define ANSL_ERR_SOCK_CONNECT			(ANSL_BASE_ERR+45)
#define ANSL_ERR_SOCK_COMM				(ANSL_BASE_ERR+46)
#define ANSL_ERR_SOCK_IPADDR			(ANSL_BASE_ERR+47)

#define ANSL_ERR_OBJ_DEFAULT			(ANSL_BASE_ERR+50)
#define ANSL_ERR_OBJ_DELETED			(ANSL_BASE_ERR+51)
#define ANSL_ERR_OBJ_UNDEFINED			(ANSL_BASE_ERR+52)
#define ANSL_ERR_OBJ_NOT_READY			(ANSL_BASE_ERR+53)

/******************************************************************************
*  ANSL Online Service Error Codes (>= ANSL_BASE_ERR+100):
******************************************************************************/

#define ANSL_ONL_ERR_UNSUPPORTED_OBJ	(ANSL_BASE_ERR+100)
#define ANSL_ONL_ERR_UNSUPPORTED_SVC	(ANSL_BASE_ERR+101)
#define ANSL_ONL_ERR_OBJ_HIERARCHY		(ANSL_BASE_ERR+102)
#define ANSL_ONL_ERR_OBJ_NOTEXIST		(ANSL_BASE_ERR+103)
#define ANSL_ONL_ERR_OBJ_INVALID		(ANSL_BASE_ERR+104)
#define ANSL_ONL_ERR_ILG_SVC_DATA		(ANSL_BASE_ERR+105)
#define ANSL_ONL_ERR_SENDJOB_NOTFOUND	(ANSL_BASE_ERR+106)
#define ANSL_ONL_ERR_SENDJOB_INUSE		(ANSL_BASE_ERR+108)
#define ANSL_ONL_ERR_OBJ_ALREADYEXIST	(ANSL_BASE_ERR+109)

#define ANSL_ONL_ERR_BRMOD_INSTALL		(ANSL_BASE_ERR+150)
#define ANSL_ONL_ERR_BRMOD_CHECKSUM		(ANSL_BASE_ERR+151)
#define ANSL_ONL_ERR_TIME_READ			(ANSL_BASE_ERR+152)
#define ANSL_ONL_ERR_TIME_WRITE			(ANSL_BASE_ERR+153)
#define ANSL_ONL_ERR_DELETE_MODUL		(ANSL_BASE_ERR+154)
#define ANSL_ONL_ERR_UPLOAD_MODUL		(ANSL_BASE_ERR+155)
#define ANSL_ONL_ERR_DOWNLOAD_MODUL		(ANSL_BASE_ERR+156)
#define ANSL_ONL_ERR_MODULLIST			(ANSL_BASE_ERR+157)
#define ANSL_ONL_ERR_PVLIST				(ANSL_BASE_ERR+158)
#define ANSL_ONL_ERR_GPVLIST			(ANSL_BASE_ERR+159)
#define ANSL_ONL_ERR_PV					(ANSL_BASE_ERR+160)
#define ANSL_ONL_ERR_ILG_PVFORMAT		(ANSL_BASE_ERR+161)


#endif // !defined(AFX_ANSLERROR_H__545480F4_36C6_4AE8_9F08_9BC0169A9DA4__INCLUDED_)
