/*****************************************************************************
 * COPYRIGHT (C) BERNECKER + RAINER, AUSTRIA, A-5142 EGGELSBERG, B&R STRASZE 1
 *****************************************************************************
 * Module:  ANSLuif
 * File:    ANSLUIF.h
 * Author:  A.T.
 * Created: 2008.05.07
 ******************************************************************************
 * Description:
 * public functions for the ANSL Online user interface
 *
 * History:
 ******************************************************************************/

#if !defined(ANSLUIF_H_INCLUDED_CC6CC22C_D3C2_4CD8_A5E3_812EC92BC513)
#define ANSLUIF_H_INCLUDED_CC6CC22C_D3C2_4CD8_A5E3_812EC92BC513

#ifdef	__cplusplus
#define ANSLUIF_API extern "C"
#else
#define ANSLUIF_API extern
#endif

#pragma once

typedef enum ANSL_OBJECT_TYPES ANSL_OBJECT_TYPES;

/* ANSL online object types: */
enum ANSL_OBJECT_TYPES
{
	// the server
	ANSL_OBJTYPE_SERVER,
	// the cpu object
	ANSL_OBJTYPE_CPU,
	// the container object for global variables
	ANSL_OBJTYPE_CPUPV,
	// the BR module object
	ANSL_OBJTYPE_MODULE,
	// the logger module object (derived from ANSL_OBJTYPE_MODULE)
	ANSL_OBJTYPE_LOGGER,
	// the library object
	ANSL_OBJTYPE_LIBRARY,
	// the logger module object (derived from ANSL_OBJTYPE_MODULE)
	ANSL_OBJTYPE_TASK,
	// the local- / task- PV object
	ANSL_OBJTYPE_TASKPV,
	// the IO variable object
	ANSL_OBJTYPE_IOPV
};

typedef enum ANSL_CALLBACK_TYPES ANSL_CALLBACK_TYPES;

/* online server service object types: */
enum ANSL_CALLBACK_TYPES
{
	// unknown call back type
	ANSL_UNKNOWN,
	// response due to a conect request
	ANSL_RESP_CONNECT,
	// response due to a disconect request
	ANSL_RESP_DISCONNECT,
	// error event
	ANSL_EVENT_ERROR,
	// status event
	ANSL_EVENT_STATUS,
	// response to a call on get date and time (this is only valid for CPU objects)
	ANSL_RESP_GET_DATETIME,
	// cancel the request to the reuqest on get date and time (this is only valid for CPU objects)
	ANSL_CANCEL_GET_DATETIME,
	// response to a call on set date and time (this is only valid for CPU objects)
	ANSL_RESP_SET_DATETIME,
	// cancel a request on set date and time (this is only valid for CPU objects)
	ANSL_CANCEL_SET_DATETIME,
	// response to a call on warm- restart (this is only valid for CPU objects)
	ANSL_RESP_WARM_RESTART,
	// cancel a call on warm- restart (this is only valid for CPU objects)
	ANSL_CANCEL_WARM_RESTART,
	// response to a call on cold- restart
	ANSL_RESP_COLD_RESTART,
	// cancel a cold- restart
	ANSL_CANCEL_COLD_RESTART,
	// response to an upload reqest for a module, task, libray or logger
	ANSL_RESP_UPLOAD,
	// cancelation of an upload reqest for a module, task, libray or logger
	ANSL_CANCEL_UPLOAD,
	// response to an download reqest for a module, task, libray or logger
	ANSL_RESP_DOWNLOAD,
	// cancelation of an downoad reqest for a module, task, libray or logger
	ANSL_CANCEL_DOWNLOAD,
	// format event for a variabe object
	ANSL_EVENT_FORMAT,
	// response to a call on get format for a variabe object
	ANSL_RESP_GET_FORMAT,
	// value change notivication for a variable object
	ANSL_EVENT_VALUE,
	// response on a request for the variable data
	ANSL_RESP_GET_VALUE,
	// cancel reading the data for a variable object
	ANSL_CANCEL_GET_VALUE,
	// response to write request on the value of a variable object
	ANSL_RESP_SET_VALUE,
	// cancel the write request of a value
	ANSL_CANCEL_SET_VALUE,
	// cancelation of a value changed event for a variable
	ANSL_CANCEL_EVENT_VALUE
};

/****************************************************************
// ANSL data type categories
// each specific datatyp is a combination of the base data type
// plus the number of bytes for the specific implementation
*****************************************************************/
// unsigned data types (unsigned -int8, -int16, -int32 and -int64)
#define	ANSL_UINT_TYPES			0x100
// unsigned data types (signed -int8, -int16, -int32 and -int64)
#define	ANSL_INT_TYPES			0x200
// data types representing a date, time, timespan or date and time
#define	ANSL_DATENTIME_TYPES	0x400
// data types representing a boolean value
#define	ANSL_BOOL_TYPES			0x800
// data types representing 32 and 64 bit floating point values
#define	ANSL_FLOAT_TYPES		0x1000
// data types representing strings (1 Byte and 2 Byte strings)
#define	ANSL_STRING_TYPES		0x2000
// signature for all structure datatypes
#define	ANSL_STRUCT_TYPES		0x4000

typedef enum ANSL_VAR_TYPES ANSL_VAR_TYPES;

/* ANSL data types; each data type contains the category and its byte length */
/* BOOL data types contain the category and the bit length */
enum ANSL_VAR_TYPES
{
	// undefined
	ANSL_VAR_TYPE_UNDEFINED		= 0,
	// 8 bit unsigned intger
	ANSL_VAR_TYPE_UINT8			= ANSL_UINT_TYPES+1,
	// 16 bit unsigned intger
	ANSL_VAR_TYPE_UINT16		= ANSL_UINT_TYPES+2,
	// 32 bit unsigned intger
	ANSL_VAR_TYPE_UINT32		= ANSL_UINT_TYPES+4,
	// 64 bit unsigned intger
	ANSL_VAR_TYPE_UINT64		= ANSL_UINT_TYPES+8,
	// 8 bit signed intger
	ANSL_VAR_TYPE_INT8			= ANSL_INT_TYPES+1,
	// 16 bit signed intger
	ANSL_VAR_TYPE_INT16			= ANSL_INT_TYPES+2,
	// 32 bit signed intger
	ANSL_VAR_TYPE_INT32			= ANSL_INT_TYPES+4,
	// 64 bit signed intger
	ANSL_VAR_TYPE_INT64			= ANSL_INT_TYPES+8,
	// 10 byte date and time buffer
	ANSL_VAR_TYPE_DT			= ANSL_DATENTIME_TYPES+10,
	// 8 bit boolean
	ANSL_VAR_TYPE_BOOL			= ANSL_BOOL_TYPES+8,
	// 32 bit float
	ANSL_VAR_TYPE_FLOAT32		= ANSL_FLOAT_TYPES+4,
	// 64 bit float
	ANSL_VAR_TYPE_FLOAT64		= ANSL_FLOAT_TYPES+8,
	// string 1 byte per character
	ANSL_VAR_TYPE_CHAR_STRING	= ANSL_STRING_TYPES+1,
	// string 2 bytes per character
	ANSL_VAR_TYPE_WCHAR_STRING	= ANSL_STRING_TYPES+2,
	// structure
	ANSL_VAR_TYPE_STRUCT		= ANSL_STRUCT_TYPES
};


typedef struct ANSL_VARIABLE_DATA_TYPE ANSL_VARIABLE_DATA_TYPE;

#pragma pack(push,1)
// Data type description for a variable
// if the variable is of type ANSL_VAR_TYPE_STRUCT the
// member m_pTypes will contain the type information of the members
struct ANSL_VARIABLE_DATA_TYPE
{
	/* the datatype of type ANSL_ONL_PV_TYPE */
	unsigned short m_DataType;
	/* the nesting level */
	unsigned short m_Nesting;
	/* the number of dimensions if the item is an array else 0 */
	unsigned short m_ArrayDimensions;
	/* the number of bytes for this element */
	unsigned int m_ElementLength;
	/* there are only 3 dimensions supported!! */
	unsigned int m_DimensionElements[3];
	/* element name (zero terminated 1 byte string) */
	const char* m_Name;
	/* the byte offset if the element is astruct member else 0 */
	unsigned int m_ByteOffset;
	/* the total number of bytes for this nesting level (nested elements will structures will be)*/
	unsigned int m_Size;
	ANSL_VARIABLE_DATA_TYPE* m_pTypes;			/* the element types for structures */
};
#ifdef PLATFORM_LINUX
	#pragma pack(pop)
#else
	#pragma pack(pop, 1)
#endif

typedef enum ANSL_VARIABLE_EVENT_FLAGS ANSL_VARIABLE_EVENT_FLAGS;

/* a bitwise combination of event flags */
enum ANSL_VARIABLE_EVENT_FLAGS
{
	ANSL_VAR_EV_FLAG_NONE,
	ANSL_VAR_EV_FLAG_FORMAT = (1<<0),
	ANSL_VAR_EV_FLAG_DATA = (1<<1)
};

typedef enum ANSL_CPU_RESTART_MODES ANSL_CPU_RESTART_MODES;

enum ANSL_CPU_RESTART_MODES
{
	ANSL_CPU_RESTART_UNDEF,	/* undefined */
	ANSL_CPU_COLD_RESTART,	/* cpu cold restart */
	ANSL_CPU_WARM_RESTART	/* cpu warm restart */
};

typedef struct ANSL_VARIABLE_EVENT_PARAMETER ANSL_VARIABLE_EVENT_PARAMETER;

#pragma pack(push,1)
struct ANSL_VARIABLE_EVENT_PARAMETER
{
	unsigned int m_EventFlags;		/* a bitwise combination of event flags see enum ANSL_VARIABLE_EVENT_FLAGS*/
	unsigned int m_RefreshTime;		/* refresh time (ms)  0 no automatic refresh*/
	int m_IntegerHyst;				/* hyst: 0=disabled   0!= integer hysterese is enabled**/
	float m_FloatHyst;				/* hyst: 0=disabled   0!= float hysterese is enabled**/
};
#ifdef PLATFORM_LINUX
	#pragma pack(pop)
#else
	#pragma pack(pop, 1)
#endif


////////////////////////////////////////////////////////
// CallBack declaration
// this function declaration is valid for all objects
//
#ifdef PLATFORM_LINUX
typedef unsigned int (__attribute__((stdcall))/*__stdcall*/ *ANSL_CALL_BACK) (unsigned int errorCode,
												  unsigned int ident,
												  ANSL_OBJECT_TYPES objectType,
												  ANSL_CALLBACK_TYPES cbType,
												  void* pData,
												  unsigned int dataLen,
												  void* pUserData);
#else
typedef unsigned int (__stdcall *ANSL_CALL_BACK) (unsigned int errorCode,
												  unsigned int ident,
												  ANSL_OBJECT_TYPES objectType,
												  ANSL_CALLBACK_TYPES cbType,
												  void* pData,
												  unsigned int dataLen,
												  void* pUserData);
#endif


///////////////////////////////////////////
// Acces functions for the CPU object

//! Initialize ANSL user interface
ANSLUIF_API unsigned int ANSLI_Initialize();
ANSLUIF_API void ANSLI_DeInitialize();


// Establish a tcpip connection to a cpu object
ANSLUIF_API unsigned int ANSLI_CPU_Connect(const char* addressCPU,
										   const char* commParameter,
										   unsigned int* pIdentCpu,
										   ANSL_CALL_BACK pCBFunction,
										   void* pUserData);

// Release the connection to the cpu object
ANSLUIF_API unsigned int ANSLI_CPU_Disconnect(unsigned int identCpu);

// Restart the cpu in a specified mode
ANSLUIF_API unsigned int ANSLI_CPU_Restart(unsigned int identCpu,
										   int booteMode);

#pragma pack(push,1)
// struct representing the cpu date and time format
typedef struct
{
	unsigned char tm_Second;			/* seconds after the minute  - [0, 59]*/
	unsigned char tm_Minute;			/* minutes after the hour  - [0, 59]*/
	unsigned char tm_Hour;				/* hours after midnight  - [0, 23]*/
	unsigned char tm_DayOfMonth;		/* day of the month  - [1, 31]*/
	unsigned char tm_Month;				/* months since January  - [0, 11]*/
	unsigned char tm_Year;				/* years since 1900*/
    unsigned char tm_WeekDay;			/* days since Sunday - [0,6] */
    unsigned char tm_DaylightSaving;	/* daylight savings time flag */
    unsigned short tm_DayOfYear;		/* days since January 1 - [0,365] */
} CPU_DATETIME_TYPE;
#ifdef PLATFORM_LINUX
	#pragma pack(pop)
#else
	#pragma pack(pop, 1)
#endif

// Reads the date and time from the cpu object
ANSLUIF_API unsigned int ANSLI_CPU_DTRead(unsigned int identCpu);

// Write the date and time to the cou object
ANSLUIF_API unsigned int ANSLI_CPU_DTWrite(unsigned int identCpu,
										   CPU_DATETIME_TYPE* pDateTime);


///////////////////////////////////////////
// Acces functions for the task object
///////////////////////////////////////////

// create a connection to a task
ANSLUIF_API unsigned int ANSLI_TASK_Connect(unsigned int parentIdent,
											const char* taskName,
											unsigned int* pIdentTask,
											ANSL_CALL_BACK pCBFunction,
											void* pUserData);

// release the connection to the task object.
ANSLUIF_API unsigned int ANSLI_TASK_Disconnect(unsigned int identTask);


///////////////////////////////////////////
// Acces functions for the variable object
///////////////////////////////////////////

// create a connection to a variable object.
// depending on the parent ID the variable will becomes a TASK local or a CPU global variable.
ANSLUIF_API unsigned int ANSLI_VAR_Connect(unsigned int parentIdent,
										   const char* addressVar,
										   unsigned int* pIdentVar,
										   ANSL_CALL_BACK pCBFunction,
										   void* pUserData,
										   ANSL_VARIABLE_EVENT_PARAMETER* pEventParameters);

// release the connection to the variable object.
ANSLUIF_API unsigned int ANSLI_VAR_Disconnect(unsigned int identVar);

// get the data type of the variable object.
ANSLUIF_API unsigned int ANSLI_VAR_ReadDataType(unsigned int identVar);

// get the value for the variable.
ANSLUIF_API unsigned int ANSLI_VAR_ReadValue(unsigned int identVar,
											 void* pDataBuffer,
											 unsigned int dataLenght);

// write the assigned value to the variable.
ANSLUIF_API unsigned int ANSLI_VAR_WriteValue(unsigned int identVar);

// assign and write the value.
ANSLUIF_API unsigned int ANSLI_VAR_SetValue(unsigned int identVar, 
											void* pDataBuffer, 
											unsigned int dataLenght);

// assign value to the variable (does not send value to PLC!).
// use this method for initialization of structures 
ANSLUIF_API unsigned int ANSLI_VAR_AssignValue(unsigned int identVar, 
											   unsigned int offset,
											   void* pDataBuffer, 
											   unsigned int dataLenght);


#endif // !defined(ANSLUIF_H_INCLUDED_CC6CC22C_D3C2_4CD8_A5E3_812EC92BC513)
