/*
 ============================================================================
 Name        : ANSLSample.c
 Author      : Werner Hager
 Version     : 1
 Copyright   : B&R
 Description : ANSL sample
 ============================================================================
 */

/***************************************************************************************************/
/** INCLUDES ***************************************************************************************/

#include <sys/syscall.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ANSLUIF.h>	/* ANSL USER INTERFACE */
#include <termios.h>
#include <unistd.h>
#include <pthread.h>
#include <semaphore.h>
#include <time.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h> // for pid_t gettid(void)
#include <stdarg.h>
#include <sys/select.h>

#ifdef APROL_SYSTEM
#define WINBASEAPI static
#else
#define WINBASEAPI
#endif
#define WINAPI

#ifndef sprintf_s
#define sprintf_s(frmt, size, args...) 			sprintf(frmt, ##args)
#endif


/***************************************************************************************************/
/** DEFINES ****************************************************************************************/

#define TARGET_IP 					"127.0.0.1"	/* IP of the target */
#define CONNECTION_PARAMS			"Timeout=\"35000\" Port=\"11169\" SendDelay=\"1\" Buffer=\"4096\""	/* connection parameters */
#define CONNECTION_PARAMS_NO_TO		"Port=\"11169\" SendDelay=\"1\" Buffer=\"4096\""	/* connection parameters */
#define CONNECTION_PARAMS_NO_TOPO	"SendDelay=\"1\" Buffer=\"4096\""	/* connection parameters */
#define TASK_NAME 					"pvitest"		/* task name of the task on target */
#define VAR_NAME_READ 				"PV1"			/* local UDINT variable */
#define VAR_NAME_SET 				"PV2"			/* local UDINT variable */
#define VAR_NAME_SET_VALUE	        20				/* VAR_NAME_SET gets this value */

#define RC_LOCK_ERROR              1
#define RC_OBJECT_NOT_CREATED      2
#define RC_STACKSIZE_ERROR         3
#define RC_THREAD_ATTR_ERROR       4
#define RC_THREAD_NOT_CREATED      5
#define RC_UNLOCK_ERROR            6
#define RC_NOT_OWNER               7

#define FMT_STR_SIZE			   1024
#define INFO_STR_SIZE			   256

#pragma pack(1)
/* read/write cpu date and time (request/response): */
typedef struct
{
    unsigned char		Sec;			/* seconds after the minute - [0,59] */
    unsigned char		Min;			/* minutes after the hour - [0,59] */
    unsigned char		Hour;			/* hours since midnight - [0,23] */
    unsigned char		MDay;			/* day of the month - [1,31] */
    unsigned char		Mon;			/* months since January - [0,11] */
    unsigned char		Year;			/* years since 1900 */
    unsigned char		WDay;			/* days since Sunday - [0,6] */
    unsigned char		IsDst;			/* daylight savings time flag */
    unsigned short		YDay;			/* days since January 1 - [0,365] */
}	CPU_DATETIME_Type;
#pragma pack()


/***************************************************************************************************/
/** GLOBAL VARIABLES  ******************************************************************************/

/* for synchronization */
int var1,var2,var3,var4 = 0;
ANSL_VARIABLE_EVENT_PARAMETER evParam;
unsigned int identCpu380 = 0;				/* ident of connection to cpu */
unsigned int identTask = 0;					/* ident of connection to task1 */
unsigned int identLogger = 0;			    /* ident of connection to $arlogsys */
unsigned int identVarRead = 0;				/* ident of connection to var1 */
unsigned int identVarSet = 0;				/* ident of connection to var2 */
unsigned short varSet = VAR_NAME_SET_VALUE;	/* value to set on target */
unsigned int status = 0;					/* state of function calls */
unsigned int userDataCpu = 380;			/* user date to identify the caller */
unsigned int userDataTask = 381;			/* user date to identify the caller */
unsigned int userDataLogger = 400;			/* user date to identify the caller */
unsigned int userDataRead = 1;				/* user date to identify the caller */
unsigned int userDataSet = 2;				/* user date to identify the caller */


#define T_PTHREAD pthread_t
#define T_PTHREAD_MUTEX pthread_mutex_t
pthread_attr_t _threadAttr;

#define T_PTHREADP T_PTHREAD*
#define T_PTHREAD_MUTEXP T_PTHREAD_MUTEX*


T_PTHREAD _ThreadId;
T_PTHREAD _ThreadId2;
T_PTHREAD _ThreadIdPrint;
T_PTHREAD_MUTEX _MutexObject;
T_PTHREAD_MUTEX _MutexObjectPrint;
int _bBreakThread;
int _CPUConnClosed;
int _bConnectionClosed;
int _RunningState;
int _RunningStatePrint;
char _strObj[10];
char _strPV1[100];
char _strPV2[100];
char _strTask[100];
char _strValue[1024];
ANSL_VARIABLE_DATA_TYPE _DataTypePV1;
ANSL_VARIABLE_DATA_TYPE _DataTypePV2;
char strIP[16]; 		/* IP-Adresse */
char strConPara[256]; 	/* Connection parameters */
unsigned int pvParentIdent;
int rfPV1;
int rfPV2;

unsigned long GetSystemTickCount(); // Zeit seit dem Booten in Sekunden
unsigned long GetSystemTickCountA(); // Zeit seit dem Booten in Sekunden
unsigned long GetSystemTickCountB(); // Zeit seit dem Booten in Sekunden
char* DataToString(char *pData, unsigned int dataLen, ANSL_VARIABLE_DATA_TYPE *pDataType);
int GetFormatString(ANSL_VARIABLE_DATA_TYPE* pTypeInfo, char strValue[FMT_STR_SIZE]);
int GetObjTypeString(unsigned int objType, char strValue[INFO_STR_SIZE]);
int GetCBTypeString(unsigned int cbType, char strValue[INFO_STR_SIZE]);
unsigned int ReadCpuDateTime(unsigned int cpuIdent);
unsigned int ReadVariable(unsigned int varIdent);
unsigned int WriteVariable(unsigned int varIdent, long varValue);
int connectCPU(char* strIP, char* strConPara, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData);
int connectTask(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData);
int connectLogger(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData);
int connectVariable(unsigned int parentID, char* name, unsigned int* ident, int flags, int refresh, ANSL_CALL_BACK fpCB, void* userData);
void AnslGetVersionInfo();
void TaskClassResume(unsigned int cpuIdent, unsigned short tcNumber);
void TaskClassSuspend(unsigned int cpuIdent, unsigned short tcNumber);
void TaskResume(unsigned int taskIdent, unsigned short numOfCycles);
void TaskSuspend(unsigned int taskIdent, unsigned short numOfCycles);
void LoggerClearBuffer(unsigned int loggerIdent);
void CpuClearMemory(unsigned int cpuIdent, ANSL_CPU_CLEARMEM_MODES memType);
static void PrintText(void *pParam);


int ioGetch();
unsigned int ConnectPV(char *name, unsigned int parentIdent, unsigned int evFlags, unsigned int refresh, void *pUserData, unsigned int *pIdent);
void DoWork();
void StartNoThread();

int CreateWorkerThread(int* pRunningState, T_PTHREAD_MUTEX* pMutex, T_PTHREAD* pThreadId, void* fpThread);
int CreatePrinterThread(int* pRunningState, T_PTHREAD_MUTEX* pMutex, T_PTHREAD* pThreadId, void* fpThread);

static void WorkerThread(void* pParam);

unsigned int ANSLCB_Cpu(unsigned int errorCode,
                          unsigned int ident,
                          unsigned int objectType,
                          unsigned int cbType,
                          void* pData,
                          unsigned int dataLen,
                          void* pUserData);

int ErrorCheck(unsigned int status, const char* error);
void DoExit();
int UninitializeObjects();

// Synchronization
/* thread synchronisation description (dependent on the platform): */
#define ERR_OK				0

#define MAX_PATH			260

// socket option error flags
#define WSANOTINITIALISED		1	/* A successful WSAStartup call must occur before using this function. */
#define WSAENETDOWN				2 	/* The network subsystem has failed. */
#define WSAEFAULT				3	/* One of the optval or the optlen parameters is not a valid part of the user address space, or the optlen parameter is too small. */
#define WSAEINPROGRESS			4	/* A blocking Windows Sockets 1.1 call is in progress, or the service provider is still processing a callback function. */
#define WSAEINVAL				5	/* The level parameter is unknown or invalid. */
#define WSAENOPROTOOPT			6	/* The option is unknown or unsupported by the indicated protocol family. */
#define WSAENOTSOCK				7	/* The descriptor is not a socket. */


#ifndef NULL
#ifdef __cplusplus
#define NULL    0
#else
#define NULL    ((void *)0)
#endif
#endif

#ifndef FALSE
#define FALSE               0
#endif

#ifndef TRUE
#define TRUE                1
#endif

#ifndef IN
#define IN
#endif

#ifndef OUT
#define OUT
#endif

#ifndef OPTIONAL
#define OPTIONAL
#endif

#undef far
#undef near
#undef pascal

#define far
#define near
#if (_MSC_VER >= 800) || defined(_STDCALL_SUPPORTED)
#define pascal __stdcall
#else
#define pascal
#endif

#ifdef DOSWIN32
#define cdecl _cdecl
#ifndef CDECL
#define CDECL _cdecl
#endif
#else
#define cdecl
#ifndef CDECL
#define CDECL
#endif
#endif

#if (_MSC_VER >= 800) || defined(_STDCALL_SUPPORTED)
#define CALLBACK    __stdcall
#define WINAPI      __stdcall
#define WINAPIV     __cdecl
#define APIENTRY    WINAPI
#define APIPRIVATE  __stdcall
#define PASCAL      __stdcall
#else
#define CALLBACK
#define WINAPI
#define WINAPIV
#define APIENTRY    WINAPI
#define APIPRIVATE
#define PASCAL      pascal
#endif

#define FAR                 far
#define NEAR                near
#ifndef CONST
#define CONST               const
#endif


/******************************************************************************
*  Base Data Types:
******************************************************************************/
#define bool_t	 unsigned char	  /* Boolean (0=false, 1=true)		*/
#define char_t	 char                   /* ASCII character representation	*/
#define CHAR	 char                   /* ASCII character representation	*/
#define TCHAR	 char                   /* ASCII character representation	*/
#define int_t		 int					  /* 16/32/64 bit signed integer		*/
#define float32_t	 float                   /* 32 bit floating point	*/
#define float64_t	 double                /* 64 bit floating point	*/
#define uint_t		 unsigned int		  /* 16/32/64 bit unsigned integer	*/
#define VOID		 void

typedef unsigned char BYTE;
typedef unsigned char near *PBYTE;
typedef unsigned char  far *LPBYTE;

#ifdef CHAR
#undef CHAR
#endif
typedef char CHAR;
typedef char near *PCHAR;
typedef char  far *LPCHAR;
typedef char *LPSTRING;

typedef unsigned short WORD;
typedef unsigned short near *PWORD;
typedef unsigned short far *LPWORD;

typedef unsigned int UINT;
typedef int near *PUINT;
typedef int far *LPUINT;

typedef int INT;
typedef int near *PINT;
typedef int far *LPINT;

typedef unsigned long DWORD;
typedef unsigned long near *PDWORD;
typedef unsigned long far *LPDWORD;

#ifdef LONG
#undef LONG
#endif
typedef long LONG;


typedef float FLOAT;
typedef float *PFLOAT;

#ifdef BOOL
#undef BOOL
#endif
typedef int BOOL;
typedef int near *PBOOL;
typedef int far *LPBOOL;



typedef long far *LPLONG;

typedef void far *LPVOID;
typedef const void far *LPCVOID;


typedef BYTE  		BOOLEAN;
typedef BOOLEAN 	*PBOOLEAN;
typedef CHAR *LPSTR;
typedef const CHAR *LPCSTR;

typedef unsigned long ULONG;
typedef unsigned long *PULONG;
typedef unsigned long *LPULONG;

typedef unsigned short USHORT;
typedef USHORT *PUSHORT;
typedef unsigned char UCHAR;
typedef UCHAR *PUCHAR;
typedef char *PSZ;

#define ANSL_NO_ERR		0
#define ANSL_BASE_ERR					10000

/******************************************************************************
*  ANSL Core Error Codes (reserved ANSL_BASE_ERR+0 - ANSL_BASE_ERR+99):
******************************************************************************/

#define ANSL_ERR_SYNC_RESOURCE				(ANSL_BASE_ERR+4)

#ifndef WIN32

typedef struct _SYSTEMTIME {
    WORD wYear;
    WORD wMonth;
    WORD wDayOfWeek;
    WORD wDay;
    WORD wHour;
    WORD wMinute;
    WORD wSecond;
    WORD wMilliseconds;
} SYSTEMTIME, *PSYSTEMTIME, *LPSYSTEMTIME;
#endif

#define SYSTEMTIME_t	SYSTEMTIME		/* system time declaration including MSec	*/

#define INFINITE		0xFFFFFFFF
#define ERR_OK			0
#define ERROR			-1
#define WAIT_FOREVER 	-1
#define FUNCPTR			void*
#define RESTRICTED_POINTER

/* invalid thread / task ID: */
#define SYS_INVALID_THREAD_ID	((uint_t) -1)

typedef void *PVOID;

/* Types use for passing & returning polymorphic values */
typedef UINT WPARAM;
typedef LONG LPARAM;
typedef LONG LRESULT;


#ifdef STRICT
typedef void *HANDLE;
#define DECLARE_HANDLE(name) struct name##__ { int unused; }; typedef struct name##__ *name
#else
typedef PVOID HANDLE;
#define DECLARE_HANDLE(name) typedef HANDLE name
#endif
typedef HANDLE *PHANDLE;

typedef struct _LIST_ENTRY {
   struct _LIST_ENTRY * volatile Flink;
   struct _LIST_ENTRY * volatile Blink;
} LIST_ENTRY, *PLIST_ENTRY, *RESTRICTED_POINTER PRLIST_ENTRY;

typedef struct _RTL_CRITICAL_SECTION_DEBUG {
    WORD   Type;
    WORD   CreatorBackTraceIndex;
    struct _RTL_CRITICAL_SECTION *CriticalSection;
    LIST_ENTRY ProcessLocksList;
    DWORD EntryCount;
    DWORD ContentionCount;
    DWORD Spare[ 2 ];
} RTL_CRITICAL_SECTION_DEBUG, *PRTL_CRITICAL_SECTION_DEBUG;



typedef struct _RTL_CRITICAL_SECTION {
    PRTL_CRITICAL_SECTION_DEBUG DebugInfo;

    //
    //  The following three fields control entering and exiting the critical
    //  section for the resource

    //LockCount field indicates the number of times that any thread has called the EnterCriticalSection
    LONG LockCount;
    //RecursionCount indicates the number of times that the owning thread has called EnterCriticalSection
    LONG RecursionCount;
    HANDLE OwningThread;        // from the thread's ClientId->UniqueThread
    HANDLE LockSemaphore;
    pthread_mutex_t SyncMutex;
} RTL_CRITICAL_SECTION, *PRTL_CRITICAL_SECTION;

typedef RTL_CRITICAL_SECTION CRITICAL_SECTION;
typedef PRTL_CRITICAL_SECTION PCRITICAL_SECTION;
typedef PRTL_CRITICAL_SECTION LPCRITICAL_SECTION;

/* thread synchronisation description (dependent on the platform): */


typedef struct _EVENT_OBJECT_T
{
    int is_set;
    BOOL is_manual_reset;
    pthread_cond_t   cond;
    pthread_mutex_t  mutex;
    pthread_t owningThread;
} EVENT_OBJECT, *LPEVENT_OBJECT;

/* win handle emulation */
typedef enum hType
{
    H_UNINITILIZED,
    H_SEM,
    H_EVENT,
    H_MUTEX,
    H_THREAD,
    H_CONDITION
}T_hType;

typedef struct _hdl_s
{
    enum hType       type;
    union _hdl_union
    {
        pthread_t thread;
        pthread_mutex_t mutex;
        EVENT_OBJECT event;
    } obj;

} hUnion;

typedef struct
{
    int          Created;
    bool_t       UseCriticalSection;
    pthread_t    threadID;
    DWORD        syncCount;
    union
    {
            CRITICAL_SECTION  CriticalSection;
            hUnion            hMutex;
    } Sync;
}	ANSL_SYNC_DESC;

typedef struct _SECURITY_ATTRIBUTES {
    DWORD nLength;
    LPVOID lpSecurityDescriptor;
    BOOL bInheritHandle;
} SECURITY_ATTRIBUTES, *PSECURITY_ATTRIBUTES, *LPSECURITY_ATTRIBUTES;

typedef struct
{
    int          Created;
    int       UseCriticalSection;
    pthread_t    threadID;
    int        syncCount;
    union
    {
            CRITICAL_SECTION  CriticalSection;
            hUnion            hMutex;
    } Sync;
}	T_SYNC_DESC;

WINBASEAPI int WINAPI XInitializeCriticalSection(LPCRITICAL_SECTION lpCriticalSection);
WINBASEAPI int WINAPI XCreateMutex(LPSECURITY_ATTRIBUTES lpSecAttribs, BOOL bManualReset,LPCSTR lpName, pthread_mutex_t* pMutex);
WINBASEAPI BOOL WINAPI XCloseHandle(hUnion* pHandle);
WINBASEAPI BOOL WINAPI XReleaseMutex(hUnion* pMutex);
WINBASEAPI void WINAPI XEnterCriticalSection(CRITICAL_SECTION *lpCriticalSection);
WINBASEAPI void WINAPI XLeaveCriticalSection(CRITICAL_SECTION *lpCriticalSection);
WINBASEAPI void WINAPI XDeleteCriticalSection(LPCRITICAL_SECTION lpCriticalSection);

static SECURITY_ATTRIBUTES g_pSecAttr;

pthread_t sysGetThreadId ();
int sysEnterSync (T_SYNC_DESC*	pSyncDesc);
int sysLeaveSync (T_SYNC_DESC*	pSyncDesc);
void sysCloseSync (T_SYNC_DESC*	pSyncDesc);
int sysCheckSync (T_SYNC_DESC*	pSyncDesc);
void sysSleep(unsigned int milliSeconds);
int sysCreateSync (T_SYNC_DESC* pSyncDesc, char_t* pIdName);




static T_SYNC_DESC	g_SyncDesc;
//static T_SYNC_DESC	g_SyncVar;

static char				g_PrintBf[1024*1024];
static volatile int		g_PrintOffs = 0;
int g = 0;



void xprintf (const char* pText, ...)
{
    
	
	va_list	argList;
    char*	pPrintBf;
    char	szTextBf[6000];


    sysEnterSync (&g_SyncDesc);
	pPrintBf = g_PrintBf;

    va_start (argList, pText);
    vsprintf (szTextBf, pText, argList);
    va_end (argList);


    if (g_PrintOffs < (sizeof (g_PrintBf) - 6000))
    {
        pPrintBf += g_PrintOffs;
        strcpy (pPrintBf, szTextBf);
        g_PrintOffs += strlen (pPrintBf);
    }

    sysLeaveSync (&g_SyncDesc);
}

void xprintfCallback (const char* pText, ...)
{
    va_list	argList;
    char	szTextBf[12000];


    sysEnterSync (&g_SyncDesc);

    va_start (argList, pText);
    vsprintf (szTextBf, pText, argList);
	va_end (argList);

    printf("%s \r", szTextBf);

    sysLeaveSync (&g_SyncDesc);
}

void PrintVInfo() {

    unsigned int uiRedV = 0;
    unsigned int uiUifV = 0;
    unsigned int uiOnlV = 0;
    unsigned short redREL = 0;
    unsigned char  redMaj = 0;
    unsigned char  redMin= 0;
    unsigned short redBld = 0;
    unsigned short uifREL = 0;
    unsigned char  uifMaj = 0;
    unsigned char  uifMin= 0;
    unsigned short uifBld = 0;
    ANSLI_GetVersionInfos(&uiUifV, &uiOnlV);
    ANSLI_DecodeVersionInfo(uiUifV, &uifREL, &uifMaj, &uifMin, &uifBld);
    uiRedV = ANSLI_GetRedundancyVersion();
    ANSLI_DecodeVersionInfo(uiRedV, &redREL, &redMaj, &redMin, &redBld);

    xprintf("UIF: V%d.%d.%d.%d  REDUNDANCY: V%d.%d.%d.%d  ONLINE: V%u\r\n",
                    uifREL,
                    uifMaj,
                    uifMin,
                    uifBld,
                    redREL,
                    redMaj,
                    redMin,
                    redBld,
                    uiOnlV);

}

/***************************************************************************************************/
/** MAIN PROGRAM ***********************************************************************************/

int main(int argc,char *argv[])
{
    int logLevel = 0;
    unsigned long t1, t2, tH, tM, tS, tMS;
    t1 = GetSystemTickCount();
    xprintf("SystemTickCount=%lu\n", t1);

    _bBreakThread = 0;
    _RunningState = 0;
    _RunningStatePrint = 0;
    _CPUConnClosed = 0;
    _bConnectionClosed = 0;
    memset(strIP, 0, 16);
    memset(strConPara, 0, 256);

    sysCreateSync (&g_SyncDesc, NULL);


    /** begin: init and connect **/
    sprintf(strIP, "%s", (argc >= 2) ? argv[1] : TARGET_IP); /* IP-Adresse */
    if (argc >= 4)
    {
        sprintf(strConPara, "Port=\"%s\" Timeout=\"%s\" %s", argv[3], argv[2], CONNECTION_PARAMS_NO_TOPO);
    } else if (argc >= 3) {
        sprintf(strConPara, "Timeout=\"%s\" %s", argv[2], CONNECTION_PARAMS_NO_TO);
    } else {
        sprintf(strConPara, "%s", CONNECTION_PARAMS);
    }
    if (argc >= 10)
    {
        sprintf(_strPV1, "%s", argv[4]);
        sprintf(_strPV2, "%s", argv[5]);
        rfPV1 = atoi(argv[6]);
        rfPV2 = atoi(argv[7]);
        sprintf(_strTask, "%s", argv[8]);
        if ('.' == _strTask[0])
        {
            _strTask[0] = 0;
        }
        logLevel = atoi(argv[9]);
    } else if (argc >= 9)
    {
        sprintf(_strPV1, "%s", argv[4]);
        sprintf(_strPV2, "%s", argv[5]);
        rfPV1 = atoi(argv[6]);
        rfPV2 = atoi(argv[7]);
        sprintf(_strTask, "%s", argv[8]);
    } else if (argc >= 8) {
        sprintf(_strPV1, "%s", argv[4]);
        sprintf(_strPV2, "%s", argv[5]);
        rfPV1 = atoi(argv[6]);
        rfPV2 = atoi(argv[7]);
    } else if (argc >= 7) {
        sprintf(_strPV1, "%s", argv[4]);
        sprintf(_strPV2, "%s", argv[5]);
        rfPV1 = atoi(argv[6]);
    } else if (argc >= 6) {
        sprintf(_strPV1, "%s", argv[4]);
        sprintf(_strPV2, "%s", argv[5]);
    } else if (argc >= 5){
        sprintf(_strPV1, "%s", argv[4]);
        sprintf(_strPV2, "%s", VAR_NAME_SET);
    } else {
        sprintf(_strPV1, "%s", VAR_NAME_READ);
        sprintf(_strPV2, "%s", VAR_NAME_SET);
    }


    xprintf("+ANSLSample  IPAddress=\"%s\"\n", strIP);
    xprintf("             %s\n", strConPara);
    xprintf("             Task=\"%s\" PV1=\"%s\" RF1=%d PV2=\"%s\" RF2=%d   LOGGING=%d\n\n", _strTask, _strPV1, rfPV1, _strPV2, rfPV2, logLevel);
    xprintf("Press h to get the command line options!\n");
    xprintf("Press 0 to stop ANSL communication!\n");
    xprintf("Press 1 to read CPU date and time!\n");
    xprintf("Press 2 to read PV1!\n");
    xprintf("Press 3 to read PV2!\n");
    xprintf("Press 4 to PV1++ and PV2++!\n");
    xprintf("Press 5 to set PV1!\n");
    xprintf("Press 6 to set PV2!\n");
    xprintf("Press 7 to PV1-- and PV2--!\n");
    xprintf("Press 8 for WarmRestart!\n");
    xprintf("Press 9 for ColdRestart!\n");
    xprintf("Press a for get ANSL user and online interface version info!\n");
    xprintf("Press b for resume task class #1!\n");
    xprintf("Press c for suspend task class #1!\n");
    xprintf("Press d for resume task for 10 cycles!\n");
    xprintf("Press e for resume task for 10 cycles!\n");
    xprintf("Press f for clear logger buffer!\n");
    xprintf("Press g for clear CPU RAM!\n");
    xprintf("Press i for reading CPU info!\n");
    xprintf("Press j for reading HW info!\n");
    xprintf("Press k for reading memory info!\n");
    xprintf("\n");

    xprintf("ANSL console TestClient for LINUX %d bit\n", (sizeof(void*)*8));
    PrintVInfo();

    /* turn on logging*/
    ANSLI_LOGGING_SetLevel(logLevel);

    /* init ANSL */
    status = ANSLI_Initialize();
    if (ErrorCheck(status, "ERROR ANSLI_Initialize") != 0) return -1;

#ifndef _NO_AUTO_CONNECT
    /* connect to CPU object */
    status = connectCPU(strIP, strConPara, (unsigned int*)&identCpu380, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataCpu);
    if (0 != status) { return -1; }

    pvParentIdent = identCpu380;


    /* connect to task object */
    if (0 != _strTask[0])
    {
        connectTask(identCpu380, _strTask, (unsigned int*)&identTask, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataTask);
        if (0 != status) { return -1; }

        pvParentIdent = identTask;
    }


    /* connect to task object */
    connectLogger(identCpu380, "$arlogsys", (unsigned int*)&identLogger, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataLogger);
    if (0 != status) { return -1; }


    // connect to variable VAR_NAME_READ
    if (0 != _strPV1[0])
    {
        connectVariable(pvParentIdent, _strPV1, (unsigned int*)&identVarRead, ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV1, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataRead);
        if (0 != status) { return -1; }
    }

    /* connect to variable VAR_NAME_SET */
    if (0 != _strPV2[0])
    {
        connectVariable(pvParentIdent, _strPV2, (unsigned int*)&identVarSet, ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV2, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataSet);
        if (0 != status) { return -1; }
    }


#endif // !#ifdef _NO_AUTO_CONNECT


    // either start a new thread
    CreatePrinterThread(&_RunningState, &_MutexObject, &_ThreadId2, (void *) PrintText);
    CreateWorkerThread(&_RunningState, &_MutexObject, &_ThreadId, (void *) WorkerThread);

    pthread_join(_ThreadId, NULL);

    t2 = GetSystemTickCount();
    tS = ((t2-t1)/1000);
    tMS = ((t2-t1)%1000);
    tM = tS/60;
    tS = tS%60;
    tH = tM/60;
    tM = tM%60;
    xprintf("-ANSLSample    Execution time: %luh %lum %lus %lums\n", tH, tM, tS, tMS);

    sysCloseSync (&g_SyncDesc);

    /** begin: deinit and disconnect **/
    return EXIT_SUCCESS;
}

/* connect to CPU object */
int connectCPU(char* strIP, char* strConPara, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData)
{
    int status = 0;
    status = ANSLI_CPU_Connect(strIP, strConPara, (unsigned int*)&identCpu380, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataCpu);
    if (ErrorCheck(status, "ERROR ANSLI_CPU_Connect") != 0)
    {
        return -1;
    }
    return status;
}

/* connect to task object */
int connectTask(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData)
{
    int status = 0;
    if (0 != name[0])
    {
        status = ANSLI_TASK_Connect(parentID, name, ident, (ANSL_CALL_BACK)fpCB, userData);
        if(ErrorCheck(status, "ERROR ANSLI_TASK_Connect TASK_NAME") != 0)
        {
            return -1;
        }
    }
    return status;
}

/* connect to task object */
int connectLogger(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData)
{
    int status = 0;
    if (0 != name[0])
    {
        status = ANSLI_LOGGER_Connect(parentID, name, ident, ANSL_LOGM_EV_FLAG_NONE, (ANSL_CALL_BACK)fpCB, userData);
        if(ErrorCheck(status, "ERROR ANSLI_LOG_Connect LOGGER_NAME") != 0)
        {
            return -1;
        }
    }
    return status;
}

/* connect to variable object */
int connectVariable(unsigned int parentID, char* name, unsigned int* ident, int flags, int refresh, ANSL_CALL_BACK fpCB, void* userData)
{
    int status = 0;
    ConnectPV(name, parentID, flags, refresh, userData, ident);
    status = ANSLI_VAR_ReadDataType(*ident);
    if(ErrorCheck(status, "ERROR ANSLI_VAR_ReadDataType VAR_NAME_READ") != 0)
    {
        return -1;
    }
    return status;
}


unsigned long GetSystemTickCount() // Zeit seit dem Booten in Sekunden
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return (unsigned long)(tv.tv_sec*1000 + (tv.tv_usec / 1000));
}

unsigned long GetSystemTickCountA() // Zeit seit dem Booten in Sekunden
{
    struct timespec abstime;
    clock_gettime(CLOCK_REALTIME, &abstime);
    return (unsigned long)(abstime.tv_sec*1000 + (abstime.tv_nsec / 1000));
}

unsigned long GetSystemTickCountB() // Zeit seit dem Booten in Sekunden
{
    struct timeb tp;
    ftime (&tp);
    return (unsigned long)(tp.time*1000 + tp.millitm);
}

int CreateWorkerThread(int* pRunningState, T_PTHREAD_MUTEX* pMutex, T_PTHREAD* pThreadId, void* fpThread)
{
    int retVal = 0;
    *pRunningState = 0;

    pthread_mutexattr_t mutex_attr;
    pthread_attr_t	threadAttr;

    /* create event objekt: */
    pthread_mutexattr_init (&mutex_attr);
    retVal = pthread_mutex_init(pMutex, &mutex_attr);
    if(retVal < 0)
    {
        xprintf( "Mutex NOT created.\n" );
        return RC_OBJECT_NOT_CREATED;
    }

    retVal = pthread_mutex_lock (pMutex);
    if (-1 == retVal)
    {
       xprintf( "Mutex LOCK ERROR.\n" );
       return RC_LOCK_ERROR;
    }

    retVal = pthread_attr_init(&threadAttr);
    if(retVal < 0)
    {
        xprintf( "pthread_attr_init ERROR.\n" );
        return RC_THREAD_ATTR_ERROR;
    }

    retVal = pthread_attr_setstacksize(&threadAttr, 120*1024);
    if(retVal < 0)
    {
        xprintf( "pthread_attr_setstacksize ERROR.\n" );
        return RC_STACKSIZE_ERROR;
    }

    retVal = pthread_mutex_unlock (pMutex);
    if (-1 == retVal)
    {
       xprintf( "Mutex UNLOCK ERROR.\n" );
       return RC_UNLOCK_ERROR;
    }

    //xprintf("pthread_create...\n");
    retVal = pthread_create(pThreadId, NULL/*&threadAttr*/, (void *(*)(void *))fpThread, (void*)pMutex);    

    if(retVal != 0)
    {
        pthread_mutex_destroy(pMutex);
        xprintf( "Thread NOT created!\n" );
        return RC_THREAD_NOT_CREATED;
    }

    return status;
}


static void WorkerThread(void* pParam)
{

    T_PTHREAD_MUTEX* pMutex = (T_PTHREAD_MUTEX*)pParam;
    int status;


    // Nun kann losgeloest von der restlichen Kommunikation
    // mit der Datenaufbereitung begonnen werden


    while (0 == _bBreakThread)
    {
        status = pthread_mutex_lock(pMutex);
        if (0 != status)
        {
        	xprintf("pthread_mutex_lock   RC_LOCK_ERROR!\n");
        }        
        pthread_mutex_unlock(pMutex);
        DoWork();
    }
    UninitializeObjects();
    while (0 == _bConnectionClosed)
    {
        status = pthread_mutex_lock(pMutex);
        if (0 != status)
        {
        	xprintf("pthread_mutex_lock   RC_LOCK_ERROR!\n");
        }
        DoExit();        
        pthread_mutex_unlock(pMutex);
    }

}

int CreatePrinterThread(int* pRunningState, T_PTHREAD_MUTEX* pMutex, T_PTHREAD* pThreadId, void* fpThread)
{
    int retVal = 0;
    *pRunningState = 0;

    pthread_mutexattr_t mutex_attr;
    pthread_attr_t	threadAttr;

    /* create event objekt: */
    pthread_mutexattr_init (&mutex_attr);
    retVal = pthread_mutex_init(pMutex, &mutex_attr);
    if(retVal < 0)
    {
        xprintf( "Mutex NOT created.\n" );
        return RC_OBJECT_NOT_CREATED;
    }

    retVal = pthread_mutex_lock (pMutex);
    if (-1 == retVal)
    {
       xprintf( "Mutex LOCK ERROR.\n" );
       return RC_LOCK_ERROR;
    }

    retVal = pthread_attr_init(&threadAttr);
    if(retVal < 0)
    {
        xprintf( "pthread_attr_init ERROR.\n" );
        return RC_THREAD_ATTR_ERROR;
    }

    retVal = pthread_attr_setstacksize(&threadAttr, 120*1024);
    if(retVal < 0)
    {
        xprintf( "pthread_attr_setstacksize ERROR.\n" );
        return RC_STACKSIZE_ERROR;
    }

    retVal = pthread_mutex_unlock (pMutex);
    if (-1 == retVal)
    {
       xprintf( "Mutex UNLOCK ERROR.\n" );
       return RC_UNLOCK_ERROR;
    }

    //xprintf("pthread_create...\n");
    retVal = pthread_create(pThreadId, NULL/*&threadAttr*/, (void *(*)(void *))fpThread, (void*)pMutex);

    if(retVal != 0)
    {
        pthread_mutex_destroy(pMutex);
        xprintf( "Thread NOT created!\n" );
        return RC_THREAD_NOT_CREATED;
    }

    return status;
}

void DoWork()
{
    unsigned int status = 0;
    int charRead = 0;
        switch (_RunningState)
        {
            case 0:
            {
                _RunningState = 1;
            } break;
            case 1:
            {
                _RunningState = 2;
            } break;
            case 2:
            {                  
                    charRead = ioGetch();       
                    switch (charRead)
                    {
                        case 48:
                        {
                            xprintf("Exiting...\n");
                            _RunningState = 3;
                        } break;
                        case 97:
                        {
                            xprintf("Getting user and online interface version info...\n\r");
                            AnslGetVersionInfo();
                        } break;
                        case 98:
                        {
                            xprintf("Resuming TaskClass...\n");
                            TaskClassResume(identCpu380, 1);
                        } break;
                        case 99:
                        {
                            xprintf("Suspend TaskClass...\n");
                            TaskClassSuspend(identCpu380, 1);
                        } break;
                        case 100:
                        {
                            xprintf("Resuming Task...\n");
                            TaskResume(identTask, 10);
                        } break;
                        case 101:
                        {
                            xprintf("Suspend Task...\n");
                            TaskSuspend(identTask, 10);
                        } break;
                        case 102:
                        {
                            xprintf("Clear Buffer for Logger...\n");
                            LoggerClearBuffer(identLogger);
                        } break;
                        case 103:
                        {
                            xprintf("Clear CPU RAM...\n");
                            CpuClearMemory(identCpu380, ANSL_CPU_CLEAR_RAM);
                        } break;
                        case 104:
                        {
                            xprintf("Optional parameters when executing this demo:\n\r");
                            xprintf("[ip address] [timeout] [port] [1st PV] [2nd PV] [refresh 1st PV] [refresh 2nd PV] [task] [verbose level for logging]\n\r");
                            xprintf("If task name is given assume the variables are TASK local else CPU global!\n\r");
                            xprintf("When using logging for global variables set task to \".\"!\n");
                        } break;
                        case 105:
                        {
                            xprintf("Reading CPU info...\n");
                            status = ANSLI_CPU_ReadInfo(identCpu380);
                            if (status != 0)
                            {
                                xprintf("ERROR CPU read info: %i\n",status);                                
                            }



                        } break;
                        case 106:
                        {
                            xprintf("Reading HW info...\n\r");
                            status = ANSLI_CPU_GetHWComponents(identCpu380);
                            if (status != 0)
                            {
                                xprintf("ERROR CPU read HW info: %i\n",status);
                            }


                        } break;
                        case 107:
                        {
                            xprintf("Reading memory info...\n");
                            status = ANSLI_CPU_ReadMemInfo(identCpu380);
                            if (status != 0)
                            {
                                xprintf("ERROR CPU read memory info: %i\n",status);                                
                            }
                        } break;
                        case 49:
                        {
                            xprintf("reading CPU date and time...\n");
                            ReadCpuDateTime(identCpu380);
                        } break;
                        case 50:
                        {
                            xprintf("reading \"%s\"...\n\r", _strPV1);
                            ReadVariable(identVarRead);
                        } break;
                        case 53:
                        {
                            xprintf("writing \"%s\"...\n\r", _strPV1);
                            WriteVariable(identVarRead, var3);
                        } break;
                        case 51:
                        {
                            xprintf("reading \"%s\"...\n\r", _strPV2);
                            ReadVariable(identVarSet);
                        } break;
                        case 54:
                        {
                            xprintf("writing \"%s\"...\n\r", _strPV2);
                            WriteVariable(identVarSet, var3);
                        } break;
                        case 52:
                        {
                            var3++;
                            var4++;
                            xprintf("increment write buffers   %d\n", var3);
                        } break;
                        case 55:
                        {
                            var3--;
                            var4--;
                            xprintf("decrement write buffers   %d\n", var3);
                        } break;
                        case 56:
                        {
                            xprintf("WarmRestart\n");
                            status = ANSLI_CPU_Restart(identCpu380, ANSL_CPU_WARM_RESTART);
                            ErrorCheck(status, "ANSLI_CPU_Restart");
                        } break;
                        case 57:
                        {
                            xprintf("ColdRestart\n");
                            status = ANSLI_CPU_Restart(identCpu380, ANSL_CPU_COLD_RESTART);
                            ErrorCheck(status, "ANSLI_CPU_Restart");
                        } break;
        #ifdef _NO_AUTO_CONNECT
                        case 99:
                        {
                            xprintf("Connect CPU\n");
                            status = connectCPU(strIP, strConPara, (unsigned int*)&identCpu380, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataCpu);
                            ErrorCheck(status, "Connect CPU");
                            pvParentIdent = identCpu380;

                        } break;
                        case 116:
                        {
                            xprintf("connect to task object\n");
                            connectTask(identCpu380, _strTask, (unsigned int*)&identTask, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataTask);
                            ErrorCheck(status, "Connect TASK");
                            pvParentIdent = identTask;                            
                        } break;
                        case 114:
                        {
                            xprintf("connect to variable VAR_NAME_READ\n");
                            connectVariable(pvParentIdent, _strPV1, (unsigned int*)&identVarRead, ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV1, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataRead);
                            ErrorCheck(status, "Connect VAR_NAME_READ");                            
                        } break;
                        case 119:
                        {
                            xprintf("connect to variable VAR_NAME_SET\n");
                            connectVariable(pvParentIdent, _strPV2, (unsigned int*)&identVarSet, ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV2, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataSet);
                            ErrorCheck(status, "Connect VAR_NAME_SET");                            
                        } break;
        #endif //_NO_AUTO_CONNECT
                        default:
                        {

                        }break;
                    }
                } break;
            case 3:
            {
                _bBreakThread = 1;
            } break;
        }    
}

static void PrintText(void *threadId)
{
    while(1)
    {
		sysEnterSync (&g_SyncDesc);

        if (g_PrintOffs > 0)
            printf ("%s \r", g_PrintBf);

        g_PrintOffs = 0;

        sysLeaveSync (&g_SyncDesc);
    }
}

void StartNoThread()
{
    //read date and time of the cpu
    status = ANSLI_CPU_DTRead(identCpu380);
    ErrorCheck(status, "ANSLI_CPU_DTRead");

    while (0 == _bBreakThread)
    {
         DoWork();
         //PrintText();
    }
    UninitializeObjects();
    while (0 == _bConnectionClosed)
    {
        DoExit();
    }
    /** end: working **/
}

int UninitializeObjects()
{
    int status = 0;
    _bConnectionClosed = 1;
    /* cancel the PV connection */
    if (0 != identVarRead)
    {
        status = ANSLI_VAR_Disconnect(identVarRead);
        if (ErrorCheck(status, "ERROR ANSLI_VAR_Disconnect") != 0) return -1;
    }
    if (0 != identVarSet)
    {
        status = ANSLI_VAR_Disconnect(identVarSet);
        if (ErrorCheck(status, "ERROR ANSLI_VAR_Disconnect") != 0) return -1;
    }

    /* cancel the task connection */
    if (0 != identTask)
    {
        status = ANSLI_TASK_Disconnect(identTask);
        if (ErrorCheck(status, "ERROR ANSLI_TASK_Disconnect") != 0) return -1;
    }

    /* cancel the CPU connection */
    if (0 != identCpu380)
    {
        status = ANSLI_CPU_Disconnect(identCpu380);
        if (ErrorCheck(status, "ERROR ANSLI_CPU_Disconnect") != 0) return -1;
        _bConnectionClosed = 0;
    }

    /* deinit ANSL */
    // Wait for deinit in CPU callback!!

    return 0;
}

void DoExit()
{
    if (1 == _CPUConnClosed)
    {
        ANSLI_DeInitialize();
        _bConnectionClosed = 1;
    }
}

unsigned int ReadCpuDateTime(unsigned int cpuIdent)
{
    unsigned int status = 0;

    status = ANSLI_CPU_DTRead(cpuIdent);
    if (status != 0)
    {
        xprintf("ERROR READ DT : %i\n",status);        
    }
    return status;
}

unsigned int ConnectPV(char *name, unsigned int parentIdent, unsigned int evFlags, unsigned int refresh, void *pUserData, unsigned int *pIdent)
{
    unsigned int status = 0;
    struct ANSL_VARIABLE_EVENT_PARAMETER 	evParam;
    evParam.m_EventFlags = evFlags;
    evParam.m_RefreshTime = refresh;
    evParam.m_IntegerHyst = 0;
    evParam.m_FloatHyst = 0;

    status = ANSLI_VAR_Connect(parentIdent, name, pIdent, (ANSL_CALL_BACK)ANSLCB_Cpu, pUserData, (ANSL_VARIABLE_EVENT_PARAMETER*)&evParam);
    if (0 != status)
    {
        xprintf("ANSLI_VAR_Connect  \"%s\"   E#%i\n", name, status);
    }
    return status;
}

unsigned int ReadVariable(unsigned int varIdent)
{
    unsigned int status = 0;

    //xprintf("Reading PV ident=%u \n", varIdent);
    status = ANSLI_VAR_ReadValue(varIdent, NULL, 0);
    if (status !=0)
    {
        xprintf("ERROR %u READ PV ident=%u \n", status, varIdent);        
    }
    return status;
}

unsigned int WriteVariable(unsigned int varIdent, long varValue)
{
    unsigned int status = 0;

    xprintf("Write PV ident=%u to %d \n", varIdent, (int)varValue);
    status = ANSLI_VAR_SetValue(varIdent, &varValue, sizeof(varValue));
    if (status !=0)
    {
        xprintf("ERROR %u SET PV ident=%u \n", status, varIdent);        
    }
    return status;
}

struct termios orig_termios;

void reset_terminal_mode()
{
    tcsetattr(0, TCSANOW, &orig_termios);
}

void set_conio_terminal_mode()
{
    struct termios new_termios;

    /* take two copies - one for now, one for later */
    tcgetattr(0, &orig_termios);
    memcpy(&new_termios, &orig_termios, sizeof(new_termios));

    /* register cleanup handler, and set the new terminal mode */
    atexit(reset_terminal_mode);
    cfmakeraw(&new_termios);
    tcsetattr(0, TCSANOW, &new_termios);
}


int kbhit()
{
    struct timeval tv = { 0L, 0L };
    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(0, &fds);
    return select(1, &fds, NULL, NULL, &tv);
}

int getch()
{
    int r;
    unsigned char c;
    if ((r = read(0, &c, sizeof(c))) < 0) {
        return r;
    } else {
        return c;
    }
}


int ioGetch ()
{
     int ch = 0;
     
	 BOOL setChar = FALSE;

     sysEnterSync (&g_SyncDesc);
     set_conio_terminal_mode();
     sysLeaveSync (&g_SyncDesc);

     while (kbhit() && !setChar)
     {
         sysEnterSync (&g_SyncDesc);
         ch = getch();
         sysLeaveSync (&g_SyncDesc);
         setChar = TRUE;
     }

     // Zeichen zurueckgeben
     return ch;
}

/***************************************************************************************************/
/** ERROR HANDLING *********************************************************************************/

int ErrorCheck(unsigned int status, const char* error)
{
    if (status != 0)
    {
        xprintf("%s: %i\n", error, status);
        return 1;
    }
    return 0;
}


/***************************************************************************************************/
/** CALLBACK FUNCTION ******************************************************************************/
unsigned int ANSLCB_Cpu(unsigned int errorCode,
                        unsigned int ident,
                        unsigned int objectType,
                        unsigned int cbType,
                        void* pData,
                        unsigned int dataLen,
                        void* pUserData)
{
    char strOut[6000];
    char strName[128];
    char strFmt[FMT_STR_SIZE];
    char strTyp[INFO_STR_SIZE];
    char strEvent[INFO_STR_SIZE];
    memset(strOut, 0, 6000);
    memset(strName, 0, 128);
    memset(strFmt, 0, FMT_STR_SIZE);
    memset(strTyp, 0, INFO_STR_SIZE);
    memset(strEvent, 0, INFO_STR_SIZE);

    GetObjTypeString(objectType, strTyp);
    switch (cbType)
    {
        case ANSL_RESP_CONNECT:
        {
            xprintfCallback("CONNECT  %s error=%u ident=%u   dataLen=%u\n",
                            strTyp, errorCode, ident, dataLen);
        } break;
        case ANSL_RESP_DISCONNECT:
        {
            xprintfCallback("DISCONNECT  %s error=%u ident=%u   dataLen=%u\n",
                            strTyp, errorCode, ident, dataLen);
            _CPUConnClosed = 1;
        } break;
        case ANSL_EVENT_ERROR:
        {
            xprintfCallback("ERROR  %s error=%u ident=%u   dataLen=%u\n",
                strTyp, errorCode, ident, dataLen);
        } break;
        case ANSL_EVENT_STATUS:
        {
            xprintfCallback("STATUS  %s error=%u ident=%u   dataLen=%u\n",
                            strTyp, errorCode, ident, dataLen);
        } break;
        case ANSL_EVENT_FINAL:
        {
            xprintfCallback("FINAL  %s error=%u ident=%u   serviceNumber=%u\n",
                            strTyp, errorCode, ident, dataLen);
        } break;
        /* read date and time */
        case ANSL_RESP_GET_DATETIME:
        {
            CPU_DATETIME_TYPE cpuDataTime;

            if	(
                    (errorCode == 0) &&
                    (sizeof(CPU_DATETIME_TYPE) == dataLen)
                )
            {
                memcpy(&cpuDataTime, pData, dataLen);
                //                        year starts at 1900         month starts at 0
                xprintfCallback("Date: %d/%d/%d ", cpuDataTime.tm_Year + 1900, cpuDataTime.tm_Month + 1, cpuDataTime.tm_DayOfMonth);
                xprintfCallback("Time: %d:%d:%d\n", cpuDataTime.tm_Hour, cpuDataTime.tm_Minute, cpuDataTime.tm_Second);
            }
            else
            {
                xprintfCallback("ERROR GET_DATETIME: %i\n", errorCode);
            }
        } break;

        /* read READ_VAR */
        case ANSL_RESP_GET_VALUE:
        /* read READ_VAR */
        case ANSL_EVENT_VALUE:
        {
                if (*((unsigned int*)pUserData) == 1)
                {
                    sprintf(strName, "%s", _strPV1);
                }
                if (*((unsigned int*)pUserData) == 2)
                {
                    sprintf(strName, "%s", _strPV2);
                }
                GetCBTypeString(cbType, strEvent);
                xprintfCallback("%s \"%s\": %s  error=%u ident=%u   dataLen=%u\n",
                                        strEvent,
                                        strName,
                                        ((0 == errorCode) ? DataToString((char *)pData, dataLen, ((*((unsigned int*)pUserData) == 1)? &_DataTypePV1 : &_DataTypePV2)) : "ERROR"),
                                        errorCode,
                                        ident,
                                        dataLen);
        } break;

        /* read format SET_VAR */
        case ANSL_EVENT_FORMAT:
        case ANSL_RESP_GET_FORMAT:
        {
            if (*((unsigned int*)pUserData) == 1)
            {
                memcpy(&_DataTypePV1, pData, sizeof(ANSL_VARIABLE_DATA_TYPE));
                sprintf(strName, "%s", _strPV1);
            }
            if (*((unsigned int*)pUserData) == 2)
            {
                memcpy(&_DataTypePV2, pData, sizeof(ANSL_VARIABLE_DATA_TYPE));
                sprintf(strName, "%s", _strPV2);
            }
            memset (strFmt, 0, FMT_STR_SIZE);
            if (0 == errorCode)
            {
                GetFormatString((ANSL_VARIABLE_DATA_TYPE*)pData, strFmt);
            } else {
                sprintf(strFmt, "???");
            }
            GetCBTypeString(cbType, strEvent);
            xprintfCallback("%s  %s \"%s\": %s  error=%u ident=%u   dataLen=%u\n",
                                strTyp,
                                strEvent,
                                strName,
                                strFmt,
                                errorCode,
                                ident,
                                dataLen);
        } break;
        case ANSL_RESP_SET_ATTRIBUTES:
        {
            if (*((unsigned int*)pUserData) == 1)
            {
                sprintf(strName, "%s", _strPV1);
            }
            if (*((unsigned int*)pUserData) == 2)
            {
                sprintf(strName, "%s", _strPV2);
            }
            GetCBTypeString(cbType, strEvent);
            xprintfCallback("%s \"%s\": error=%u ident=%u\n",
                                strEvent,
                                strName,
                                errorCode,
                                ident);
        } break;
        /* set SET_VAR */
        case ANSL_RESP_SET_VALUE:
        {
            if (*((unsigned int*)pUserData) == 1)
            {
                sprintf(strName, "%s", _strPV1);
            }
            if (*((unsigned int*)pUserData) == 2)
            {
                sprintf(strName, "%s", _strPV2);
            }
            GetCBTypeString(cbType, strEvent);
            xprintfCallback("%s \"%s\": error=%u ident=%u\n",
                                strEvent,
                                strName,
                                errorCode,
                                ident);
        } break;
        case ANSL_RESP_GET_HARDWARE:
        {
            xprintfCallback("%s \n", (char*)pData);
        } break;
        case ANSL_RESP_GET_INFO:
        {
            xprintfCallback("%s\n", (char*)pData);
        } break;
        case ANSL_RESP_GET_STORAGE:
        {
            xprintfCallback("%s\n", (char*)pData);
        } break;
        default:
        {
            GetCBTypeString(cbType, strEvent);
            xprintfCallback("%s  %s error=%u ident=%u   dataLen=%u\n",
                                strTyp,
                                strEvent,
                                errorCode,
                                ident,
                                dataLen);
        } break;
    }    
    return 0;
}

int GetCBTypeString(unsigned int eventCode, char strValue[INFO_STR_SIZE])
{
    memset(strValue, 0, INFO_STR_SIZE);
    switch (eventCode)
    {
        case ANSL_EVENT_FINAL: { strcpy(strValue, "EVENT_FINAL"); } break;
        case ANSL_EVENT_ERROR: { strcpy(strValue, "EVENT_ERROR"); } break;
        case ANSL_EVENT_STATUS: { strcpy(strValue, "EVENT_STATUS"); } break;
        case ANSL_REQ_CONNECT: { strcpy(strValue, "REQ_CONNECT"); } break;
        case ANSL_RESP_CONNECT: { strcpy(strValue, "RESP_CONNECT"); } break;
        case ANSL_CANCEL_CONNECT: { strcpy(strValue, "CANCEL_CONNECT"); } break;
        case ANSL_REQ_DISCONNECT: { strcpy(strValue, "REQ_DISCONNECT"); } break;
        case ANSL_RESP_DISCONNECT: { strcpy(strValue, "RESP_DISCONNECT"); } break;
        case ANSL_CANCEL_DISCONNECT: { strcpy(strValue, "CANCEL_DISCONNECT"); } break;
        case ANSL_REQ_GET_DATETIME: { strcpy(strValue, "REQ_GET_DATETIME"); } break;
        case ANSL_RESP_GET_DATETIME: { strcpy(strValue, "RESP_GET_DATETIME"); } break;
        case ANSL_CANCEL_GET_DATETIME: { strcpy(strValue, "CANCEL_GET_DATETIME"); } break;
        case ANSL_REQ_SET_DATETIME: { strcpy(strValue, "REQ_SET_DATETIME"); } break;
        case ANSL_RESP_SET_DATETIME: { strcpy(strValue, "RESP_SET_DATETIME"); } break;
        case ANSL_CANCEL_SET_DATETIME: { strcpy(strValue, "CANCEL_SET_DATETIME"); } break;
        case ANSL_REQ_WARM_RESTART: { strcpy(strValue, "REQ_WARM_RESTART"); } break;
        case ANSL_RESP_WARM_RESTART: { strcpy(strValue, "RESP_WARM_RESTART"); } break;
        case ANSL_CANCEL_WARM_RESTART: { strcpy(strValue, "CANCEL_WARM_RESTART"); } break;
        case ANSL_REQ_COLD_RESTART: { strcpy(strValue, "REQ_COLD_RESTART"); } break;
        case ANSL_RESP_COLD_RESTART: { strcpy(strValue, "RESP_COLD_RESTART"); } break;
        case ANSL_CANCEL_COLD_RESTART: { strcpy(strValue, "CANCEL_COLD_RESTART"); } break;
        case ANSL_REQ_GET_INFO: { strcpy(strValue, "REQ_GET_INFO"); } break;
        case ANSL_RESP_GET_INFO: { strcpy(strValue, "RESP_GET_INFO"); } break;
        case ANSL_CANCEL_GET_INFO: { strcpy(strValue, "CANCEL_GET_INFO"); } break;
        case ANSL_REQ_GET_STORAGE: { strcpy(strValue, "REQ_GET_STORAGE"); } break;
        case ANSL_RESP_GET_STORAGE: { strcpy(strValue, "RESP_GET_STORAGE"); } break;
        case ANSL_CANCEL_GET_STORAGE: { strcpy(strValue, "CANCEL_GET_STORAGE"); } break;
        case ANSL_REQ_GET_TASKCLASSES: { strcpy(strValue, "REQ_GET_TASKCLASSES"); } break;
        case ANSL_RESP_GET_TASKCLASSES: { strcpy(strValue, "RESP_GET_TASKCLASSES"); } break;
        case ANSL_CANCEL_GET_TASKCLASSES: { strcpy(strValue, "CANCEL_GET_TASKCLASSES"); } break;
        case ANSL_REQ_GET_HARDWARE: { strcpy(strValue, "REQ_GET_HARDWARE"); } break;
        case ANSL_RESP_GET_HARDWARE: { strcpy(strValue, "RESP_GET_HARDWARE"); } break;
        case ANSL_CANCEL_GET_HARDWARE: { strcpy(strValue, "CANCEL_GET_HARDWARE"); } break;
        case ANSL_REQ_UPLOAD: { strcpy(strValue, "REQ_UPLOAD"); } break;
        case ANSL_RESP_UPLOAD: { strcpy(strValue, "RESP_UPLOAD"); } break;
        case ANSL_CANCEL_UPLOAD: { strcpy(strValue, "CANCEL_UPLOAD"); } break;
        case ANSL_REQ_DOWNLOAD: { strcpy(strValue, "REQ_DOWNLOAD"); } break;
        case ANSL_RESP_DOWNLOAD: { strcpy(strValue, "RESP_DOWNLOAD"); } break;
        case ANSL_CANCEL_DOWNLOAD: { strcpy(strValue, "CANCEL_DOWNLOAD"); } break;
        case ANSL_REQU_CREATE_MODULE: { strcpy(strValue, "REQU_CREATE_MODULE"); } break;
        case ANSL_RESP_CREATE_MODULE: { strcpy(strValue, "RESP_CREATE_MODULE"); } break;
        case ANSL_CANCEL_CREATE_MODULE: { strcpy(strValue, "CANCEL_CREATE_MODULE"); } break;
        case ANSL_REQU_DELETE_MODULE: { strcpy(strValue, "REQU_DELETE_MODULE"); } break;
        case ANSL_RESP_DELETE_MODULE: { strcpy(strValue, "RESP_DELETE_MODULE"); } break;
        case ANSL_CANCEL_DELETE_MODULE: { strcpy(strValue, "CANCEL_DELETE_MODULE"); } break;
        case ANSL_REQU_LIST_MODULES: { strcpy(strValue, "REQU_LIST_MODULES"); } break;
        case ANSL_RESP_LIST_MODULES: { strcpy(strValue, "RESP_LIST_MODULES"); } break;
        case ANSL_CANCEL_LIST_MODULES: { strcpy(strValue, "CANCEL_LIST_MODULES"); } break;
        case ANSL_EVENT_LIST_MODULES: { strcpy(strValue, "EVENT_LIST_MODULES"); } break;
        case ANSL_EVENT_FORMAT: { strcpy(strValue, "EVENT_FORMAT"); } break;
        case ANSL_REQ_GET_FORMAT: { strcpy(strValue, "REQ_GET_FORMAT"); } break;
        case ANSL_RESP_GET_FORMAT: { strcpy(strValue, "RESP_GET_FORMAT"); } break;
        case ANSL_CANCEL_GET_FORMAT: { strcpy(strValue, "CANCEL_GET_FORMAT"); } break;
        case ANSL_EVENT_VALUE: { strcpy(strValue, "EVENT_VALUE"); } break;
        case ANSL_CANCEL_EVENT_VALUE: { strcpy(strValue, "CANCEL_EVENT_VALUE"); } break;
        case ANSL_REQ_GET_VALUE: { strcpy(strValue, "REQ_GET_VALUE"); } break;
        case ANSL_RESP_GET_VALUE: { strcpy(strValue, "RESP_GET_VALUE"); } break;
        case ANSL_CANCEL_GET_VALUE: { strcpy(strValue, "CANCEL_GET_VALUE"); } break;
        case ANSL_REQ_SET_ATTRIBUTES: { strcpy(strValue, "REQ_SET_ATTRIBUTES"); } break;
        case ANSL_RESP_SET_ATTRIBUTES: { strcpy(strValue, "RESP_SET_ATTRIBUTES"); } break;
        case ANSL_CANCEL_SET_ATTRIBUTES: { strcpy(strValue, "CANCEL_SET_ATTRIBUTES"); } break;
        case ANSL_REQ_SET_VALUE: { strcpy(strValue, "REQ_SET_VALUE"); } break;
        case ANSL_RESP_SET_VALUE: { strcpy(strValue, "RESP_SET_VALUE"); } break;
        case ANSL_CANCEL_SET_VALUE: { strcpy(strValue, "CANCEL_SET_VALUE"); } break;
        case ANSL_REQ_GET_CONTENT: { strcpy(strValue, "REQ_GET_CONTENT"); } break;
        case ANSL_RESP_GET_CONTENT: { strcpy(strValue, "RESP_GET_CONTENT"); } break;
        case ANSL_CANCEL_GET_CONTENT: { strcpy(strValue, "CANCEL_GET_CONTENT"); } break;
        case ANSL_EVENT_CONTENT: { strcpy(strValue, "EVENT_CONTENT"); } break;
        case ANSL_CANCEL_EVENT_CONTENT: { strcpy(strValue, "CANCEL_EVENT_CONTENT"); } break;
        case ANSL_REQ_READ_ENTRIES: { strcpy(strValue, "REQ_READ_ENTRIES"); } break;
        case ANSL_RESP_READ_ENTRIES: { strcpy(strValue, "RESP_READ_ENTRIES"); } break;
        case ANSL_CANCEL_READ_ENTRIES: { strcpy(strValue, "CANCEL_READ_ENTRIES"); } break;
        case ANSL_REQ_NEW_LENTRY: { strcpy(strValue, "REQ_NEW_LENTRY"); } break;
        case ANSL_RESP_NEW_LENTRY: { strcpy(strValue, "RESP_NEW_LENTRY"); } break;
        case ANSL_CANCEL_NEW_LENTRY: { strcpy(strValue, "CANCEL_NEW_LENTRY"); } break;
        case ANSL_REQ_GLKN_LINKLIST: { strcpy(strValue, "REQ_GLKN_LINKLIST"); } break;
        case ANSL_RESP_GLKN_LINKLIST: { strcpy(strValue, "RESP_GLKN_LINKLIST"); } break;
        case ANSL_CANCEL_GLKN_LINKLIST: { strcpy(strValue, "CANCEL_GLKN_LINKLIST"); } break;
        case ANSL_REQ_GLKN_FORCEOFF: { strcpy(strValue, "REQ_GLKN_FORCEOFF"); } break;
        case ANSL_RESP_GLKN_FORCEOFF: { strcpy(strValue, "RESP_GLKN_FORCEOFF"); } break;
        case ANSL_CANCEL_GLKN_FORCEOFF: { strcpy(strValue, "CANCEL_GLKN_FORCEOFF"); } break;
        case ANSL_REQ_LKN_FORMAT: { strcpy(strValue, "REQ_LKN_FORMAT"); } break;
        case ANSL_RESP_LKN_FORMAT: { strcpy(strValue, "RESP_LKN_FORMAT"); } break;
        case ANSL_CANCEL_LKN_FORMAT: { strcpy(strValue, "CANCEL_LKN_FORMAT"); } break;
        case ANSL_REQ_LKN_DATA: { strcpy(strValue, "REQ_LKN_DATA"); } break;
        case ANSL_RESP_LKN_DATA: { strcpy(strValue, "RESP_LKN_DATA"); } break;
        case ANSL_CANCEL_LKN_DATA: { strcpy(strValue, "CANCEL_LKN_DATA"); } break;
        case ANSL_REQ_LKN_FORCEDATA: { strcpy(strValue, "REQ_LKN_FORCEDATA"); } break;
        case ANSL_RESP_LKN_FORCEDATA: { strcpy(strValue, "RESP_LKN_FORCEDATA"); } break;
        case ANSL_CANCEL_LKN_FORCEDATA: { strcpy(strValue, "CANCEL_LKN_FORCEDATA"); } break;
        case ANSL_REQ_LKN_FORCEON: { strcpy(strValue, "REQ_LKN_FORCEON"); } break;
        case ANSL_RESP_LKN_FORCEON: { strcpy(strValue, "RESP_LKN_FORCEON"); } break;
        case ANSL_CANCEL_LKN_FORCEON: { strcpy(strValue, "CANCEL_LKN_FORCEON"); } break;
        case ANSL_REQ_LKN_FORCEOFF: { strcpy(strValue, "REQ_LKN_FORCEOFF"); } break;
        case ANSL_RESP_LKN_FORCEOFF: { strcpy(strValue, "RESP_LKN_FORCEOFF"); } break;
        case ANSL_CANCEL_LKN_FORCEOFF: { strcpy(strValue, "CANCEL_LKN_FORCEOFF"); } break;
        case ANSL_REQ_LOGGER_INFO: { strcpy(strValue, "REQ_LOGGER_INFO"); } break;
        case ANSL_RESP_LOGGER_INFO: { strcpy(strValue, "RESP_LOGGER_INFO"); } break;
        case ANSL_CANCEL_LOGGER_INFO: { strcpy(strValue, "CANCEL_LOGGER_INFO"); } break;
        case ANSL_REQ_MODULE_INFO: { strcpy(strValue, "REQ_MODULE_INFO"); } break;
        case ANSL_RESP_MODULE_INFO: { strcpy(strValue, "RESP_MODULE_INFO"); } break;
        case ANSL_CANCEL_MODULE_INFO: { strcpy(strValue, "CANCEL_MODULE_INFO"); } break;
        case ANSL_REQ_TASK_INFO: { strcpy(strValue, "REQ_TASK_INFO"); } break;
        case ANSL_RESP_TASK_INFO: { strcpy(strValue, "RESP_TASK_INFO"); } break;
        case ANSL_CANCEL_TASK_INFO: { strcpy(strValue, "CANCEL_TASK_INFO"); } break;
        case ANSL_REQU_TASK_SECTION: { strcpy(strValue, "REQU_TASK_SECTION"); } break;
        case ANSL_RESP_TASK_SECTION: { strcpy(strValue, "RESP_TASK_SECTION"); } break;
        case ANSL_CANCEL_TASK_SECTION: { strcpy(strValue, "CANCEL_TASK_SECTION"); } break;
        case ANSL_REQU_LIST_VARIABLES: { strcpy(strValue, "REQU_LIST_VARIABLES"); } break;
        case ANSL_RESP_LIST_VARIABLES: { strcpy(strValue, "RESP_LIST_VARIABLES"); } break;
        case ANSL_CANCEL_LIST_VARIABLES: { strcpy(strValue, "CANCEL_LIST_VARIABLES"); } break;
        case ANSL_SERVER_CLIENT_CONNECT: { strcpy(strValue, "SERVER_CLIENT_CONNECT"); } break;
        case ANSL_SERVER_CLIENT_DISCONNECT: { strcpy(strValue, "SERVER_CLIENT_DISCONNECT"); } break;
        case ANSL_DATA_CHANGED: { strcpy(strValue, "DATA_CHANGED"); } break;
        case ANSL_CANCEL_DATA_CHANGED: { strcpy(strValue, "CANCEL_DATA_CHANGED"); } break;
        case ANSL_EVENT_DATA: { strcpy(strValue, "EVENT_DATA"); } break;
        case ANSL_CANCEL_EVENT_DATA: { strcpy(strValue, "CANCEL_EVENT_DATA"); } break;
        case ANSL_REQ_READ_DATA: { strcpy(strValue, "REQ_READ_DATA"); } break;
        case ANSL_RESP_READ_DATA: { strcpy(strValue, "RESP_READ_DATA"); } break;
        case ANSL_CANCEL_READ_DATA: { strcpy(strValue, "CANCEL_READ_DATA"); } break;
        case ANSL_REQ_WRITE_DATA: { strcpy(strValue, "REQ_WRITE_DATA"); } break;
        case ANSL_RESP_WRITE_DATA: { strcpy(strValue, "RESP_WRITE_DATA"); } break;
        case ANSL_CANCEL_WRITE_DATA: { strcpy(strValue, "CANCEL_WRITE_DATA"); } break;
        case ANSL_REQU_LIST_DATA_OBJECTS: { strcpy(strValue, "REQU_LIST_DATA_OBJECTS"); } break;
        case ANSL_RESP_LIST_DATA_OBJECTS: { strcpy(strValue, "RESP_LIST_DATA_OBJECTS"); } break;
        case ANSL_CANCEL_LIST_DATA_OBJECTS: { strcpy(strValue, "CANCEL_LIST_DATA_OBJECTS"); } break;
        case ANSL_CLIENT_DATA_OBJ_CONNECT: { strcpy(strValue, "CLIENT_DATA_OBJ_CONNECT"); } break;
        case ANSL_CLIENT_DATA_OBJ_DISCONNECT: { strcpy(strValue, "CLIENT_DATA_OBJ_DISCONNECT"); } break;
        case ANSL_REQ_RESET: { strcpy(strValue, "REQ_RESET"); } break;
        case ANSL_RESP_RESET: { strcpy(strValue, "RESP_RESET"); } break;
        case ANSL_CANCEL_RESET: { strcpy(strValue, "CANCEL_RESET"); } break;
        case ANSL_REQ_CLEAR_MEMORY: { strcpy(strValue, "REQ_CLEAR_MEMORY"); } break;
        case ANSL_RESP_CLEAR_MEMORY: { strcpy(strValue, "RESP_CLEAR_MEMORY"); } break;
        case ANSL_CANCEL_CLEAR_MEMORY: { strcpy(strValue, "CANCEL_CLEAR_MEMORY"); } break;
        case ANSL_REQ_RESUME: { strcpy(strValue, "REQ_RESUME"); } break;
        case ANSL_RESP_RESUME: { strcpy(strValue, "RESP_RESUME"); } break;
        case ANSL_CANCEL_RESUME: { strcpy(strValue, "CANCEL_RESUME"); } break;
        case ANSL_REQ_SUSPEND: { strcpy(strValue, "REQ_SUSPEND"); } break;
        case ANSL_RESP_SUSPEND: { strcpy(strValue, "RESP_SUSPEND"); } break;
        case ANSL_CANCEL_SUSPEND: { strcpy(strValue, "CANCEL_SUSPEND"); } break;
        case ANSL_REQ_CLEAR_LOGBUFFER: { strcpy(strValue, "REQ_CLEAR_LOGBUFFER"); } break;
        case ANSL_RESP_CLEAR_LOGBUFFER: { strcpy(strValue, "RESP_CLEAR_LOGBUFFER"); } break;
        case ANSL_CANCEL_CLEAR_LOGBUFFER: { strcpy(strValue, "CANCEL_CLEAR_LOGBUFFER"); } break;
        case ANSL_REQ_BITMASK: { strcpy(strValue, "REQ_BITMASK"); } break;
        case ANSL_RESP_BITMASK: { strcpy(strValue, "RESP_BITMASK"); } break;
        case ANSL_CANCEL_BITMASK: { strcpy(strValue, "CANCEL_BITMASK"); } break;
        case ANSL_PROCEEDING_LIST_VARIABLES: { strcpy(strValue, "PROCEEDING_LIST_VARIABLES"); } break;
        case ANSL_EVENT_MODULE_CHANGED: { strcpy(strValue, "EVENT_MODULE_CHANGED"); } break;
        case ANSL_CANCEL_MODULE_CHANGED: { strcpy(strValue, "CANCEL_MODULE_CHANGED"); } break;
        case ANSL_SRV_REGISTER_OBJECT: { strcpy(strValue, "SRV_REGISTER_OBJECT"); } break;
        case ANSL_SRV_OBJECT_REGISTERED: { strcpy(strValue, "SRV_OBJECT_REGISTERED"); } break;
        case ANSL_SRV_REG_OBJECT_FINAL: { strcpy(strValue, "SRV_REG_OBJECT_FINAL"); } break;
        case ANSL_SRV_DEREGISTER_OBJECT: { strcpy(strValue, "SRV_DEREGISTER_OBJECT"); } break;
        case ANSL_SRV_OBJECT_DEREGISTER: { strcpy(strValue, "SRV_OBJECT_DEREGISTER"); } break;
        case ANSL_SRV_TELEGRAM_SEND: { strcpy(strValue, "SRV_TELEGRAM_SEND"); } break;
        case ANSL_SRV_TELEGRAM_SENT: { strcpy(strValue, "SRV_TELEGRAM_SENT"); } break;
        case ANSL_SRV_STREAM_SEND: { strcpy(strValue, "SRV_STREAM_SEND"); } break;
        //case ANSL_SRV_STREAM_SENT: { strcpy(strValue, "SRV_STREAM_SENT"); } break;
        case ANSL_SRV_TELEGRAM_RECEIVE: { strcpy(strValue, "SRV_TELEGRAM_RECEIVE"); } break;
        case ANSL_SRV_TELEGRAM_RECEIVED: { strcpy(strValue, "SRV_TELEGRAM_RECEIVED"); } break;
        case ANSL_SRV_STREAM_RECEIVE: { strcpy(strValue, "SRV_STREAM_RECEIVE"); } break;
        case ANSL_SRV_STREAM_RECEIVED: { strcpy(strValue, "SRV_STREAM_RECEIVED"); } break;
        case ANSL_REQ_CONTROL: { strcpy(strValue, "REQ_CONTROL"); } break;
        case ANSL_RESP_CONTROL: { strcpy(strValue, "RESP_CONTROL"); } break;
        case ANSL_CANCEL_CONTROL: { strcpy(strValue, "CANCEL_CONTROL"); } break;
        case ANSL_CANCEL_EVENT_FORMAT: { strcpy(strValue, "CANCEL_EVENT_FORMAT"); } break;
        case ANSL_PROCEEDING_EVENT: { strcpy(strValue, "PROCEEDING_EVENT"); } break;
        case ANSL_REQ_ASSIGN_VALUE: { strcpy(strValue, "REQ_ASSIGN_VALUE"); } break;
        case ANSL_EVENT_TASKCLASSES_CHANGED: { strcpy(strValue, "EVENT_TASKCLASSES_CHANGED"); } break;
        case ANSL_CANCEL_TASKCLASSES_CHANGED: { strcpy(strValue, "CANCEL_TASKCLASSES_CHANGED"); } break;
        case ANSL_CB_TYPES_END: { strcpy(strValue, "CB_TYPES_END"); } break;
        case ANSL_PROCEEDING_DOWNLOAD: { strcpy(strValue, "PROCEEDING_DOWNLOAD"); } break;
        case ANSL_PROCEEDING_UPLOAD: { strcpy(strValue, "PROCEEDING_UPLOAD"); } break;
        case ANSL_INITIALIZE: { strcpy(strValue, "INITIALIZE"); } break;
        case ANSL_DEINITIALIZE: { strcpy(strValue, "DEINITIALIZE"); } break;
        case ANSL_CONNECTION_STATES: { strcpy(strValue, "CONNECTION_STATES"); } break;
        case ANSL_CONNECTION_ROUTER_ERROR: { strcpy(strValue, "CONNECTION_ROUTER_ERROR"); } break;
        case ANSL_CONNECTION_LISTEN_ERROR: { strcpy(strValue, "CONNECTION_LISTEN_ERROR"); } break;
        case ANSL_CONNECTION_LISTEN_CLOSE: { strcpy(strValue, "CONNECTION_LISTEN_CLOSE"); } break;
        case ANSL_CONNECTION_READY: { strcpy(strValue, "CONNECTION_READY"); } break;
        case ANSL_CONNECTION_ERROR: { strcpy(strValue, "CONNECTION_ERROR"); } break;
        case ANSL_CONNECTION_CLOSE: { strcpy(strValue, "CONNECTION_CLOSE"); } break;
        case ANSL_SERVER_START: { strcpy(strValue, "SERVER_START"); } break;
        case ANSL_SERVER_STARTED: { strcpy(strValue, "SERVER_STARTED"); } break;
        case ANSL_SERVER_STOP: { strcpy(strValue, "SERVER_STOP"); } break;
        case ANSL_SERVER_STOPPED: { strcpy(strValue, "SERVER_STOPPED"); } break;
        case ANSL_SERVER_ERROR: { strcpy(strValue, "SERVER_ERROR"); } break;
        case ANSL_ERROR_LISTEN: { strcpy(strValue, "ERROR_LISTEN"); } break;
        case ANSL_LISTEN_CLOSE: { strcpy(strValue, "LISTEN_CLOSE"); } break;
        case ANSL_ERROR_CONN: { strcpy(strValue, "ERROR_CONN"); } break;
        case ANSL_CONN_CLOSE: { strcpy(strValue, "CONN_CLOSE"); } break;
        case ANSL_CONN_READY: { strcpy(strValue, "CONN_READY"); } break;
        case ANSL_SERVER_STATUS: { strcpy(strValue, "SERVER_STATUS"); } break;
        default:
        {
            sprintf_s(strValue, INFO_STR_SIZE-1, "%u", eventCode);
        } break;
    }
    return strlen(strValue);
}

int GetObjTypeString(unsigned int objType, char strValue[INFO_STR_SIZE])
{
    switch (objType)
    {
        case ANSL_OBJTYPE_ROOT: { strcpy(strValue, "ROOT"); } break;
        case ANSL_OBJTYPE_SERVER: { strcpy(strValue, "SERVER"); } break;
        case ANSL_OBJTYPE_CONNECTION: { strcpy(strValue, "CONNECTION"); } break;
        case ANSL_OBJTYPE_CPU: { strcpy(strValue, "CPU"); } break;
        case ANSL_OBJTYPE_CPUPVS: { strcpy(strValue, "CPUPVS"); } break;
        case ANSL_OBJTYPE_CPUPV: { strcpy(strValue, "CPUPV"); } break;
        case ANSL_OBJTYPE_MODULE: { strcpy(strValue, "MODULE"); } break;
        case ANSL_OBJTYPE_LOGGER: { strcpy(strValue, "LOGGER"); } break;
        case ANSL_OBJTYPE_LIBRARY: { strcpy(strValue, "LIBRARY"); } break;
        case ANSL_OBJTYPE_TASK: { strcpy(strValue, "TASK"); } break;
        case ANSL_OBJTYPE_TASKPV: { strcpy(strValue, "TASKPV"); } break;
        case ANSL_OBJTYPE_GLOBAL_LINK_NODES: { strcpy(strValue, "GLOBAL_LINK_NODES"); } break;
        case ANSL_OBJTYPE_LINK_NODE: { strcpy(strValue, "LINK_NODE"); } break;
        case ANSL_OBJTYPE_REGISTERD: { strcpy(strValue, "REGISTERD"); } break;
        case ANSL_OBJTYPE_REGED_STREAM: { strcpy(strValue, "REGED_STREAM"); } break;
        case ANSL_OBJTYPE_REGED_TELEGRAM: { strcpy(strValue, "REGED_TELEGRAM"); } break;
        case ANSL_OBJTYPE_STREAM: { strcpy(strValue, "STREAM"); } break;
        case ANSL_OBJTYPE_TELEGRAM: { strcpy(strValue, "TELEGRAM"); } break;
        case ANSL_OBJTYPE_VARIABLE: { strcpy(strValue, "VARIABLE"); } break;
        default:
        {
            sprintf_s(strValue, INFO_STR_SIZE-1, "%u", objType);
        } break;
    }
    return strlen(strValue);
}

int GetFormatString(ANSL_VARIABLE_DATA_TYPE* pTypeInfo, char strValue[FMT_STR_SIZE])
{
    int dims = 0;
    char strDim[40];

    memset(strValue, 0, FMT_STR_SIZE);
    switch (pTypeInfo->m_DataType)
    {
        case ANSL_VAR_TYPE_UINT8:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=ui8 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=ui8 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_UINT16:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=ui16 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=ui16 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_UINT32:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=ui32 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=ui32 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_UINT64:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=ui64 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=ui64 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_INT8:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=i8 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=i8 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_INT16:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=i16 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=i16 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_INT32:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=i32 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=i32 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_INT64:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=i64 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=i64 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_DT:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=DateNTime Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=DateNTime Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_BOOL:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=bool Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=bool Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_FLOAT32:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=f32 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=f32 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_FLOAT64:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=f64 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=f64 Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_CHAR_STRING:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=string Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=string Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_WCHAR_STRING:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=wstring Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=wstring Len=%d", pTypeInfo->m_ElementLength);
            }
        } break;
        case ANSL_VAR_TYPE_STRUCT:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=struct Len=%d ArDim=%d Size=%d Name=%s ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size, pTypeInfo->m_Name);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=struct Len=%d Name=%s", pTypeInfo->m_ElementLength, pTypeInfo->m_Name);
            }
        } break;
        default:
        {
            if (0 < pTypeInfo->m_ArrayDimensions)
            {
                sprintf(strValue, "Type=UNKNOWN Len=%d ArDim=%d Size=%d Name=%s ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size, pTypeInfo->m_Name);
                for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
                {
                    memset(strDim, 0, 40);
                    sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
                    sprintf(strValue, "%s%s", strValue, strDim);
                }
            } else {
                sprintf(strValue, "Type=UNKNOWN Len=%d Name=%s", pTypeInfo->m_ElementLength, pTypeInfo->m_Name);
            }
        } break;
    }
    return strlen(strValue);
}

char* DataToString(char *pData, unsigned int dataLen, ANSL_VARIABLE_DATA_TYPE *pDataType)
{
    char strTmp[256];

    int i32V;
    unsigned int ui32V;
    short i16V;
    unsigned short ui16V;
    char i8V;
    unsigned char ui8V;
    double f64V = 0.0;
    float f32V;
    unsigned int xCnt, yCnt, xOffset;


    if ((NULL != pData) && (0 < dataLen))
    {
        switch (pDataType->m_DataType)
        {
            case ANSL_VAR_TYPE_BOOL:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    memcpy(&i8V, pData, 1);
                    sprintf(_strValue, "%s", ((0 != i8V) ? "true" : "false"));
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memcpy(&i8V, pData+xOffset, 1);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%s;",  ((0 != i8V) ? "true" : "false"));
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                            xOffset++;
                        }
                    }
                }
            } break;
            case ANSL_VAR_TYPE_INT8:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    memcpy(&i8V, pData, 1);
                    sprintf(_strValue, "%d",  i8V);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memcpy(&i8V, pData+xOffset, 1);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%d;",  i8V);
                            xOffset+=1;
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                        }
                    }
                }
            } break;
            case ANSL_VAR_TYPE_INT16:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    memcpy(&i16V, pData, 2);
                    sprintf(_strValue, "%d",  i16V);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memcpy(&i16V, pData+xOffset, 2);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%d;",  i16V);
                            xOffset+=2;
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                        }
                    }
                }
            } break;
            case ANSL_VAR_TYPE_INT32:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    memcpy(&i32V, pData, 4);
                    sprintf(_strValue, "%d",  i32V);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memcpy(&i32V, pData+xOffset, 4);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%d;",  i32V);
                            xOffset+=4;
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                        }
                    }
                }
            } break;
            case ANSL_VAR_TYPE_UINT8:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    memcpy(&ui8V, pData, 1);
                    sprintf(_strValue, "%u",  ui8V);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memcpy(&ui8V, pData+xOffset, 1);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%u;",  ui8V);
                            xOffset+=1;
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                        }
                    }
                }
            } break;
            case ANSL_VAR_TYPE_UINT16:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    memcpy(&ui16V, pData, 2);
                    sprintf(_strValue, "%u",  ui16V);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memcpy(&ui16V, pData+xOffset, 2);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%u;",  ui16V);
                            xOffset+=2;
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                        }
                    }
                }
            } break;
            case ANSL_VAR_TYPE_UINT32:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    memcpy(&ui32V, pData, 4);
                    sprintf(_strValue, "%u",  ui32V);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memcpy(&ui32V, pData+xOffset, 4);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%u;",  ui32V);
                            xOffset+=4;
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                        }
                    }
                }
            } break;
            case ANSL_VAR_TYPE_FLOAT32:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    memcpy(&f32V, pData, 4);
                    sprintf(_strValue, "%f",  f32V);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memcpy(&f32V, pData+xOffset, 4);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%f;",  f32V);
                            xOffset+=4;
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                        }
                    }
                }
            } break;
            case ANSL_VAR_TYPE_FLOAT64:
            {
                //if (0 == pDataType->m_ArrayDimensions)
                //{
                    memcpy(&f64V, pData, 8);
                    sprintf(_strValue, "%g",  f64V);
                /*} else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memcpy(&f64V, pData+xOffset, 8);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%g;",  f64V);
                            xOffset+=8;
                            sprintf(_strValue, "%s%s", _strValue, strTmp;
                        }
                    }
                }
                */
            } break;
            case ANSL_VAR_TYPE_CHAR_STRING:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    sprintf(_strValue, "%s", pData);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%s;", (char*)(pData+xOffset));
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                            xOffset+=pDataType->m_ElementLength;
                        }
                    }
                }
            } break;
            case ANSL_VAR_TYPE_STRUCT:
            {
                return "";
            } break;
            case ANSL_VAR_TYPE_DT:
            {
                CPU_DATETIME_Type* pTmpDT;
                if (0 == pDataType->m_ArrayDimensions)
                {
                    pTmpDT = (CPU_DATETIME_Type*)pData;
                    sprintf(_strValue,
                            "%04d/%02d/%02d %02d:%02d:%02d",
                            (pTmpDT->Year + 1900),
                            pTmpDT->Mon,
                            pTmpDT->MDay,
                            pTmpDT->Hour,
                            pTmpDT->Min,
                            pTmpDT->Sec);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            pTmpDT = (CPU_DATETIME_Type*)(pData+xOffset);
                            memset(strTmp, 0, 256);
                            sprintf(strTmp,
                                    "%04d/%02d/%02d %02d:%02d:%02d",
                                    pTmpDT->Year + 1900,
                                    pTmpDT->Mon,
                                    pTmpDT->MDay,
                                    pTmpDT->Hour,
                                    pTmpDT->Min,
                                    pTmpDT->Sec);
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                            xOffset+=pDataType->m_ElementLength;
                        }
                    }
                }
            } break;
            default:
            {
                if (0 == pDataType->m_ArrayDimensions)
                {
                    switch (pDataType->m_ElementLength)
                    {
                        case 1:
                        {
                            memcpy(&i8V, pData, 1);
                            i32V = i8V;
                        } break;
                        case 2:
                        {
                            memcpy(&i16V, pData, 2);
                            i32V = i16V;
                        } break;
                        case 4:
                        {
                            memcpy(&i32V, pData, 4);
                        } break;
                    }
                    sprintf(_strValue, "%d",  i32V);
                } else {
                    xOffset = 0;
                    for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
                    {
                        for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
                        {
                            switch (pDataType->m_ElementLength)
                            {
                                case 1:
                                {
                                    memcpy(&i8V, pData+xOffset, 1);
                                    i32V = i8V;
                                    xOffset+=1;
                                } break;
                                case 2:
                                {
                                    memcpy(&i16V, pData+xOffset, 2);
                                    i32V = i16V;
                                    xOffset+=2;
                                } break;
                                case 4:
                                {
                                    memcpy(&i32V, pData+xOffset, 4);
                                    xOffset+=4;
                                } break;
                            }
                            memset(strTmp, 0, 256);
                            sprintf(strTmp, "%d;", i32V);
                            sprintf(_strValue, "%s%s", _strValue, strTmp);
                        }
                    }
                }
            } break;
        }
    } else {
        sprintf(_strValue, "INVALID DATA");
    }
    return _strValue;
}

void AnslGetVersionInfo()
{
    unsigned int status = 0;
    unsigned int uifVersion = 0;
    unsigned int oifVersion = 0;

    status = ANSLI_GetVersionInfos(&uifVersion, &oifVersion);
    if (status != 0)
    {
        xprintf("ERROR ANSLI_GetVersionInfos : %i\n\r",status);
    } else {
		xprintf("ANSLI_GetVersionInfos: ");
		PrintVInfo();
    }    
}

void TaskClassResume(unsigned int cpuIdent, unsigned short tcNumber)
{
    unsigned int status = 0;

    status = ANSLI_CPU_ResumeTaskClass(cpuIdent, tcNumber);
    if (status != 0)
    {
        xprintf("ERROR RESUME TC : %i\n",status);        
    }
}

void TaskClassSuspend(unsigned int cpuIdent, unsigned short tcNumber)
{
    unsigned int status = 0;

    status = ANSLI_CPU_SuspendTaskClass(cpuIdent, tcNumber);
    if (status != 0)
    {
        xprintf("ERROR SUSPEND TC : %i\n",status);        
    }
}

void TaskResume(unsigned int taskIdent, unsigned short numOfCycles)
{
    unsigned int status = 0;

    status = ANSLI_TASK_Resume(taskIdent, numOfCycles);
    if (status != 0)
    {
        xprintf("ERROR RESUME TASK : %i\n",status);        
    }
}

void TaskSuspend(unsigned int taskIdent, unsigned short numOfCycles)
{
    unsigned int status = 0;

    status = ANSLI_TASK_Suspend(taskIdent, numOfCycles);
    if (status != 0)
    {
        xprintf("ERROR SUSPEND TASK : %i\n",status);        
    }
}

void LoggerClearBuffer(unsigned int loggerIdent)
{
    unsigned int status = 0;

    status = ANSLI_LOGGER_ClearBuffer(loggerIdent);
    if (status != 0)
    {
        xprintf("ERROR CLEAR BUFFER  LOGGER: %i\n",status);        
    }
}

void CpuClearMemory(unsigned int cpuIdent, ANSL_CPU_CLEARMEM_MODES memType)
{
    unsigned int status = 0;

    status = ANSLI_CPU_ClearMemory(cpuIdent, memType);
    if (status != 0)
    {
        xprintf("ERROR CLEAR MEM  CPU : %i\n",status);        
    }
}

/******************************************************************************
* Function: Enter synchronize point.
* Input:  pSyncDesc:	pointer to valid synchronize object descriptor
******************************************************************************/
int sysEnterSync (T_SYNC_DESC*	pSyncDesc)
{
    if ((pSyncDesc->UseCriticalSection) || (1 == pSyncDesc->Created))
    {
        XEnterCriticalSection (&pSyncDesc->Sync.CriticalSection);
    }
    return (0);
}


/******************************************************************************
* Function: Get unique thread (or task) identification number.
******************************************************************************/
pthread_t sysGetThreadId ()
{
    // obtain ID of the calling thread
    // The thread ID returned by pthread_self() is not the same thing as the kernel thread ID returned by a call to gettid(2).
    //return pthread_self();

    // gettid() returns the caller's thread ID (TID).
    // In a single-threaded process, the thread ID is equal to the process ID (PID, as returned by getpid(2)).
    // In a multithreaded process, all threads have the same PID, but each one has a unique TID.
    // gettid is not ported on all platforms therefore use syscall!!
    return syscall(SYS_gettid);
}

/******************************************************************************
* Function: Leave synchronize point.
* Input:  pSyncDesc:	pointer to valid synchronize object descriptor
******************************************************************************/
int sysLeaveSync (T_SYNC_DESC*	pSyncDesc)
{
     pthread_t locThreadId = 0;
     if (1 == pSyncDesc->Created)
     {
         locThreadId = sysGetThreadId();
        if ((locThreadId == pSyncDesc->threadID) && (0 < pSyncDesc->syncCount))
        {
            pSyncDesc->syncCount--;
        } else {
            pSyncDesc->threadID = SYS_INVALID_THREAD_ID;
            pSyncDesc->syncCount = 0;
			if (pSyncDesc->UseCriticalSection) {
				XLeaveCriticalSection(&pSyncDesc->Sync.CriticalSection);
            } else {
				if (!XReleaseMutex(&pSyncDesc->Sync.hMutex)) {
                    return (ANSL_ERR_SYNC_RESOURCE);
                }
            }
        }
     }
    return (ANSL_NO_ERR);
}

/******************************************************************************
* Function: Close synchronize object.
* Input:  pSyncDesc:	pointer to valid synchronize object descriptor
******************************************************************************/
void sysCloseSync (T_SYNC_DESC*	pSyncDesc)
{
    pSyncDesc->syncCount = 0;
    pSyncDesc->threadID = SYS_INVALID_THREAD_ID;
    /* close critical section: */
	if (pSyncDesc->UseCriticalSection) {
		XDeleteCriticalSection(&pSyncDesc->Sync.CriticalSection);
	} else {
        /* close mutex object: */
		XCloseHandle(&pSyncDesc->Sync.hMutex);
        pSyncDesc->Sync.hMutex.type = H_UNINITILIZED;
    }
    pSyncDesc->UseCriticalSection = FALSE;
    pSyncDesc->Created = 0;

}

/******************************************************************************
* Function: Check synchronize object.
******************************************************************************/

int sysCheckSync (T_SYNC_DESC*	pSyncDesc)
{
    if (H_UNINITILIZED != pSyncDesc->Sync.hMutex.type)
    {
        return TRUE;
    }
    return FALSE;
}

WINBASEAPI void WINAPI XEnterCriticalSection(CRITICAL_SECTION *lpCriticalSection)
{
    int status = 0;

    //check if there is a valid mutex available
    if (PTHREAD_MUTEX_RECURSIVE != lpCriticalSection->SyncMutex.__data.__kind)
    {
#ifdef DEBUG_CS
        fprintf( stderr, "pthread_mutex_lock() called on non initialized mutex\n");fflush(stderr);
#endif
        return;
    }
#ifdef _TRY_LOCK_MUTEX
    status = pthread_mutex_trylock(&lpCriticalSection->SyncMutex);
    while (EBUSY == status)
    {
        //nanoWait(0, 10);
        status = pthread_mutex_trylock(&lpCriticalSection->SyncMutex);
    }
    if (0 != status)
    {
        nanoWait(0, 10);
    }
#else
#ifdef DEBUG_CS
        xprintf ("# +ECS -   (%d:%p)\n", anslSysGetThreadId(), lpCriticalSection);
#endif
    status = pthread_mutex_lock(&lpCriticalSection->SyncMutex);
    lpCriticalSection->LockCount++;
#endif
    if( status != 0 )
    {
#ifdef DEBUG_CS
        fprintf( stderr, "MT : pthread_mutex_lock() failed (0x%p) - errno #%d\n", lpCriticalSection, status );fflush(stderr);
#endif
    }

#ifdef DEBUG_CS
        xprintf ("# -ECS -   (%d:%p)\n", anslSysGetThreadId(), lpCriticalSection);
#endif
}

WINBASEAPI void WINAPI XLeaveCriticalSection(CRITICAL_SECTION *lpCriticalSection)
{
    int status = 0;

    if(0 == lpCriticalSection->SyncMutex.__data.__owner)
    {
#ifdef DEBUG_CS
        fprintf( stderr, "MT : uninitialized critical section (%p)\n", lpCriticalSection );fflush(stderr);
#endif
        return;
    }

    status = pthread_mutex_unlock(&lpCriticalSection->SyncMutex);
    lpCriticalSection->LockCount--;

    if( status != 0 )
    {
#ifdef DEBUG_CS
        fprintf(stderr, "MT : pthread_mutex_unlock() failed (0x%p) - errno #%d\n", lpCriticalSection, status );fflush(stderr);
#endif
        return;
    }

#ifdef DEBUG_CS
        xprintf ("# LCS   - (%d:%p)\n", anslSysGetThreadId(), lpCriticalSection);
#endif
}

WINBASEAPI void WINAPI XDeleteCriticalSection(LPCRITICAL_SECTION lpCriticalSection)
{
    int retVal = 0;
#ifdef DEBUG_CS
        xprintf ("#DCS    - (%d:%p)\n", anslSysGetThreadId(), lpCriticalSection);
#endif

    if (0 != lpCriticalSection->SyncMutex.__data.__owner)
    {
        retVal = pthread_mutex_trylock(&lpCriticalSection->SyncMutex);
        while (EBUSY == retVal)
        {
            retVal = pthread_mutex_trylock(&lpCriticalSection->SyncMutex);
        }
        while (0 < lpCriticalSection->SyncMutex.__data.__count)
        {
            pthread_mutex_unlock(&lpCriticalSection->SyncMutex);
        }
    }
    retVal = pthread_mutex_destroy(&lpCriticalSection->SyncMutex);
    if(0 != retVal)
    {
#ifdef DEBUG_CS
        fprintf( stderr, "MT : pthread_mutex_destroy() failed (EBUSY)\n" );fflush(stderr);
#endif
        retVal = retVal;
        return;
    }
}

WINBASEAPI BOOL WINAPI
XReleaseMutex(hUnion* pMutex) {
	return XCloseHandle(pMutex);
}

WINBASEAPI BOOL WINAPI
XCloseHandle(hUnion* pHandle) {
    int retVal = 0;

    if (NULL == pHandle)
        return FALSE;

    switch (pHandle->type)
    {
        case H_MUTEX:
        {
            if (0 != pHandle->obj.mutex.__data.__owner)
            {
                retVal = pthread_mutex_trylock(&pHandle->obj.mutex);
                while (EBUSY == retVal)
                {
                    retVal = pthread_mutex_trylock(&pHandle->obj.mutex);
                }
                while (0 < pHandle->obj.mutex.__data.__count)
                {
                    pthread_mutex_unlock(&pHandle->obj.mutex);
                }
            }
            pthread_mutex_destroy(&pHandle->obj.mutex);
        } break;
        case H_THREAD:
        {
            pHandle->type 	= H_THREAD;
        }break;
        case H_EVENT:
        {
            pthread_mutex_lock(&pHandle->obj.event.mutex);
            pthread_cond_broadcast(&pHandle->obj.event.cond);
            pthread_mutex_unlock(&pHandle->obj.event.mutex);
            pthread_cond_destroy(&pHandle->obj.event.cond);
            pthread_mutex_destroy(&pHandle->obj.event.mutex);
        }break;
        case H_CONDITION:
        {
            while(EBUSY == pthread_cond_destroy(&pHandle->obj.event.cond))
            {
                //SetEvent(&pHandle->obj.event.cond);
                sysSleep(1);
            }
            pthread_mutex_destroy(&pHandle->obj.event.mutex);
        }break;
        case H_UNINITILIZED:
        {
            pHandle->type = H_UNINITILIZED;
            // Close kann auch f�r nie eingerichtete Objekte aufgerufen werden!!!!!
        }break;
        default:
        {
            pHandle->type = H_UNINITILIZED;
            raise(SIGINT);
        }break;
    }

    return TRUE;
}


/******************************************************************************
* Function: Slepp the specified number of milli seconds.
******************************************************************************/
void sysSleep(unsigned int milliSeconds)
{
    // The usleep() function suspends execution of the calling process for (at least) usec microseconds.
    // The sleep may be lengthened slightly by any system activity or by the time spent processing the call or by the granularity of system timers.
    // 4.3BSD, POSIX.1-2001.  POSIX.1-2001 declares this function obsolete; use
    // nanosleep(2) instead.  POSIX.1-2008 removes the specification of usleep().
    //
    // The interaction of this function with the SIGALRM signal, and with other timer
    // functions such as alarm(2), sleep(3), nanosleep(2), setitimer(2),
    // timer_create(2), timer_delete(2), timer_getoverrun(2), timer_gettime(2),
    // timer_settime(2), ualarm(3) is unspecified.
    //usleep(milliSeconds);

    struct timespec tsSleep;
    struct timespec tsSleepRem;
    tsSleep.tv_sec = 0;
    tsSleep.tv_nsec = 1;
    if (0 < milliSeconds)
    {
        tsSleep.tv_sec = milliSeconds/1000;
        tsSleep.tv_nsec = ((milliSeconds - (tsSleep.tv_sec*1000))*1000000); /* nanoseconds [0 .. 999999999] */
    }
    // delays the execution of the program for at least the time specified in *tsSleep.
    // The function can return earlier if a signal has been delivered to the process.
    // In this case, it returns -1, sets errno to EINTR, and writes the remaining time into the structure pointed to by rem unless rem is NULL.
    // The value of *tsSleepRem will then be used to call nanosleep() again and complete the specified pause.
    while (0 != nanosleep(&tsSleep, &tsSleepRem))
    {
        tsSleep.tv_sec = tsSleepRem.tv_sec;
        tsSleep.tv_nsec = tsSleepRem.tv_nsec;
    }
}

/******************************************************************************
* Function: Open synchronize object.
* Input:  pIdName:		== NULL -> local synchronize object
*						!= NULL -> null-terminated global object name
* Output: pSyncDesc:	descriptor of the created synchronize object
******************************************************************************/
int sysCreateSync (T_SYNC_DESC* pSyncDesc,
                                char_t* pIdName)
{
    TCHAR	szIdNameBf[256];
    TCHAR*	szIdName;
    int retVal = 0;

    memset(pSyncDesc, 0, sizeof(ANSL_SYNC_DESC));
    pSyncDesc->syncCount = 0;
    pSyncDesc->threadID = SYS_INVALID_THREAD_ID;
    if (pIdName == NULL)
    {
        /* create unnamed synchronize object: */
        if (0 == XInitializeCriticalSection(&pSyncDesc->Sync.CriticalSection))
        {
            pSyncDesc->Created = 1;
            pSyncDesc->UseCriticalSection = TRUE;
        }
    }
    else
    {
#ifdef UNICODE
        MultiByteToWideChar (CP_ACP, 0, pIdName, -1, szIdNameBf, 256-1);
        szIdNameBf[256-1] = _T('\0');
        szIdName = szIdNameBf;
#else
        szIdNameBf[0] = '\0';
        szIdName = (TCHAR*) pIdName;
#endif
        /* create named synchronize object: */
        pSyncDesc->UseCriticalSection = FALSE;
        pSyncDesc->Sync.hMutex.type    = H_MUTEX;
		retVal = XCreateMutex(&g_pSecAttr, FALSE, szIdName,
				&pSyncDesc->Sync.hMutex.obj.mutex);
		if (0 != retVal) {
            return (ANSL_ERR_SYNC_RESOURCE);
        } else {
            pSyncDesc->Created = 1;
        }
    }

    return retVal;
}

WINBASEAPI int WINAPI
XInitializeCriticalSection(LPCRITICAL_SECTION lpCriticalSection) {
    int retVal = 0;
    pthread_mutexattr_t	mutex_attr;

    lpCriticalSection->LockCount      = 0;
    lpCriticalSection->RecursionCount = 0;
    lpCriticalSection->OwningThread   = 0;
    lpCriticalSection->LockSemaphore  = 0;

#ifdef DEBUG_CS
        printf ("#ICS    - (%d:%p)\n", anslSysGetThreadId(), lpCriticalSection);
#endif

    retVal = pthread_mutexattr_init( &mutex_attr );

    // The pthread_mutex_trylock() function shall be equivalent to pthread_mutex_lock(),
    // except that if the mutex object referenced by mutex is currently locked
    // (by any thread, including the current thread), the call shall return immediately.
    // If the mutex type is PTHREAD_MUTEX_RECURSIVE and the mutex is currently owned by the calling thread,
    //the mutex lock count shall be incremented by one and the pthread_mutex_trylock() function shall immediately return success.

    //retVal = pthread_mutexattr_settype(&mutex_attr, PTHREAD_MUTEX_ERRORCHECK);
    retVal = pthread_mutexattr_settype(&mutex_attr, PTHREAD_MUTEX_RECURSIVE);

    retVal = pthread_mutex_init(&lpCriticalSection->SyncMutex, &mutex_attr );
    if (0 != retVal)
    {
        fprintf(stderr, "pthread_mutex_init E=%d\n", retVal);fflush(stderr);
    }
    return retVal;
}

WINBASEAPI int WINAPI
XCreateMutex(LPSECURITY_ATTRIBUTES lpSecAttribs, BOOL bManualReset,
		LPCSTR lpName, pthread_mutex_t* pMutex) {
    int retVal = 0;
    pthread_mutexattr_t mutexAttr;
    (void)lpSecAttribs;
    (void)bManualReset;

    //memset(&pMutex->obj.mutex, 0, sizeof(pthread_cond_t));
    if(NULL != lpName)
    {
        // System V semaphores can be used to provide the named mutex functionality in Linux
        //key_t key;
        //int semid;

        //  Linux does not directly support named mutex!
        //  An exception will be raised instead.
        raise(SIGINT);

        //Generate a unique key or supply a value
        //key = ftok();
        // a unique identifier to identify semaphore set
        //      |        number of semaphore in the semaphore set
        //      |_______    |      _____permissions (rwxrwxrwx) on the new semaphore set and creation flag
        //              |   |     |
        //semid = semget(key, 1, (0666 | IPC_CREAT));
    } else {
        pthread_mutexattr_init( &mutexAttr );

        //retVal = pthread_mutexattr_settype(&mutexAttr, PTHREAD_MUTEX_ERRORCHECK);
        retVal = pthread_mutexattr_settype(&mutexAttr, PTHREAD_MUTEX_RECURSIVE);

        pthread_mutex_init(pMutex, &mutexAttr );

        retVal = pthread_mutexattr_destroy(&mutexAttr);
    }

#ifdef DEBUG_CS
    fprintf( stderr, "CreateMutex            : %p (%d)\n",
                        pMutex, anslSysGetThreadId() );fflush(stderr);
#endif

    return retVal;
}


