/*
 ============================================================================
 Name        : ANSLSample.c
 Author      : Werner Hager
 Version     : 1
 Copyright   : B&R
 Description : ANSL sample
 ============================================================================
 */

/***************************************************************************************************/
/** INCLUDES ***************************************************************************************/

#ifdef WIN32
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include "..\..\..\..\Exp\ANSLUIF.h"	/* ANSL USER INTERFACE */
#include <Windows.h>
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ANSLUIF.h>	/* ANSL USER INTERFACE */
#include <termios.h>
#include <unistd.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/timeb.h>
#endif

#ifndef sprintf_s
#define sprintf_s(frmt, size, args...) 			sprintf(frmt, ##args)
#endif

/***************************************************************************************************/
/** DEFINES ****************************************************************************************/

#define TARGET_IP 					"127.0.0.1"	/* IP of the target */
#define CONNECTION_PARAMS			"Timeout=\"35000\" Port=\"11169\" SendDelay=\"1\" Buffer=\"4096\""	/* connection parameters */
#define CONNECTION_PARAMS_NO_TO		"Port=\"11169\" SendDelay=\"1\" Buffer=\"4096\""	/* connection parameters */
#define CONNECTION_PARAMS_NO_TOPO	"SendDelay=\"1\" Buffer=\"4096\""	/* connection parameters */
#define TASK_NAME 					"pvitest"		/* task name of the task on target */
#define VAR_NAME_READ 				"PV1"			/* local UDINT variable */
#define VAR_NAME_SET 				"PV2"			/* local UDINT variable */
#define VAR_NAME_SET_VALUE	        20				/* VAR_NAME_SET gets this value */

#define RC_LOCK_ERROR              1
#define RC_OBJECT_NOT_CREATED      2
#define RC_STACKSIZE_ERROR         3
#define RC_THREAD_ATTR_ERROR       4
#define RC_THREAD_NOT_CREATED      5
#define RC_UNLOCK_ERROR            6
#define RC_NOT_OWNER               7

#define FMT_STR_SIZE			   1024
#define INFO_STR_SIZE			   256

#pragma pack(1)
/* read/write cpu date and time (request/response): */
typedef struct
{
    unsigned char		Sec;			/* seconds after the minute - [0,59] */
	unsigned char		Min;			/* minutes after the hour - [0,59] */
	unsigned char		Hour;			/* hours since midnight - [0,23] */
	unsigned char		MDay;			/* day of the month - [1,31] */
	unsigned char		Mon;			/* months since January - [0,11] */
    unsigned char		Year;			/* years since 1900 */
    unsigned char		WDay;			/* days since Sunday - [0,6] */
    unsigned char		IsDst;			/* daylight savings time flag */
    unsigned short		YDay;			/* days since January 1 - [0,365] */
}	CPU_DATETIME_Type;
#pragma pack()


/***************************************************************************************************/
/** GLOBAL VARIABLES  ******************************************************************************/

/* for synchronization */
int var1,var2,var3,var4 = 0;
ANSL_VARIABLE_EVENT_PARAMETER evParam;
unsigned int identCpu380 = 0;				/* ident of connection to cpu */
unsigned int identTask = 0;					/* ident of connection to task1 */
unsigned int identLogger = 0;			    /* ident of connection to $arlogsys */
unsigned int identVarRead = 0;				/* ident of connection to var1 */
unsigned int identVarSet = 0;				/* ident of connection to var2 */
unsigned short varSet = VAR_NAME_SET_VALUE;	/* value to set on target */
unsigned int status = 0;					/* state of function calls */
unsigned int userDataCpu = 380;			/* user date to identify the caller */
unsigned int userDataTask = 381;			/* user date to identify the caller */
unsigned int userDataLogger = 400;			/* user date to identify the caller */
unsigned int userDataRead = 1;				/* user date to identify the caller */
unsigned int userDataSet = 2;				/* user date to identify the caller */

#ifdef WIN32
#define T_PTHREAD void*
#define T_PTHREAD_MUTEX int
#else
#define T_PTHREAD pthread_t
#define T_PTHREAD_MUTEX pthread_mutex_t
pthread_attr_t _threadAttr;
#endif

#define T_PTHREADP T_PTHREAD*
#define T_PTHREAD_MUTEXP T_PTHREAD_MUTEX*

#ifdef WIN32
int pthread_mutex_lock(T_PTHREAD_MUTEXP pMutex) { return 0; };
int pthread_mutex_unlock(T_PTHREAD_MUTEXP pMutex) { return 0; };

__declspec(dllimport) void* __stdcall CreateThread(LPSECURITY_ATTRIBUTES lpThreadAttributes, 
												  unsigned long dwStackSize, 
												  LPTHREAD_START_ROUTINE lpStartAddress, 
												  void* lpParameter, 
												  unsigned long dwCreationFlags, 
												  unsigned long* lpThreadId);
__declspec(dllimport) unsigned long __stdcall WaitForSingleObject(void* hHandle, unsigned long dwMilliseconds);
__declspec(dllimport) int __stdcall CloseHandle(void* hHandle);
#endif

T_PTHREAD _ThreadId;
T_PTHREAD_MUTEX _MutexObject;
int _bBreakThread;
int _CPUConnClosed;
int _bConnectionClosed;
int _RunningState;
char _strObj[10];
char _strPV1[100];
char _strPV2[100];
char _strTask[100];
char _strValue[1024];
ANSL_VARIABLE_DATA_TYPE _DataTypePV1;
ANSL_VARIABLE_DATA_TYPE _DataTypePV2;
char strIP[16]; 		/* IP-Adresse */
char strConPara[256]; 	/* Connection parameters */
unsigned int pvParentIdent;
int rfPV1;
int rfPV2;

unsigned long GetSystemTickCount(); // Zeit seit dem Booten in Sekunden
unsigned long GetSystemTickCountA(); // Zeit seit dem Booten in Sekunden
unsigned long GetSystemTickCountB(); // Zeit seit dem Booten in Sekunden
char* DataToString(char *pData, unsigned int dataLen, ANSL_VARIABLE_DATA_TYPE *pDataType);
int GetFormatString(ANSL_VARIABLE_DATA_TYPE* pTypeInfo, char strValue[FMT_STR_SIZE]);
int GetObjTypeString(unsigned int objType, char strValue[INFO_STR_SIZE]);
int GetCBTypeString(unsigned int cbType, char strValue[INFO_STR_SIZE]);
unsigned int ReadCpuDateTime(unsigned int cpuIdent);
unsigned int ReadVariable(unsigned int varIdent);
unsigned int WriteVariable(unsigned int varIdent, long varValue);
int connectCPU(char* strIP, char* strConPara, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData);
int connectTask(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData);
int connectLogger(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData);
int connectVariable(unsigned int parentID, char* name, unsigned int* ident, int flags, int refresh, ANSL_CALL_BACK fpCB, void* userData);
void AnslGetVersionInfo();
void TaskClassResume(unsigned int cpuIdent, unsigned short tcNumber);
void TaskClassSuspend(unsigned int cpuIdent, unsigned short tcNumber);
void TaskResume(unsigned int taskIdent, unsigned short numOfCycles);
void TaskSuspend(unsigned int taskIdent, unsigned short numOfCycles);
void LoggerClearBuffer(unsigned int loggerIdent);
void CpuClearMemory(unsigned int cpuIdent, ANSL_CPU_CLEARMEM_MODES memType);


int ioGetch();
unsigned int ConnectPV(char *name, unsigned int parentIdent, unsigned int evFlags, unsigned int refresh, void *pUserData, unsigned int *pIdent);
void DoWork();
void StartNoThread();

#ifdef _WORKERTHREAD
int CreateWorkerThread(int* pRunningState, T_PTHREAD_MUTEX* pMutex, T_PTHREAD* pThreadId, void* fpThread);
#ifdef WIN32
static unsigned long __stdcall WorkerThread (void* pParam);
void pthread_join(T_PTHREAD threadId, void* pAttribs) { return; };
#else
static void* WorkerThread(void* pParam);
#endif
#endif

#ifdef WIN32
static unsigned int __stdcall ANSLCB_Cpu(unsigned int errorCode, 
										 unsigned int ident, 
										 unsigned int objectType, 
										 unsigned int cbType, 
										 void* pData, 
										 unsigned int dataLen, 
										 void* pUserData);
#else
unsigned int ANSLCB_Cpu(unsigned int errorCode, 
						  unsigned int ident, 
						  unsigned int objectType, 
						  unsigned int cbType, 
						  void* pData, 
						  unsigned int dataLen, 
						  void* pUserData);
#endif
int ErrorCheck(unsigned int status, const char* error);
void DoExit();
int UninitializeObjects();

void PrintVInfo() 
{
	unsigned int uifVersion = 0;
	unsigned int oifVersion = 0;
	int retVal = 0;
    int verMaj = 0;
    int verMin = 0;
    int verBuild = 0;

	retVal = ANSLI_GetVersionInfos(&uifVersion, &oifVersion);
    verMaj = uifVersion / 10000;
    verMin = (uifVersion - verMaj * 10000) / 100;
    verBuild = (uifVersion - verMaj * 10000) - verMin * 100;
	printf("ANSLI VersionInfo:  ANSLUIF: V%d.%d.%d   ANSLONLINE: V%d\n", verMaj, verMin, verBuild, oifVersion);
}

/***************************************************************************************************/
/** MAIN PROGRAM ***********************************************************************************/

int main(int argc,char *argv[]) 
{
	int logLevel = 0;
	unsigned long t1, t2, tH, tM, tS, tMS;
	t1 = GetSystemTickCount();
	printf("SystemTickCount=%lu\n", t1);
	
	_bBreakThread = 0;
	_RunningState = 0;
	_CPUConnClosed = 0;
	_bConnectionClosed = 0;
	memset(strIP, 0, 16); 		
	memset(strConPara, 0, 256);

	/** begin: init and connect **/
	sprintf(strIP, "%s", (argc >= 2) ? argv[1] : TARGET_IP); /* IP-Adresse */
	if (argc >= 4)
	{
		sprintf(strConPara, "Port=\"%s\" Timeout=\"%s\" %s", argv[3], argv[2], CONNECTION_PARAMS_NO_TOPO);
	} else if (argc >= 3) {
		sprintf(strConPara, "Timeout=\"%s\" %s", argv[2], CONNECTION_PARAMS_NO_TO);
	} else {
		sprintf(strConPara, "%s", CONNECTION_PARAMS);
	}
	if (argc >= 10)
	{
		sprintf(_strPV1, "%s", argv[4]);
		sprintf(_strPV2, "%s", argv[5]);
		rfPV1 = atoi(argv[6]);
		rfPV2 = atoi(argv[7]);
		sprintf(_strTask, "%s", argv[8]);
		if ('.' == _strTask[0])
		{
		    _strTask[0] = 0;
		}
		logLevel = atoi(argv[9]);
	} else if (argc >= 9)
	{
		sprintf(_strPV1, "%s", argv[4]);
		sprintf(_strPV2, "%s", argv[5]);
		rfPV1 = atoi(argv[6]);
		rfPV2 = atoi(argv[7]);
		sprintf(_strTask, "%s", argv[8]);
	} else if (argc >= 8) {
		sprintf(_strPV1, "%s", argv[4]);
		sprintf(_strPV2, "%s", argv[5]);
		rfPV1 = atoi(argv[6]);
		rfPV2 = atoi(argv[7]);
	} else if (argc >= 7) {
		sprintf(_strPV1, "%s", argv[4]);
		sprintf(_strPV2, "%s", argv[5]);
		rfPV1 = atoi(argv[6]);
	} else if (argc >= 6) {
		sprintf(_strPV1, "%s", argv[4]);
		sprintf(_strPV2, "%s", argv[5]);
	} else if (argc >= 5){
		sprintf(_strPV1, "%s", argv[4]);
		sprintf(_strPV2, "%s", VAR_NAME_SET);
	} else {
		sprintf(_strPV1, "%s", VAR_NAME_READ);
		sprintf(_strPV2, "%s", VAR_NAME_SET);
	}


	printf("+ANSLSample  IPAddress=\"%s\"\n", strIP);
	printf("             %s\n", strConPara);
	printf("             Task=\"%s\" PV1=\"%s\" RF1=%d PV2=\"%s\" RF2=%d   LOGGING=%d\n\n", _strTask, _strPV1, rfPV1, _strPV2, rfPV2, logLevel);
	printf("Press h to get the command line options!\n");
	printf("Press 0 to stop ANSL communication!\n");
	printf("Press 1 to read CPU date and time!\n");
	printf("Press 2 to read PV1!\n");
	printf("Press 3 to read PV2!\n");
	printf("Press 4 to PV1++ and PV2++!\n");
	printf("Press 5 to set PV1!\n");
	printf("Press 6 to set PV2!\n");
	printf("Press 7 to PV1-- and PV2--!\n");
	printf("Press 8 for WarmRestart!\n");
	printf("Press 9 for ColdRestart!\n");
	printf("Press a for get ANSL user and online interface version info!\n");
	printf("Press b for resume task class #1!\n");
	printf("Press c for suspend task class #1!\n");
	printf("Press d for resume task for 10 cycles!\n");
	printf("Press e for resume task for 10 cycles!\n");
	printf("Press f for clear logger buffer!\n");
	printf("Press g for clear CPU RAM!\n");
	printf("Press i for reading CPU info!\n");
	printf("Press j for reading HW info!\n");
	printf("Press k for reading memory info!\n");
	printf("Press l for list the variables!\n");
	printf("\n");

	PrintVInfo(); 
	if(sizeof(int*) == 4)
	{
		printf("Running as a 32 BIT application!\n\n\n");
	} else {
		printf("Running as a 64 BIT application!\n\n\n");
	}
	/* turn on logging*/
	ANSLI_LOGGING_SetLevel(logLevel);

	/* init ANSL */
	status = ANSLI_Initialize();
	if (ErrorCheck(status, "ERROR ANSLI_Initialize") != 0) return -1;

#ifndef _NO_AUTO_CONNECT
	/* connect to CPU object */
	status = connectCPU(strIP, strConPara, (unsigned int*)&identCpu380, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataCpu);
	if (0 != status) { return -1; }
	pvParentIdent = identCpu380;
	
	/* connect to task object */
	if (0 != _strTask[0])
	{
		connectTask(identCpu380, _strTask, (unsigned int*)&identTask, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataTask);
		if (0 != status) { return -1; }
		pvParentIdent = identTask;
	}
	
	/* connect to task object */
	connectLogger(identCpu380, "$arlogsys", (unsigned int*)&identLogger, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataLogger);
	if (0 != status) { return -1; }
	
	// connect to variable VAR_NAME_READ
	if (0 != _strPV1[0])
	{
		connectVariable(pvParentIdent, _strPV1, (unsigned int*)&identVarRead, ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV1, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataRead);
		if (0 != status) { return -1; }
	}
	
	/* connect to variable VAR_NAME_SET */
	if (0 != _strPV2[0])
	{
		connectVariable(pvParentIdent, _strPV2, (unsigned int*)&identVarSet, ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV2, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataSet);
		if (0 != status) { return -1; }
	}
#endif // !#ifdef _NO_AUTO_CONNECT

#ifdef _WORKERTHREAD
	// either start a new thread
	CreateWorkerThread(&_RunningState, &_MutexObject, &_ThreadId, WorkerThread);
#else
	// or do the work in a loop until 0 pressed
	StartNoThread();
#endif

#ifdef _WORKERTHREAD
	pthread_join(_ThreadId, NULL);
#endif

	t2 = GetSystemTickCount();
	tS = ((t2-t1)/1000);
	tMS = ((t2-t1)%1000);
	tM = tS/60;
	tS = tS%60;
	tH = tM/60;
	tM = tM%60;
	printf("-ANSLSample    Execution time: %luh %lum %lus %lums\n", tH, tM, tS, tMS);

	/** begin: deinit and disconnect **/
	return EXIT_SUCCESS;
}

/* connect to CPU object */
int connectCPU(char* strIP, char* strConPara, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData)
{
	int status = 0;
	status = ANSLI_CPU_Connect(strIP, strConPara, (unsigned int*)&identCpu380, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataCpu);
	if (ErrorCheck(status, "ERROR ANSLI_CPU_Connect") != 0) return -1;
	return status;
}

/* connect to task object */
int connectTask(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData)
{
	int status = 0;
	if (0 != name[0])
	{
		status = ANSLI_TASK_Connect(parentID, name, ident, (ANSL_CALL_BACK)fpCB, userData);
		ErrorCheck(status, "ERROR ANSLI_TASK_Connect TASK_NAME");
	}
	return status;
}

/* connect to task object */
int connectLogger(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData)
{
	int status = 0;
	if (0 != name[0])
	{
		status = ANSLI_LOGGER_Connect(parentID, name, ident, ANSL_LOGM_EV_FLAG_NONE, (ANSL_CALL_BACK)fpCB, userData);
		ErrorCheck(status, "ERROR ANSLI_LOG_Connect LOGGER_NAME");
	}
	return status;
}

/* connect to variable object */
int connectVariable(unsigned int parentID, char* name, unsigned int* ident, int flags, int refresh, ANSL_CALL_BACK fpCB, void* userData)
{
	int status = 0;
	ConnectPV(name, parentID, flags, refresh, userData, ident);
	status = ANSLI_VAR_ReadDataType(*ident);
	ErrorCheck(status, "ERROR ANSLI_VAR_ReadDataType VAR_NAME_READ");
	return status;
}


unsigned long GetSystemTickCount() // Zeit seit dem Booten in Sekunden
{
#ifdef WIN32
	return GetTickCount();
#else
	struct timeval tv;
	gettimeofday(&tv,NULL);
	return (unsigned long)(tv.tv_sec*1000 + (tv.tv_usec / 1000));
#endif
}

unsigned long GetSystemTickCountA() // Zeit seit dem Booten in Sekunden
{
#ifdef WIN32
	return GetTickCount();
#else
	struct timespec abstime;
	clock_gettime(CLOCK_REALTIME, &abstime);
	return (unsigned long)(abstime.tv_sec*1000 + (abstime.tv_nsec / 1000));
#endif
}

unsigned long GetSystemTickCountB() // Zeit seit dem Booten in Sekunden
{
#ifdef WIN32
	return GetTickCount();
#else
	struct timeb tp;
	ftime (&tp);
	return (unsigned long)(tp.time*1000 + tp.millitm);
#endif
}

#ifdef _WORKERTHREAD
int CreateWorkerThread(int* pRunningState, T_PTHREAD_MUTEX* pMutex, T_PTHREAD* pThreadId, void* fpThread)
{
	int retVal = 0; 
	*pRunningState = 0;
#ifdef WIN32
	*pThreadId = CreateThread(NULL, 0, fpThread, NULL, 0, NULL);

	WaitForSingleObject(*pThreadId, 0xFFFFFFFF);
	
	CloseHandle(*pThreadId);
#else

	pthread_mutexattr_t mutex_attr;
	pthread_attr_t	threadAttr;

	/* create event objekt: */
	pthread_mutexattr_init (&mutex_attr);
	retVal = pthread_mutex_init(pMutex, &mutex_attr);
	if(retVal < 0)
	{
		printf( "Mutex NOT created.\n" );
		return RC_OBJECT_NOT_CREATED;
	}

	retVal = pthread_mutex_lock (pMutex);
	if (-1 == retVal)
	{
	   printf( "Mutex LOCK ERROR.\n" );
	   return RC_LOCK_ERROR;
	}

	retVal = pthread_attr_init(&threadAttr);
	if(retVal < 0)
	{
		printf( "pthread_attr_init ERROR.\n" );
		return RC_THREAD_ATTR_ERROR;
	}

	retVal = pthread_attr_setstacksize(&threadAttr, 120*1024);
	if(retVal < 0)
	{
		printf( "pthread_attr_setstacksize ERROR.\n" );
		return RC_STACKSIZE_ERROR;
	}

	retVal = pthread_mutex_unlock (pMutex);
	if (-1 == retVal)
	{
	   printf( "Mutex UNLOCK ERROR.\n" );
	   return RC_UNLOCK_ERROR;
	}

	//printf("pthread_create...\n");
	retVal = pthread_create(pThreadId, NULL/*&threadAttr*/, (void*)fpThread, pMutex);

	if(retVal != 0)
	{
		pthread_mutex_destroy(pMutex);
		printf( "Thread NOT created!\n" );
		return RC_THREAD_NOT_CREATED;
	}
#endif
	return status;
}
#endif

#ifdef _WORKERTHREAD
#ifdef WIN32
static unsigned long __stdcall WorkerThread (void* pParam)
#else
static void* WorkerThread(void* pParam)
#endif
{

	T_PTHREAD_MUTEX* pMutex = (T_PTHREAD_MUTEX*)pParam;

	int status;

	// Nun kann losgeloest von der restlichen Kommunikation
	// mit der Datenaufbereitung begonnen werden
	while (0 == _bBreakThread)
	{
		status = pthread_mutex_lock(pMutex);
		if (0 != status)
		{
			printf("pthread_mutex_lock   RC_LOCK_ERROR!\n");
		}
		DoWork();
		pthread_mutex_unlock(pMutex);
	}
	UninitializeObjects();	
	while (0 == _bConnectionClosed)
	{
		status = pthread_mutex_lock(pMutex);
		if (0 != status)
		{
			printf("pthread_mutex_lock   RC_LOCK_ERROR!\n");
		}
		DoExit();
		pthread_mutex_unlock(pMutex);
	}

	return (0);
}
#endif

void DoWork()
{
	unsigned int status = 0;
	int charRead = 0;
	switch (_RunningState)
	{
		case 0:
		{
			_RunningState = 1;
		} break;
		case 1:
		{
			_RunningState = 2;
		} break;
		case 2:
		{
			charRead = ioGetch();
			switch (charRead)
			{
				case 48:
				{
					printf("Exiting...\n");
					_RunningState = 3;
				} break;
				case 97:
				{
					printf("Getting user and online interface version info...\n");
					AnslGetVersionInfo();
				} break;
				case 98:
				{
					printf("Resuming TaskClass...\n");
					TaskClassResume(identCpu380, 1);
				} break;
				case 99:
				{
					printf("Suspend TaskClass...\n");
					TaskClassSuspend(identCpu380, 1);
				} break;
				case 100:
				{
					printf("Resuming Task...\n");
					TaskResume(identTask, 10);
				} break;
				case 101:
				{
					printf("Suspend Task...\n");
					TaskSuspend(identTask, 10);
				} break;
				case 102:
				{
					printf("Clear Buffer for Logger...\n");
					LoggerClearBuffer(identLogger);
				} break;
				case 103:
				{
					printf("Clear CPU RAM...\n");
					CpuClearMemory(identCpu380, ANSL_CPU_CLEAR_RAM);
				} break;
				case 104:
				{
					printf("Optional parameters when executing this demo:\n");
					printf("[ip address] [timeout] [port] [1st PV] [2nd PV] [refresh 1st PV] [refresh 2nd PV] [task] [verbose level for logging]\n");
					printf("If task name is given assume the variables are TASK local else CPU global!\n");
					printf("When using logging for global variables set task to \".\"!\n");
				} break;
				case 105:
				{
					printf("Reading CPU info...\n");
					status = ANSLI_CPU_ReadInfo(identCpu380);
					if (status != 0)
					{
						printf("ERROR CPU read info: %i\n",status);
					}
				} break;
				case 106:
				{
					printf("Reading HW info...\n");
					status = ANSLI_CPU_GetHWComponents(identCpu380);
					if (status != 0)
					{
						printf("ERROR CPU read HW info: %i\n",status);
					}
				} break;
				case 107:
				{
					printf("Reading memory info...\n");
					status = ANSLI_CPU_ReadMemInfo(identCpu380);
					if (status != 0)
					{
						printf("ERROR CPU read memory info: %i\n",status);
					}
				} break;
				
				case 108:
				{
					printf("List variables\n");
					status = ANSLI_TASK_ListVariables(pvParentIdent,0xFFFF);
					ErrorCheck(status,"ERROR List Variables");
				}break;
				case 109:	
				{
					printf("List modules\n");
					status = ANSLI_CPU_ListModules(identCpu380,ANSL_MODULE_SYS_TASK,0);
					ErrorCheck(status,"ERROR List Module");
				}break;	
				case 110:
				{
					printf("List CPU variables\n");
					status = ANSLI_CPU_ListVariables(identCpu380,0xFFFF);
					ErrorCheck(status,"ERROR List Variables");
				}break;	
				case 111:
				{
					printf("UploadModule\n");
					status = ANSLI_CPU_UpLoadModule(identCpu380,"pvitest","/home/bur/pvitest.br");
					ErrorCheck(status, "ERROR Upload Module");
				} break;
				case 112:
				{
					printf("DeleteModule\n");
					status = ANSLI_CPU_DeleteModule(identCpu380,"pvitest");
				}break;
				case 113:
				{
					printf("DownloadModule\n");
					status = ANSLI_CPU_DownLoadModule(identCpu380,"/home/bur/pvitest.br",ANSL_MEMTYP_TRSF,ANSL_INST_OVERLOAD,"pvitest");
					ErrorCheck(status,"ERROR Download Modules.\n");
				}break;

				case 114:
				{
					printf("UploadModule\n");
					status = ANSLI_CPU_UpLoadModule(identCpu380,"arconfig","/home/bur/arconfig.br");
					ErrorCheck(status, "ERROR Upload Module");
				} break;
				case 115:
				{
					printf("DeleteModule\n");
					status = ANSLI_CPU_DeleteModule(identCpu380,"arconfig");
				}break;
				case 116:
				{
					printf("DownloadModule\n");
					status = ANSLI_CPU_DownLoadModule(identCpu380,"/home/bur/arconfig.br",ANSL_MEMTYPE_USER_FLASH,ANSL_INST_ONECYCLE,"arconfig");
					ErrorCheck(status,"ERROR Download Modules.\n");
				}break;
				case 49:
				{
					printf("reading CPU date and time...\n");
					ReadCpuDateTime(identCpu380);
				} break;
				case 50:
				{
					printf("reading \"%s\"...\n", _strPV1);
					ReadVariable(identVarRead);
				} break;
				case 53:
				{
					printf("writing \"%s\"...\n", _strPV1);
					WriteVariable(identVarRead, var3);
				} break;
				case 51:
				{
					printf("reading \"%s\"...\n", _strPV2);
					ReadVariable(identVarSet);
				} break;
				case 54:
				{
					printf("writing \"%s\"...\n", _strPV2);
					WriteVariable(identVarSet, var3);
				} break;
				case 52:
				{
					var3++;
					var4++;
					printf("increment write buffers   %d\n", var3);
				} break;
				case 55:
				{
					var3--;
					var4--;
					printf("decrement write buffers   %d\n", var3);
				} break;
				case 56:
				{
					printf("WarmRestart\n");
					status = ANSLI_CPU_Restart(identCpu380, ANSL_CPU_WARM_RESTART);
					ErrorCheck(status, "ANSLI_CPU_Restart");
				} break;
				case 57:
				{
					printf("ColdRestart\n");
					status = ANSLI_CPU_Restart(identCpu380, ANSL_CPU_COLD_RESTART);
					ErrorCheck(status, "ANSLI_CPU_Restart");
				} break;
				
#ifdef _NO_AUTO_CONNECT
				case 99:
				{
					printf("Connect CPU\n");
					status = connectCPU(strIP, strConPara, (unsigned int*)&identCpu380, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataCpu);
					ErrorCheck(status, "Connect CPU");
					pvParentIdent = identCpu380;
				} break;
				case 1160:
				{
					printf("connect to task object\n");
					connectTask(identCpu380, _strTask, (unsigned int*)&identTask, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataTask);
					ErrorCheck(status, "Connect TASK");
					pvParentIdent = identTask;
				} break;
				case 1140:
				{
					printf("connect to variable VAR_NAME_READ\n");
					connectVariable(pvParentIdent, _strPV1, (unsigned int*)&identVarRead, ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV1, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataRead);
					ErrorCheck(status, "Connect VAR_NAME_READ");				
				} break;
				case 119:
				{
					printf("connect to variable VAR_NAME_SET\n");
					connectVariable(pvParentIdent, _strPV2, (unsigned int*)&identVarSet, ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV2, (ANSL_CALL_BACK)ANSLCB_Cpu, (void*)&userDataSet);
					ErrorCheck(status, "Connect VAR_NAME_SET");
				} break;
#endif //_NO_AUTO_CONNECT
			}
		} break;
		case 3:
		{
			_bBreakThread = 1;
		} break;
	}
}

void StartNoThread()
{
	// read date and time of the cpu
	status = ANSLI_CPU_DTRead(identCpu380);
	ErrorCheck(status, "ANSLI_CPU_DTRead");

	while (0 == _bBreakThread)
	{
		DoWork();
	}
	UninitializeObjects();	
	while (0 == _bConnectionClosed)
	{
		DoExit();
	}
	/** end: working **/
}

int UninitializeObjects()
{
	int status = 0;
	_bConnectionClosed = 1;
	/* cancel the PV connection */
	if (0 != identVarRead)
	{
		status = ANSLI_VAR_Disconnect(identVarRead);
		if (ErrorCheck(status, "ERROR ANSLI_VAR_Disconnect") != 0) return -1;
	}
	if (0 != identVarSet)
	{
		status = ANSLI_VAR_Disconnect(identVarSet);
		if (ErrorCheck(status, "ERROR ANSLI_VAR_Disconnect") != 0) return -1;
	}

	/* cancel the task connection */
	if (0 != identTask)
	{
		status = ANSLI_TASK_Disconnect(identTask);
		if (ErrorCheck(status, "ERROR ANSLI_TASK_Disconnect") != 0) return -1;
	}

	/* cancel the CPU connection */
	if (0 != identCpu380)
	{
		status = ANSLI_CPU_Disconnect(identCpu380);
		if (ErrorCheck(status, "ERROR ANSLI_CPU_Disconnect") != 0) return -1;
		_bConnectionClosed = 0;
	}

	/* deinit ANSL */
	// Wait for deinit in CPU callback!!

	return 0;
}

void DoExit()
{
	if (1 == _CPUConnClosed)
	{
		ANSLI_DeInitialize();
		_bConnectionClosed = 1;
	}
}

unsigned int ReadCpuDateTime(unsigned int cpuIdent)
{
	unsigned int status = 0;

	status = ANSLI_CPU_DTRead(cpuIdent);
	if (status != 0)
	{
		printf("ERROR READ DT : %i\n",status);
	}
	return status;
}

unsigned int ConnectPV(char *name, unsigned int parentIdent, unsigned int evFlags, unsigned int refresh, void *pUserData, unsigned int *pIdent)
{
	unsigned int status = 0;
	struct ANSL_VARIABLE_EVENT_PARAMETER 	evParam;
	evParam.m_EventFlags = evFlags;
	evParam.m_RefreshTime = refresh;
	evParam.m_IntegerHyst = 0;
	evParam.m_FloatHyst = 0;

	status = ANSLI_VAR_Connect(parentIdent, name, pIdent, (ANSL_CALL_BACK)ANSLCB_Cpu, pUserData, (ANSL_VARIABLE_EVENT_PARAMETER*)&evParam);
	if (0 != status)
	{
		printf("ANSLI_VAR_Connect  \"%s\"   E#%i\n", name, status);
	}
	return status;
}

unsigned int ReadVariable(unsigned int varIdent)
{
	unsigned int status = 0;

	//printf("Reading PV ident=%u \n", varIdent);
	status = ANSLI_VAR_ReadValue(varIdent, NULL, 0);
	if (status !=0)
	{
		printf("ERROR %u READ PV ident=%u \n", status, varIdent);
	}
	return status;
}

unsigned int WriteVariable(unsigned int varIdent, long varValue)
{
	unsigned int status = 0;

	printf("Write PV ident=%u to %d\n", varIdent, (int)varValue);
	status = ANSLI_VAR_SetValue(varIdent, &varValue, sizeof(varValue));
	if (status !=0)
	{
		printf("ERROR %u SET PV ident=%u \n", status, varIdent);
	}
	return status;
}

int ioGetch ()
{
	 int ch;
#ifdef WIN32
	 ch = _getch();
#else
	 struct termios old_t, tmp_t;

	 // alte Attribute holen
	 if (tcgetattr (STDIN_FILENO, &old_t))
	 {
		  // moeglicherweise kein Terminal
		  // moeglicherweise anderer Fehler
		  return -1;
	 }
	 // struktur kopieren
	 memcpy (&tmp_t, &old_t, sizeof (old_t));
	 // Zeilenmodus sowie Spezialzeichen deaktivieren, Bildschirmausgabe von eingegebenen Zeichen deaktivieren
	 tmp_t.c_lflag &= ~ICANON & ~ECHO;
	 // Neue Flags setzen
	 if (tcsetattr (STDIN_FILENO, TCSANOW, (const struct termios *)&tmp_t))
	 {
		  // moeglicherweise irgend ein Fehler
		  return -1;
	 }
	 // zeichen einlesen (darauf verlassen, STDIN_FILENO blockierend ist)
	 ch = getchar ();
	 // Alte Flags zuruecksetzen
	 // Rueckgabewert ignorieren, da das Terminal bei Fehler sowieso
	 // nicht korrekt reagieren wird, wir dagegen aber nichts tun koennen
	 tcsetattr (STDIN_FILENO, TCSANOW, (const struct termios *)&old_t);
#endif
	 // Zeichen zurueckgeben
	 return ch;
}

/***************************************************************************************************/
/** ERROR HANDLING *********************************************************************************/

int ErrorCheck(unsigned int status, const char* error)
{
	if (status != 0)
	{
		printf("%s: %i\n", error, status);
		return 1;
	}
	return 0;
}

/***************************************************************************************************/
/** CALLBACK FUNCTION ******************************************************************************/
#ifdef WIN32
static unsigned int __stdcall ANSLCB_Cpu(unsigned int errorCode, 
										 unsigned int ident, 
										 unsigned int objectType, 
										 unsigned int cbType, 
										 void* pData, 
										 unsigned int dataLen, 
										 void* pUserData)
#else
unsigned int ANSLCB_Cpu(unsigned int errorCode, 
						unsigned int ident, 
						unsigned int objectType,
						unsigned int cbType, 
						void* pData, 
						unsigned int dataLen, 
						void* pUserData)
#endif
{	
	char strOut[6000];
	char strName[128];
	char strFmt[FMT_STR_SIZE];
	char strTyp[INFO_STR_SIZE];
	char strEvent[INFO_STR_SIZE];
	memset(strOut, 0, 6000);
	memset(strName, 0, 128);
	memset(strFmt, 0, FMT_STR_SIZE);
	memset(strTyp, 0, INFO_STR_SIZE);
	memset(strEvent, 0, INFO_STR_SIZE);

	GetObjTypeString(objectType, strTyp);
	switch (cbType)
	{
		case ANSL_RESP_CONNECT:
		{
			sprintf(strOut, "CONNECT  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			printf(strOut);
		} break;
		case ANSL_RESP_DISCONNECT:
		{
			sprintf(strOut, "DISCONNECT  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			printf(strOut);
			_CPUConnClosed = 1;
		} break;
		case ANSL_EVENT_ERROR:
		{
			sprintf(strOut, "ERROR  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			printf(strOut);
		} break;
		case ANSL_EVENT_STATUS:
		{
			sprintf(strOut, "STATUS  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			printf(strOut);
		} break;
		case ANSL_EVENT_FINAL:
		{
			sprintf(strOut, "FINAL  %s error=%u ident=%u   serviceNumber=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			printf(strOut);
		} break;
		/* read date and time */
		case ANSL_RESP_GET_DATETIME:
		{
			CPU_DATETIME_TYPE cpuDataTime;

			if	(
					(errorCode == 0) &&
					(sizeof(CPU_DATETIME_TYPE) == dataLen)
				)
			{
				memcpy(&cpuDataTime, pData, dataLen);
				//                        year starts at 1900         month starts at 0    
				printf("Date: %i/%i/%i ", cpuDataTime.tm_Year + 1900, cpuDataTime.tm_Month + 1, cpuDataTime.tm_DayOfMonth);
				printf("Time: %i:%i:%i\n", cpuDataTime.tm_Hour, cpuDataTime.tm_Minute, cpuDataTime.tm_Second);
			}
			else
			{
				printf("ERROR GET_DATETIME: %i\n", errorCode);
			}
		} break;

		/* read READ_VAR */
		case ANSL_RESP_GET_VALUE:
		/* read READ_VAR */
		case ANSL_EVENT_VALUE:
		{
			if (*((unsigned int*)pUserData) == 1)
			{
				sprintf(strName, "%s", _strPV1);
			}
			if (*((unsigned int*)pUserData) == 2)
			{
				sprintf(strName, "%s", _strPV2);
			}
			GetCBTypeString(cbType, strEvent);
			sprintf(strOut, "%s \"%s\": %s  error=%u ident=%u   dataLen=%u\n", 
					strEvent,
					strName,
					((0 == errorCode) ? DataToString(pData, dataLen, ((*((unsigned int*)pUserData) == 1)? &_DataTypePV1 : &_DataTypePV2)) : "ERROR"), 
					errorCode, 
					ident, 
					dataLen);	
			printf(strOut);
		} break;

		/* read format SET_VAR */
		case ANSL_EVENT_FORMAT:
		case ANSL_RESP_GET_FORMAT:
		{
			if (*((unsigned int*)pUserData) == 1)
			{
				memcpy(&_DataTypePV1, pData, sizeof(ANSL_VARIABLE_DATA_TYPE)); 
				sprintf(strName, "%s", _strPV1);
			}
			if (*((unsigned int*)pUserData) == 2)
			{
				memcpy(&_DataTypePV2, pData, sizeof(ANSL_VARIABLE_DATA_TYPE)); 
				sprintf(strName, "%s", _strPV2);
			}
			memset (strFmt, 0, FMT_STR_SIZE);
			if (0 == errorCode)
			{
				GetFormatString((ANSL_VARIABLE_DATA_TYPE*)pData, strFmt);
			} else {
				sprintf(strFmt, "???");
			}
			GetCBTypeString(cbType, strEvent);
			sprintf(strOut, "%s  %s \"%s\": %s  error=%u ident=%u   dataLen=%u\n", 
					strTyp, 
					strEvent,
					strName,
					strFmt, 
					errorCode, 
					ident, 
					dataLen);	
			printf(strOut);

		} break;
		case ANSL_RESP_SET_ATTRIBUTES:
		{
			if (*((unsigned int*)pUserData) == 1)
			{
				sprintf(strName, "%s", _strPV1);
			}
			if (*((unsigned int*)pUserData) == 2)
			{
				sprintf(strName, "%s", _strPV2);
			}
			GetCBTypeString(cbType, strEvent);
			sprintf(strOut, "%s \"%s\": error=%u ident=%u\n", 
					strEvent,
					strName,					 
					errorCode, 
					ident);	
			printf(strOut);
		} break;
		/* set SET_VAR */
		case ANSL_RESP_SET_VALUE:
		{
			if (*((unsigned int*)pUserData) == 1)
			{
				sprintf(strName, "%s", _strPV1);
			}
			if (*((unsigned int*)pUserData) == 2)
			{
				sprintf(strName, "%s", _strPV2);
			}
			GetCBTypeString(cbType, strEvent);
			sprintf(strOut, "%s \"%s\": error=%u ident=%u\n", 
					strEvent,
					strName,					 
					errorCode, 
					ident);	
			printf(strOut);
		} break;
		case ANSL_RESP_GET_HARDWARE:
		{
			sprintf(strOut, "%s\n", (char*)pData);	
			printf(strOut);
		} break;
		case ANSL_RESP_GET_INFO:
		{
			sprintf(strOut, "%s\n", (char*)pData);	
			printf(strOut);
		} break;
		case ANSL_RESP_GET_STORAGE:
		{
			sprintf(strOut, "%s\n", (char*)pData);	
			printf(strOut);
		} break;
		case ANSL_RESP_LIST_MODULES:
		{
			ANSL_MODULE_DESC* pModDesc;
			pModDesc = (ANSL_MODULE_DESC*)pData;
			while(NULL != pModDesc)
			{
				if(NULL != pModDesc->m_Name)
				{
					sprintf(strOut,"Module Name=%s Type=%02X\n",pModDesc->m_Name,pModDesc->m_Type);
					printf(strOut);
				}
				pModDesc = pModDesc->m_pNext;
			}
		}break;
		case ANSL_RESP_LIST_VARIABLES: //List Task Variables.
		{
			ANSL_VARIABLE_DESC* pPVDesc;
			pPVDesc = (ANSL_VARIABLE_DESC*)pData;
			while(NULL != pPVDesc)
			{
				if(NULL != pPVDesc->m_Name)
				{
					sprintf(strOut,"PV name = %s  \n",pPVDesc->m_Name);
					printf(strOut);
				}
				pPVDesc = pPVDesc->m_pNext;
			}
		}break;
		default:
		{
			GetCBTypeString(cbType, strEvent);
			sprintf(strOut, "%s  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, 
					strEvent, 
					errorCode, 
					ident, 
					dataLen);	
			printf(strOut);
		} break;
	}
	return 0;
}

int GetCBTypeString(unsigned int eventCode, char strValue[INFO_STR_SIZE])
{
	memset(strValue, 0, INFO_STR_SIZE);
	switch (eventCode)
	{
		case ANSL_EVENT_FINAL: { strcpy(strValue, "EVENT_FINAL"); } break;
		case ANSL_EVENT_ERROR: { strcpy(strValue, "EVENT_ERROR"); } break;
		case ANSL_EVENT_STATUS: { strcpy(strValue, "EVENT_STATUS"); } break;
		case ANSL_REQ_CONNECT: { strcpy(strValue, "REQ_CONNECT"); } break;
		case ANSL_RESP_CONNECT: { strcpy(strValue, "RESP_CONNECT"); } break;
		case ANSL_CANCEL_CONNECT: { strcpy(strValue, "CANCEL_CONNECT"); } break;
		case ANSL_REQ_DISCONNECT: { strcpy(strValue, "REQ_DISCONNECT"); } break;
		case ANSL_RESP_DISCONNECT: { strcpy(strValue, "RESP_DISCONNECT"); } break;
		case ANSL_CANCEL_DISCONNECT: { strcpy(strValue, "CANCEL_DISCONNECT"); } break;
		case ANSL_REQ_GET_DATETIME: { strcpy(strValue, "REQ_GET_DATETIME"); } break;
		case ANSL_RESP_GET_DATETIME: { strcpy(strValue, "RESP_GET_DATETIME"); } break;
		case ANSL_CANCEL_GET_DATETIME: { strcpy(strValue, "CANCEL_GET_DATETIME"); } break;
		case ANSL_REQ_SET_DATETIME: { strcpy(strValue, "REQ_SET_DATETIME"); } break;
		case ANSL_RESP_SET_DATETIME: { strcpy(strValue, "RESP_SET_DATETIME"); } break;
		case ANSL_CANCEL_SET_DATETIME: { strcpy(strValue, "CANCEL_SET_DATETIME"); } break;
		case ANSL_REQ_WARM_RESTART: { strcpy(strValue, "REQ_WARM_RESTART"); } break;
		case ANSL_RESP_WARM_RESTART: { strcpy(strValue, "RESP_WARM_RESTART"); } break;
		case ANSL_CANCEL_WARM_RESTART: { strcpy(strValue, "CANCEL_WARM_RESTART"); } break;
		case ANSL_REQ_COLD_RESTART: { strcpy(strValue, "REQ_COLD_RESTART"); } break;
		case ANSL_RESP_COLD_RESTART: { strcpy(strValue, "RESP_COLD_RESTART"); } break;
		case ANSL_CANCEL_COLD_RESTART: { strcpy(strValue, "CANCEL_COLD_RESTART"); } break;
		case ANSL_REQ_GET_INFO: { strcpy(strValue, "REQ_GET_INFO"); } break;
		case ANSL_RESP_GET_INFO: { strcpy(strValue, "RESP_GET_INFO"); } break;
		case ANSL_CANCEL_GET_INFO: { strcpy(strValue, "CANCEL_GET_INFO"); } break;
		case ANSL_REQ_GET_STORAGE: { strcpy(strValue, "REQ_GET_STORAGE"); } break;
		case ANSL_RESP_GET_STORAGE: { strcpy(strValue, "RESP_GET_STORAGE"); } break;
		case ANSL_CANCEL_GET_STORAGE: { strcpy(strValue, "CANCEL_GET_STORAGE"); } break;
		case ANSL_REQ_GET_TASKCLASSES: { strcpy(strValue, "REQ_GET_TASKCLASSES"); } break;
		case ANSL_RESP_GET_TASKCLASSES: { strcpy(strValue, "RESP_GET_TASKCLASSES"); } break;
		case ANSL_CANCEL_GET_TASKCLASSES: { strcpy(strValue, "CANCEL_GET_TASKCLASSES"); } break;
		case ANSL_REQ_GET_HARDWARE: { strcpy(strValue, "REQ_GET_HARDWARE"); } break;
		case ANSL_RESP_GET_HARDWARE: { strcpy(strValue, "RESP_GET_HARDWARE"); } break;
		case ANSL_CANCEL_GET_HARDWARE: { strcpy(strValue, "CANCEL_GET_HARDWARE"); } break;
		case ANSL_REQ_UPLOAD: { strcpy(strValue, "REQ_UPLOAD"); } break;
		case ANSL_RESP_UPLOAD: { strcpy(strValue, "RESP_UPLOAD"); } break;
		case ANSL_CANCEL_UPLOAD: { strcpy(strValue, "CANCEL_UPLOAD"); } break;
		case ANSL_REQ_DOWNLOAD: { strcpy(strValue, "REQ_DOWNLOAD"); } break;
		case ANSL_RESP_DOWNLOAD: { strcpy(strValue, "RESP_DOWNLOAD"); } break;
		case ANSL_CANCEL_DOWNLOAD: { strcpy(strValue, "CANCEL_DOWNLOAD"); } break;
		case ANSL_REQU_CREATE_MODULE: { strcpy(strValue, "REQU_CREATE_MODULE"); } break;
		case ANSL_RESP_CREATE_MODULE: { strcpy(strValue, "RESP_CREATE_MODULE"); } break;
		case ANSL_CANCEL_CREATE_MODULE: { strcpy(strValue, "CANCEL_CREATE_MODULE"); } break;
		case ANSL_REQU_DELETE_MODULE: { strcpy(strValue, "REQU_DELETE_MODULE"); } break;
		case ANSL_RESP_DELETE_MODULE: { strcpy(strValue, "RESP_DELETE_MODULE"); } break;
		case ANSL_CANCEL_DELETE_MODULE: { strcpy(strValue, "CANCEL_DELETE_MODULE"); } break;
		case ANSL_REQU_LIST_MODULES: { strcpy(strValue, "REQU_LIST_MODULES"); } break;
		case ANSL_RESP_LIST_MODULES: { strcpy(strValue, "RESP_LIST_MODULES"); } break;
		case ANSL_CANCEL_LIST_MODULES: { strcpy(strValue, "CANCEL_LIST_MODULES"); } break;
		case ANSL_EVENT_LIST_MODULES: { strcpy(strValue, "EVENT_LIST_MODULES"); } break;
		case ANSL_EVENT_FORMAT: { strcpy(strValue, "EVENT_FORMAT"); } break;
		case ANSL_REQ_GET_FORMAT: { strcpy(strValue, "REQ_GET_FORMAT"); } break;
		case ANSL_RESP_GET_FORMAT: { strcpy(strValue, "RESP_GET_FORMAT"); } break;
		case ANSL_CANCEL_GET_FORMAT: { strcpy(strValue, "CANCEL_GET_FORMAT"); } break;
		case ANSL_EVENT_VALUE: { strcpy(strValue, "EVENT_VALUE"); } break;
		case ANSL_CANCEL_EVENT_VALUE: { strcpy(strValue, "CANCEL_EVENT_VALUE"); } break;
		case ANSL_REQ_GET_VALUE: { strcpy(strValue, "REQ_GET_VALUE"); } break;
		case ANSL_RESP_GET_VALUE: { strcpy(strValue, "RESP_GET_VALUE"); } break;
		case ANSL_CANCEL_GET_VALUE: { strcpy(strValue, "CANCEL_GET_VALUE"); } break;
		case ANSL_REQ_SET_ATTRIBUTES: { strcpy(strValue, "REQ_SET_ATTRIBUTES"); } break;
		case ANSL_RESP_SET_ATTRIBUTES: { strcpy(strValue, "RESP_SET_ATTRIBUTES"); } break;
		case ANSL_CANCEL_SET_ATTRIBUTES: { strcpy(strValue, "CANCEL_SET_ATTRIBUTES"); } break;
		case ANSL_REQ_SET_VALUE: { strcpy(strValue, "REQ_SET_VALUE"); } break;
		case ANSL_RESP_SET_VALUE: { strcpy(strValue, "RESP_SET_VALUE"); } break;
		case ANSL_CANCEL_SET_VALUE: { strcpy(strValue, "CANCEL_SET_VALUE"); } break;
		case ANSL_REQ_GET_CONTENT: { strcpy(strValue, "REQ_GET_CONTENT"); } break;
		case ANSL_RESP_GET_CONTENT: { strcpy(strValue, "RESP_GET_CONTENT"); } break;
		case ANSL_CANCEL_GET_CONTENT: { strcpy(strValue, "CANCEL_GET_CONTENT"); } break;
		case ANSL_EVENT_CONTENT: { strcpy(strValue, "EVENT_CONTENT"); } break;
		case ANSL_CANCEL_EVENT_CONTENT: { strcpy(strValue, "CANCEL_EVENT_CONTENT"); } break;
		case ANSL_REQ_READ_ENTRIES: { strcpy(strValue, "REQ_READ_ENTRIES"); } break;
		case ANSL_RESP_READ_ENTRIES: { strcpy(strValue, "RESP_READ_ENTRIES"); } break;
		case ANSL_CANCEL_READ_ENTRIES: { strcpy(strValue, "CANCEL_READ_ENTRIES"); } break;
		case ANSL_REQ_NEW_LENTRY: { strcpy(strValue, "REQ_NEW_LENTRY"); } break;
		case ANSL_RESP_NEW_LENTRY: { strcpy(strValue, "RESP_NEW_LENTRY"); } break;
		case ANSL_CANCEL_NEW_LENTRY: { strcpy(strValue, "CANCEL_NEW_LENTRY"); } break;
		case ANSL_REQ_GLKN_LINKLIST: { strcpy(strValue, "REQ_GLKN_LINKLIST"); } break;
		case ANSL_RESP_GLKN_LINKLIST: { strcpy(strValue, "RESP_GLKN_LINKLIST"); } break;
		case ANSL_CANCEL_GLKN_LINKLIST: { strcpy(strValue, "CANCEL_GLKN_LINKLIST"); } break;
		case ANSL_REQ_GLKN_FORCEOFF: { strcpy(strValue, "REQ_GLKN_FORCEOFF"); } break;
		case ANSL_RESP_GLKN_FORCEOFF: { strcpy(strValue, "RESP_GLKN_FORCEOFF"); } break;
		case ANSL_CANCEL_GLKN_FORCEOFF: { strcpy(strValue, "CANCEL_GLKN_FORCEOFF"); } break;
		case ANSL_REQ_LKN_FORMAT: { strcpy(strValue, "REQ_LKN_FORMAT"); } break;
		case ANSL_RESP_LKN_FORMAT: { strcpy(strValue, "RESP_LKN_FORMAT"); } break;
		case ANSL_CANCEL_LKN_FORMAT: { strcpy(strValue, "CANCEL_LKN_FORMAT"); } break;
		case ANSL_REQ_LKN_DATA: { strcpy(strValue, "REQ_LKN_DATA"); } break;
		case ANSL_RESP_LKN_DATA: { strcpy(strValue, "RESP_LKN_DATA"); } break;
		case ANSL_CANCEL_LKN_DATA: { strcpy(strValue, "CANCEL_LKN_DATA"); } break;
		case ANSL_REQ_LKN_FORCEDATA: { strcpy(strValue, "REQ_LKN_FORCEDATA"); } break;
		case ANSL_RESP_LKN_FORCEDATA: { strcpy(strValue, "RESP_LKN_FORCEDATA"); } break;
		case ANSL_CANCEL_LKN_FORCEDATA: { strcpy(strValue, "CANCEL_LKN_FORCEDATA"); } break;
		case ANSL_REQ_LKN_FORCEON: { strcpy(strValue, "REQ_LKN_FORCEON"); } break;
		case ANSL_RESP_LKN_FORCEON: { strcpy(strValue, "RESP_LKN_FORCEON"); } break;
		case ANSL_CANCEL_LKN_FORCEON: { strcpy(strValue, "CANCEL_LKN_FORCEON"); } break;
		case ANSL_REQ_LKN_FORCEOFF: { strcpy(strValue, "REQ_LKN_FORCEOFF"); } break;
		case ANSL_RESP_LKN_FORCEOFF: { strcpy(strValue, "RESP_LKN_FORCEOFF"); } break;
		case ANSL_CANCEL_LKN_FORCEOFF: { strcpy(strValue, "CANCEL_LKN_FORCEOFF"); } break;
		case ANSL_REQ_LOGGER_INFO: { strcpy(strValue, "REQ_LOGGER_INFO"); } break;
		case ANSL_RESP_LOGGER_INFO: { strcpy(strValue, "RESP_LOGGER_INFO"); } break;
		case ANSL_CANCEL_LOGGER_INFO: { strcpy(strValue, "CANCEL_LOGGER_INFO"); } break;
		case ANSL_REQ_MODULE_INFO: { strcpy(strValue, "REQ_MODULE_INFO"); } break;
		case ANSL_RESP_MODULE_INFO: { strcpy(strValue, "RESP_MODULE_INFO"); } break;
		case ANSL_CANCEL_MODULE_INFO: { strcpy(strValue, "CANCEL_MODULE_INFO"); } break;
		case ANSL_REQ_TASK_INFO: { strcpy(strValue, "REQ_TASK_INFO"); } break;
		case ANSL_RESP_TASK_INFO: { strcpy(strValue, "RESP_TASK_INFO"); } break;
		case ANSL_CANCEL_TASK_INFO: { strcpy(strValue, "CANCEL_TASK_INFO"); } break;
		case ANSL_REQU_TASK_SECTION: { strcpy(strValue, "REQU_TASK_SECTION"); } break;
		case ANSL_RESP_TASK_SECTION: { strcpy(strValue, "RESP_TASK_SECTION"); } break;
		case ANSL_CANCEL_TASK_SECTION: { strcpy(strValue, "CANCEL_TASK_SECTION"); } break;
		case ANSL_REQU_LIST_VARIABLES: { strcpy(strValue, "REQU_LIST_VARIABLES"); } break;
		case ANSL_RESP_LIST_VARIABLES: { strcpy(strValue, "RESP_LIST_VARIABLES"); } break;
		case ANSL_CANCEL_LIST_VARIABLES: { strcpy(strValue, "CANCEL_LIST_VARIABLES"); } break;
		case ANSL_SERVER_CLIENT_CONNECT: { strcpy(strValue, "SERVER_CLIENT_CONNECT"); } break;
		case ANSL_SERVER_CLIENT_DISCONNECT: { strcpy(strValue, "SERVER_CLIENT_DISCONNECT"); } break;
		case ANSL_DATA_CHANGED: { strcpy(strValue, "DATA_CHANGED"); } break;
		case ANSL_CANCEL_DATA_CHANGED: { strcpy(strValue, "CANCEL_DATA_CHANGED"); } break;
		case ANSL_EVENT_DATA: { strcpy(strValue, "EVENT_DATA"); } break;
		case ANSL_CANCEL_EVENT_DATA: { strcpy(strValue, "CANCEL_EVENT_DATA"); } break;
		case ANSL_REQ_READ_DATA: { strcpy(strValue, "REQ_READ_DATA"); } break;
		case ANSL_RESP_READ_DATA: { strcpy(strValue, "RESP_READ_DATA"); } break;
		case ANSL_CANCEL_READ_DATA: { strcpy(strValue, "CANCEL_READ_DATA"); } break;
		case ANSL_REQ_WRITE_DATA: { strcpy(strValue, "REQ_WRITE_DATA"); } break;
		case ANSL_RESP_WRITE_DATA: { strcpy(strValue, "RESP_WRITE_DATA"); } break;
		case ANSL_CANCEL_WRITE_DATA: { strcpy(strValue, "CANCEL_WRITE_DATA"); } break;
		case ANSL_REQU_LIST_DATA_OBJECTS: { strcpy(strValue, "REQU_LIST_DATA_OBJECTS"); } break;
		case ANSL_RESP_LIST_DATA_OBJECTS: { strcpy(strValue, "RESP_LIST_DATA_OBJECTS"); } break;
		case ANSL_CANCEL_LIST_DATA_OBJECTS: { strcpy(strValue, "CANCEL_LIST_DATA_OBJECTS"); } break;
		case ANSL_CLIENT_DATA_OBJ_CONNECT: { strcpy(strValue, "CLIENT_DATA_OBJ_CONNECT"); } break;
		case ANSL_CLIENT_DATA_OBJ_DISCONNECT: { strcpy(strValue, "CLIENT_DATA_OBJ_DISCONNECT"); } break;
		case ANSL_REQ_RESET: { strcpy(strValue, "REQ_RESET"); } break;
		case ANSL_RESP_RESET: { strcpy(strValue, "RESP_RESET"); } break;
		case ANSL_CANCEL_RESET: { strcpy(strValue, "CANCEL_RESET"); } break;
		case ANSL_REQ_CLEAR_MEMORY: { strcpy(strValue, "REQ_CLEAR_MEMORY"); } break;
		case ANSL_RESP_CLEAR_MEMORY: { strcpy(strValue, "RESP_CLEAR_MEMORY"); } break;
		case ANSL_CANCEL_CLEAR_MEMORY: { strcpy(strValue, "CANCEL_CLEAR_MEMORY"); } break;
		case ANSL_REQ_RESUME: { strcpy(strValue, "REQ_RESUME"); } break;
		case ANSL_RESP_RESUME: { strcpy(strValue, "RESP_RESUME"); } break;
		case ANSL_CANCEL_RESUME: { strcpy(strValue, "CANCEL_RESUME"); } break;
		case ANSL_REQ_SUSPEND: { strcpy(strValue, "REQ_SUSPEND"); } break;
		case ANSL_RESP_SUSPEND: { strcpy(strValue, "RESP_SUSPEND"); } break;
		case ANSL_CANCEL_SUSPEND: { strcpy(strValue, "CANCEL_SUSPEND"); } break;
		case ANSL_REQ_CLEAR_LOGBUFFER: { strcpy(strValue, "REQ_CLEAR_LOGBUFFER"); } break;
		case ANSL_RESP_CLEAR_LOGBUFFER: { strcpy(strValue, "RESP_CLEAR_LOGBUFFER"); } break;
		case ANSL_CANCEL_CLEAR_LOGBUFFER: { strcpy(strValue, "CANCEL_CLEAR_LOGBUFFER"); } break;
		case ANSL_REQ_BITMASK: { strcpy(strValue, "REQ_BITMASK"); } break;
		case ANSL_RESP_BITMASK: { strcpy(strValue, "RESP_BITMASK"); } break;
		case ANSL_CANCEL_BITMASK: { strcpy(strValue, "CANCEL_BITMASK"); } break;
		case ANSL_PROCEEDING_LIST_VARIABLES: { strcpy(strValue, "PROCEEDING_LIST_VARIABLES"); } break;
		case ANSL_EVENT_MODULE_CHANGED: { strcpy(strValue, "EVENT_MODULE_CHANGED"); } break;
		case ANSL_CANCEL_MODULE_CHANGED: { strcpy(strValue, "CANCEL_MODULE_CHANGED"); } break;
		case ANSL_SRV_REGISTER_OBJECT: { strcpy(strValue, "SRV_REGISTER_OBJECT"); } break;
		case ANSL_SRV_OBJECT_REGISTERED: { strcpy(strValue, "SRV_OBJECT_REGISTERED"); } break;
		case ANSL_SRV_REG_OBJECT_FINAL: { strcpy(strValue, "SRV_REG_OBJECT_FINAL"); } break;
		case ANSL_SRV_DEREGISTER_OBJECT: { strcpy(strValue, "SRV_DEREGISTER_OBJECT"); } break;
		case ANSL_SRV_OBJECT_DEREGISTER: { strcpy(strValue, "SRV_OBJECT_DEREGISTER"); } break;
		case ANSL_SRV_TELEGRAM_SEND: { strcpy(strValue, "SRV_TELEGRAM_SEND"); } break;
		case ANSL_SRV_TELEGRAM_SENT: { strcpy(strValue, "SRV_TELEGRAM_SENT"); } break;
		case ANSL_SRV_STREAM_SEND: { strcpy(strValue, "SRV_STREAM_SEND"); } break;
//		case ANSL_SRV_STREAM_SENT: { strcpy(strValue, "SRV_STREAM_SENT"); } break;
		case ANSL_SRV_TELEGRAM_RECEIVE: { strcpy(strValue, "SRV_TELEGRAM_RECEIVE"); } break;
		case ANSL_SRV_TELEGRAM_RECEIVED: { strcpy(strValue, "SRV_TELEGRAM_RECEIVED"); } break;
		case ANSL_SRV_STREAM_RECEIVE: { strcpy(strValue, "SRV_STREAM_RECEIVE"); } break;
		case ANSL_SRV_STREAM_RECEIVED: { strcpy(strValue, "SRV_STREAM_RECEIVED"); } break;
		case ANSL_REQ_CONTROL: { strcpy(strValue, "REQ_CONTROL"); } break;
		case ANSL_RESP_CONTROL: { strcpy(strValue, "RESP_CONTROL"); } break;
		case ANSL_CANCEL_CONTROL: { strcpy(strValue, "CANCEL_CONTROL"); } break;
		case ANSL_CANCEL_EVENT_FORMAT: { strcpy(strValue, "CANCEL_EVENT_FORMAT"); } break;
		case ANSL_PROCEEDING_EVENT: { strcpy(strValue, "PROCEEDING_EVENT"); } break;
		case ANSL_REQ_ASSIGN_VALUE: { strcpy(strValue, "REQ_ASSIGN_VALUE"); } break;
		case ANSL_EVENT_TASKCLASSES_CHANGED: { strcpy(strValue, "EVENT_TASKCLASSES_CHANGED"); } break;
		case ANSL_CANCEL_TASKCLASSES_CHANGED: { strcpy(strValue, "CANCEL_TASKCLASSES_CHANGED"); } break;
		case ANSL_CB_TYPES_END: { strcpy(strValue, "CB_TYPES_END"); } break;
		case ANSL_PROCEEDING_DOWNLOAD: { strcpy(strValue, "PROCEEDING_DOWNLOAD"); } break;
		case ANSL_PROCEEDING_UPLOAD: { strcpy(strValue, "PROCEEDING_UPLOAD"); } break;
		case ANSL_INITIALIZE: { strcpy(strValue, "INITIALIZE"); } break;
		case ANSL_DEINITIALIZE: { strcpy(strValue, "DEINITIALIZE"); } break;
		case ANSL_CONNECTION_STATES: { strcpy(strValue, "CONNECTION_STATES"); } break;
		case ANSL_CONNECTION_ROUTER_ERROR: { strcpy(strValue, "CONNECTION_ROUTER_ERROR"); } break;
		case ANSL_CONNECTION_LISTEN_ERROR: { strcpy(strValue, "CONNECTION_LISTEN_ERROR"); } break;
		case ANSL_CONNECTION_LISTEN_CLOSE: { strcpy(strValue, "CONNECTION_LISTEN_CLOSE"); } break;
		case ANSL_CONNECTION_READY: { strcpy(strValue, "CONNECTION_READY"); } break;
		case ANSL_CONNECTION_ERROR: { strcpy(strValue, "CONNECTION_ERROR"); } break;
		case ANSL_CONNECTION_CLOSE: { strcpy(strValue, "CONNECTION_CLOSE"); } break;
		case ANSL_SERVER_START: { strcpy(strValue, "SERVER_START"); } break;
		case ANSL_SERVER_STARTED: { strcpy(strValue, "SERVER_STARTED"); } break;
		case ANSL_SERVER_STOP: { strcpy(strValue, "SERVER_STOP"); } break;
		case ANSL_SERVER_STOPPED: { strcpy(strValue, "SERVER_STOPPED"); } break;
		case ANSL_SERVER_ERROR: { strcpy(strValue, "SERVER_ERROR"); } break;
		case ANSL_ERROR_LISTEN: { strcpy(strValue, "ERROR_LISTEN"); } break;
		case ANSL_LISTEN_CLOSE: { strcpy(strValue, "LISTEN_CLOSE"); } break;
		case ANSL_ERROR_CONN: { strcpy(strValue, "ERROR_CONN"); } break;
		case ANSL_CONN_CLOSE: { strcpy(strValue, "CONN_CLOSE"); } break;
		case ANSL_CONN_READY: { strcpy(strValue, "CONN_READY"); } break;
		case ANSL_SERVER_STATUS: { strcpy(strValue, "SERVER_STATUS"); } break;
		default:
		{
			sprintf_s(strValue, INFO_STR_SIZE-1, "%u", eventCode);
		} break;
	}
	return strlen(strValue);
}

int GetObjTypeString(unsigned int objType, char strValue[INFO_STR_SIZE])
{
	switch (objType)
	{
		case ANSL_OBJTYPE_ROOT: { strcpy(strValue, "ROOT"); } break;
		case ANSL_OBJTYPE_SERVER: { strcpy(strValue, "SERVER"); } break;
		case ANSL_OBJTYPE_CONNECTION: { strcpy(strValue, "CONNECTION"); } break;
		case ANSL_OBJTYPE_CPU: { strcpy(strValue, "CPU"); } break;
		case ANSL_OBJTYPE_CPUPVS: { strcpy(strValue, "CPUPVS"); } break;
		case ANSL_OBJTYPE_CPUPV: { strcpy(strValue, "CPUPV"); } break;
		case ANSL_OBJTYPE_MODULE: { strcpy(strValue, "MODULE"); } break;
		case ANSL_OBJTYPE_LOGGER: { strcpy(strValue, "LOGGER"); } break;
		case ANSL_OBJTYPE_LIBRARY: { strcpy(strValue, "LIBRARY"); } break;
		case ANSL_OBJTYPE_TASK: { strcpy(strValue, "TASK"); } break;
		case ANSL_OBJTYPE_TASKPV: { strcpy(strValue, "TASKPV"); } break;
		case ANSL_OBJTYPE_GLOBAL_LINK_NODES: { strcpy(strValue, "GLOBAL_LINK_NODES"); } break;
		case ANSL_OBJTYPE_LINK_NODE: { strcpy(strValue, "LINK_NODE"); } break;
		case ANSL_OBJTYPE_REGISTERD: { strcpy(strValue, "REGISTERD"); } break;
		case ANSL_OBJTYPE_REGED_STREAM: { strcpy(strValue, "REGED_STREAM"); } break;
		case ANSL_OBJTYPE_REGED_TELEGRAM: { strcpy(strValue, "REGED_TELEGRAM"); } break;
		case ANSL_OBJTYPE_STREAM: { strcpy(strValue, "STREAM"); } break;
		case ANSL_OBJTYPE_TELEGRAM: { strcpy(strValue, "TELEGRAM"); } break;
		case ANSL_OBJTYPE_VARIABLE: { strcpy(strValue, "VARIABLE"); } break;
		default:
		{
			sprintf_s(strValue, INFO_STR_SIZE-1, "%u", objType);
		} break;
	}
	return strlen(strValue);
}

int GetFormatString(ANSL_VARIABLE_DATA_TYPE* pTypeInfo, char strValue[FMT_STR_SIZE])
{
	int dims = 0;
	char strDim[40];
	
	memset(strValue, 0, FMT_STR_SIZE);
	switch (pTypeInfo->m_DataType)
	{
		case ANSL_VAR_TYPE_UINT8: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=ui8 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=ui8 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_UINT16: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=ui16 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=ui16 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_UINT32: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=ui32 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=ui32 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_UINT64: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=ui64 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=ui64 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_INT8: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=i8 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=i8 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_INT16: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=i16 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=i16 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_INT32: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=i32 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=i32 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_INT64: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=i64 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=i64 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_DT: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=DateNTime Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=DateNTime Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_BOOL: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=bool Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=bool Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_FLOAT32: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=f32 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=f32 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_FLOAT64: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=f64 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=f64 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_CHAR_STRING: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=string Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=string Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_WCHAR_STRING: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=wstring Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=wstring Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_STRUCT: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=struct Len=%d ArDim=%d Size=%d Name=%s ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size, pTypeInfo->m_Name);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=struct Len=%d Name=%s", pTypeInfo->m_ElementLength, pTypeInfo->m_Name);
			}
		} break;
		default: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=UNKNOWN Len=%d ArDim=%d Size=%d Name=%s ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size, pTypeInfo->m_Name);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=UNKNOWN Len=%d Name=%s", pTypeInfo->m_ElementLength, pTypeInfo->m_Name);
			}
		} break;
	}
	return strlen(strValue);
}

char* DataToString(char *pData, unsigned int dataLen, ANSL_VARIABLE_DATA_TYPE *pDataType)
{
	char strTmp[256];
	
	int i32V;
	unsigned int ui32V;
	short i16V;
	unsigned short ui16V;
	char i8V;
	unsigned char ui8V;
	double f64V = 0.0;
	float f32V;
	unsigned int xCnt, yCnt, xOffset;


	if ((NULL != pData) && (0 < dataLen))
	{
		switch (pDataType->m_DataType)
		{
			case ANSL_VAR_TYPE_BOOL:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&i8V, pData, 1);
					sprintf(_strValue, "%s", ((0 != i8V) ? "true" : "false"));
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&i8V, pData+xOffset, 1);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%s;",  ((0 != i8V) ? "true" : "false"));
							sprintf(_strValue, "%s%s", _strValue, strTmp);
							xOffset++;
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_INT8:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&i8V, pData, 1);
					sprintf(_strValue, "%d",  i8V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&i8V, pData+xOffset, 1);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%d;",  i8V);
							xOffset+=1;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_INT16:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&i16V, pData, 2);
					sprintf(_strValue, "%d",  i16V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&i16V, pData+xOffset, 2);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%d;",  i16V);
							xOffset+=2;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_INT32:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&i32V, pData, 4);
					sprintf(_strValue, "%d",  i32V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&i32V, pData+xOffset, 4);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%d;",  i32V);
							xOffset+=4;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_UINT8:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&ui8V, pData, 1);
					sprintf(_strValue, "%u",  ui8V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&ui8V, pData+xOffset, 1);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%u;",  ui8V);
							xOffset+=1;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_UINT16:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&ui16V, pData, 2);
					sprintf(_strValue, "%u",  ui16V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&ui16V, pData+xOffset, 2);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%u;",  ui16V);
							xOffset+=2;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_UINT32:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&ui32V, pData, 4);
					sprintf(_strValue, "%u",  ui32V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&ui32V, pData+xOffset, 4);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%u;",  ui32V);
							xOffset+=4;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_FLOAT32:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&f32V, pData, 4);
					sprintf(_strValue, "%f",  f32V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&f32V, pData+xOffset, 4);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%f;",  f32V);
							xOffset+=4;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_FLOAT64:
			{
				//if (0 == pDataType->m_ArrayDimensions)
				//{
					memcpy(&f64V, pData, 8);
					sprintf(_strValue, "%g",  f64V);
				/*} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&f64V, pData+xOffset, 8);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%g;",  f64V);
							xOffset+=8;
							sprintf(_strValue, "%s%s", _strValue, strTmp;
						}
					}
				}
				*/
			} break;
			case ANSL_VAR_TYPE_CHAR_STRING:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					sprintf(_strValue, "%s", pData);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%s;", (char*)(pData+xOffset));
							sprintf(_strValue, "%s%s", _strValue, strTmp);
							xOffset+=pDataType->m_ElementLength;
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_STRUCT:
			{
				return "";
			} break;
			case ANSL_VAR_TYPE_DT:
			{
				CPU_DATETIME_Type* pTmpDT;
				if (0 == pDataType->m_ArrayDimensions)
				{
					pTmpDT = (CPU_DATETIME_Type*)pData;
					sprintf(_strValue, 
							"%04d/%02d/%02d %02d:%02d:%02d",
							(pTmpDT->Year + 1900),
							pTmpDT->Mon,
							pTmpDT->MDay,
							pTmpDT->Hour,
							pTmpDT->Min,
							pTmpDT->Sec);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							pTmpDT = (CPU_DATETIME_Type*)(pData+xOffset);
							memset(strTmp, 0, 256);
							sprintf(strTmp, 
									"%04d/%02d/%02d %02d:%02d:%02d",
									pTmpDT->Year + 1900,
									pTmpDT->Mon,
									pTmpDT->MDay,
									pTmpDT->Hour,
									pTmpDT->Min,
									pTmpDT->Sec);
							sprintf(_strValue, "%s%s", _strValue, strTmp);
							xOffset+=pDataType->m_ElementLength;
						}
					}
				}
			} break;
			default:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					switch (pDataType->m_ElementLength)
					{
						case 1:
						{
							memcpy(&i8V, pData, 1);
							i32V = i8V;
						} break;
						case 2:
						{
							memcpy(&i16V, pData, 2);
							i32V = i16V;
						} break;
						case 4:
						{
							memcpy(&i32V, pData, 4);
						} break;
					}
					sprintf(_strValue, "%d",  i32V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							switch (pDataType->m_ElementLength)
							{
								case 1:
								{
									memcpy(&i8V, pData+xOffset, 1);
									i32V = i8V;
									xOffset+=1;
								} break;
								case 2:
								{
									memcpy(&i16V, pData+xOffset, 2);
									i32V = i16V;
									xOffset+=2;
								} break;
								case 4:
								{
									memcpy(&i32V, pData+xOffset, 4);
									xOffset+=4;
								} break;
							}
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%d;", i32V);
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
		}
	} else {
		sprintf(_strValue, "INVALID DATA");
	}
	return _strValue;
}

void AnslGetVersionInfo()
{
	unsigned int status = 0;
	unsigned int uifVersion = 0;
	unsigned int oifVersion = 0;

	status = ANSLI_GetVersionInfos(&uifVersion, &oifVersion);
	if (status != 0)
	{
		printf("ERROR ANSLI_GetVersionInfos : %i\n",status);
	} else {
		printf("ANSLI_GetVersionInfos: uifV=%d.%d  oifV=%d\n", (uifVersion/100), (uifVersion%100), oifVersion);
	}
}

void TaskClassResume(unsigned int cpuIdent, unsigned short tcNumber)
{
	unsigned int status = 0;

	status = ANSLI_CPU_ResumeTaskClass(cpuIdent, tcNumber);
	if (status != 0)
	{
		printf("ERROR RESUME TC : %i\n",status);
	}
}

void TaskClassSuspend(unsigned int cpuIdent, unsigned short tcNumber)
{
	unsigned int status = 0;

	status = ANSLI_CPU_SuspendTaskClass(cpuIdent, tcNumber);
	if (status != 0)
	{
		printf("ERROR SUSPEND TC : %i\n",status);
	}
}

void TaskResume(unsigned int taskIdent, unsigned short numOfCycles)
{
	unsigned int status = 0;

	status = ANSLI_TASK_Resume(taskIdent, numOfCycles);
	if (status != 0)
	{
		printf("ERROR RESUME TASK : %i\n",status);
	}
}

void TaskSuspend(unsigned int taskIdent, unsigned short numOfCycles)
{
	unsigned int status = 0;

	status = ANSLI_TASK_Suspend(taskIdent, numOfCycles);
	if (status != 0)
	{
		printf("ERROR SUSPEND TASK : %i\n",status);
	}
}

void LoggerClearBuffer(unsigned int loggerIdent)
{
	unsigned int status = 0;

	status = ANSLI_LOGGER_ClearBuffer(loggerIdent);
	if (status != 0)
	{
		printf("ERROR CLEAR BUFFER  LOGGER: %i\n",status);
	}
}

void CpuClearMemory(unsigned int cpuIdent, ANSL_CPU_CLEARMEM_MODES memType)
{
	unsigned int status = 0;

	status = ANSLI_CPU_ClearMemory(cpuIdent, memType);
	if (status != 0)
	{
		printf("ERROR CLEAR MEM  CPU : %i\n",status);
	}
}

