/*
 ============================================================================
 Name        : ANSLSample.c
 Author      : Werner Hager
 Version     : 3
 Copyright   : B&R
 Description : ANSL sample
 ============================================================================
 */

/***************************************************************************************************/
/** INCLUDES ***************************************************************************************/

#ifdef WIN32
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include "..\..\..\..\Exp\ANSLUIF.h"	/* ANSL USER INTERFACE */
#include <Windows.h>
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ANSLUIF.h>	/* ANSL USER INTERFACE */
#include <AnslError.h>
#include <termios.h>
#include <unistd.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/timeb.h>
#endif

#ifndef sprintf_s
#define sprintf_s(frmt, size, args...) 			sprintf(frmt, ##args)
#endif

/***************************************************************************************************/
/** DEFINES ****************************************************************************************/

#define TARGET_IP 					"10.86.13.220"	/* IP of the target */
#define CONNECTION_PARAMS			"Timeout=\"35000\" Port=\"11169\" SendDelay=\"1\" Buffer=\"4096\""	/* connection parameters */
#define CONNECTION_PARAMS_NO_TO		"Port=\"11169\" SendDelay=\"1\" Buffer=\"4096\""	/* connection parameters */
#define CONNECTION_PARAMS_NO_TOPO	"SendDelay=\"1\" Buffer=\"4096\""	/* connection parameters */
#define TASK_NAME 					"pvitest"		/* task name of the task on target */
#define VAR_NAME_READ 				"PV1"			/* local UDINT variable */
#define VAR_NAME_SET 				"PV2"			/* local UDINT variable */
#define VAR_NAME_SET_VALUE	        20				/* VAR_NAME_SET gets this value */
/*my*/
#define VAR_NAME_READ1 				"PviVar.PVBool01"			/* local UDINT variable */
#define VAR_NAME_READ2				"PviVar.PVDint01"
#define VAR_NAME_READ3				"PviVar.PVInt01"
#define VAR_NAME_READ4				"PviVar.PVSint01"
#define VAR_NAME_READ5				"PviVar.PVUdint01"
#define VAR_NAME_READ6				"PviVar.PVUint01"
#define VAR_NAME_READ7				"PviVar.PVUsint01"
#define VAR_NAME_READ8				"PviVar.PVReal01"
#define VAR_NO                     		 8
#define rfPV1					1000	
#define GVAR_NAME_READ1             		 "gPviVar.PVBool01"
#define GVAR_NAME_READ2             		 "gPviVar.PVDint01"
#define GVAR_NAME_READ3             		 "gPviVar.PVInt01"
#define GVAR_NAME_READ4             		 "gPviVar.PVSint01"
#define GVAR_NAME_READ5             		 "gPviVar.PVUdint01"
#define GVAR_NAME_READ6             		 "gPviVar.PVUint01"
#define GVAR_NAME_READ7             		 "gPviVar.PVUsint01"
#define GVAR_NAME_READ8             		 "gPviVar.PVReal01"
#define GVAR_NAME_READ9             		 "gPV"
#define GVAR_NAME_READ10             		 "gPV1"
#define GVAR_NAME_READ11             		 "gPV2"
#define GVAR_NAME_READ12             		 "gPV3"
#define GVAR_NAME_READ13             		 "gPV4"
#define GVAR_NAME_READ14             		 "gPV5"
#define GVAR_NAME_READ15             		 "gPV6"
#define GVAR_NAME_READ16             		 "gPV7"
#define GVAR_NAME_READ17             		 "gPV8"
#define GVAR_NAME_READ18             		 "gPV9"
#define GVAR_NAME_READ19             		 "gPV10"
#define GVAR_NAME_READ20             		 "gPV11"
#define GVAR_NAME_READ21             		 "gPV12"
#define GVAR_NAME_READ22             		 "gPV13"
#define GVAR_NAME_READ23             		 "gPV14"
#define GVAR_NAME_READ24            		 "gPV15"
#define GVAR_NAME_READ25             		 "gPV16"
#define GVAR_NAME_READ26             		 "gPV17"
#define GVAR_NAME_READ27             		 "gPV18"
#define GVAR_NAME_READ28            		 "gPV19"
#define GVAR_NAME_READ29             		 "gPV20"


#define VAR_TOTAL					16


#define RC_LOCK_ERROR              1
#define RC_OBJECT_NOT_CREATED      2
#define RC_STACKSIZE_ERROR         3
#define RC_THREAD_ATTR_ERROR       4
#define RC_THREAD_NOT_CREATED      5
#define RC_UNLOCK_ERROR            6
#define RC_NOT_OWNER               7

#define FMT_STR_SIZE			   1024
#define INFO_STR_SIZE			   256

#pragma pack(1)
/* read/write cpu date and time (request/response): */
typedef struct
{
    unsigned char		Sec;			/* seconds after the minute - [0,59] */
	unsigned char		Min;			/* minutes after the hour - [0,59] */
	unsigned char		Hour;			/* hours since midnight - [0,23] */
	unsigned char		MDay;			/* day of the month - [1,31] */
	unsigned char		Mon;			/* months since January - [0,11] */
    unsigned char		Year;			/* years since 1900 */
    unsigned char		WDay;			/* days since Sunday - [0,6] */
    unsigned char		IsDst;			/* daylight savings time flag */
    unsigned short		YDay;			/* days since January 1 - [0,365] */
}	CPU_DATETIME_Type;
#pragma pack()


/***************************************************************************************************/
/** GLOBAL VARIABLES  ******************************************************************************/


/*my*/
int varNo = 0;
_Bool var2 = 0; 
int var3 = 0; 
unsigned long var4 = 0;
float var5 = 0 ;

/* for synchronization */

//int var1,var2,var3,var4 = 0;
ANSL_VARIABLE_EVENT_PARAMETER evParam;
unsigned int identCpu380 = 0;				/* ident of connection to cpu */
unsigned int identTask = 0;					/* ident of connection to task1 */
unsigned int identLogger = 0;			    /* ident of connection to $arlogsys */
unsigned int identVarRead = 0;				/* ident of connection to var1 */
unsigned int identVarSet = 0;				/* ident of connection to var2 */
unsigned short varSet = VAR_NAME_SET_VALUE;	/* value to set on target */
unsigned int status = 0;					/* state of function calls */
unsigned int userDataCpu = 380;			/* user date to identify the caller */
unsigned int userDataTask = 381;			/* user date to identify the caller */
unsigned int userDataLogger = 400;			/* user date to identify the caller */
unsigned int userDataRead = 1;				/* user date to identify the caller */
unsigned int userDataSet = 2;				/* user date to identify the caller */

#ifdef WIN32
#define T_PTHREAD void*
#define T_PTHREAD_MUTEX int
#else
#define T_PTHREAD pthread_t
#define T_PTHREAD_MUTEX pthread_mutex_t
pthread_attr_t _threadAttr;
#endif

#define T_PTHREADP T_PTHREAD*
#define T_PTHREAD_MUTEXP T_PTHREAD_MUTEX*

#ifdef WIN32
int pthread_mutex_lock(T_PTHREAD_MUTEXP pMutex) { return 0; };
int pthread_mutex_unlock(T_PTHREAD_MUTEXP pMutex) { return 0; };

__declspec(dllimport) void* __stdcall CreateThread(LPSECURITY_ATTRIBUTES lpThreadAttributes, 
												  unsigned long dwStackSize, 
												  LPTHREAD_START_ROUTINE lpStartAddress, 
												  void* lpParameter, 
												  unsigned long dwCreationFlags, 
												  unsigned long* lpThreadId);
__declspec(dllimport) unsigned long __stdcall WaitForSingleObject(void* hHandle, unsigned long dwMilliseconds);
__declspec(dllimport) int __stdcall CloseHandle(void* hHandle);
#endif

T_PTHREAD _ThreadId;
T_PTHREAD_MUTEX _MutexObject;
int _bBreakThread;
int _CPUConnClosed;
int _bConnectionClosed;
int _RunningState;
char _strObj[10];
char _strPV1[100];
char _strPV2[100];
char _strTask[100];
char _strValue[1024];
ANSL_VARIABLE_DATA_TYPE _DataTypePV1;
ANSL_VARIABLE_DATA_TYPE _DataTypePV2;
char strIP[16]; 		/* IP-Adresse */
char strConPara[256]; 	/* Connection parameters */
unsigned int pvParentIdent;
//int rfPV1 = 2000;
int rfPV2;

/*my*/

unsigned int myIdentVarRead[VAR_TOTAL];
char _myStrFormat[VAR_TOTAL][FMT_STR_SIZE];
ANSL_VARIABLE_DATA_TYPE _DataTypePV[200];
char *_strPV[400] = { VAR_NAME_READ1, VAR_NAME_READ2, VAR_NAME_READ3, VAR_NAME_READ4, VAR_NAME_READ5, VAR_NAME_READ6, VAR_NAME_READ7, VAR_NAME_READ8,
GVAR_NAME_READ1,GVAR_NAME_READ2,GVAR_NAME_READ3,GVAR_NAME_READ4,GVAR_NAME_READ5,GVAR_NAME_READ6,GVAR_NAME_READ7,GVAR_NAME_READ8,GVAR_NAME_READ9,GVAR_NAME_READ10,
GVAR_NAME_READ11,GVAR_NAME_READ12,GVAR_NAME_READ13,GVAR_NAME_READ14,GVAR_NAME_READ15,GVAR_NAME_READ16,GVAR_NAME_READ17,GVAR_NAME_READ18,GVAR_NAME_READ19,GVAR_NAME_READ20,
GVAR_NAME_READ21,GVAR_NAME_READ22,GVAR_NAME_READ23,GVAR_NAME_READ24,GVAR_NAME_READ25,GVAR_NAME_READ26,GVAR_NAME_READ27,GVAR_NAME_READ28,GVAR_NAME_READ29};
unsigned int mypvParentIdent;
unsigned int mygpvParentIdent;
int i = 0;
void delay(unsigned int ms);
unsigned int WriteVariableInt(unsigned int varIdent, long varValue);
unsigned int WriteVariableReal(unsigned int varIdent, float varValue);
unsigned int ANSLCB_all(unsigned int errorCode, 
						  unsigned int ident, 
						  unsigned int objectType, 
						  unsigned int cbType, 
						  void* pData, 
						  unsigned int dataLen, 
						  void* pUserData);


unsigned long GetSystemTickCount(); // Zeit seit dem Booten in Sekunden
unsigned long GetSystemTickCountA(); // Zeit seit dem Booten in Sekunden
unsigned long GetSystemTickCountB(); // Zeit seit dem Booten in Sekunden
char* DataToString(char *pData, unsigned int dataLen, ANSL_VARIABLE_DATA_TYPE *pDataType);
int GetFormatString(ANSL_VARIABLE_DATA_TYPE* pTypeInfo, char strValue[FMT_STR_SIZE]);
int GetObjTypeString(unsigned int objType, char strValue[INFO_STR_SIZE]);
int GetCBTypeString(unsigned int cbType, char strValue[INFO_STR_SIZE]);
unsigned int ReadCpuDateTime(unsigned int cpuIdent);
unsigned int ReadVariable(unsigned int varIdent);
unsigned int WriteVariable(unsigned int varIdent, long varValue);
int connectCPU(char* strIP, char* strConPara, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData);
int connectTask(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData);
int connectLogger(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData);
int connectVariable(unsigned int parentID, char* name, unsigned int* ident, int flags, int refresh, ANSL_CALL_BACK fpCB, void* userData);
void AnslGetVersionInfo();
void TaskClassResume(unsigned int cpuIdent, unsigned short tcNumber);
void TaskClassSuspend(unsigned int cpuIdent, unsigned short tcNumber);
void TaskResume(unsigned int taskIdent, unsigned short numOfCycles);
void TaskSuspend(unsigned int taskIdent, unsigned short numOfCycles);
void LoggerClearBuffer(unsigned int loggerIdent);
void CpuClearMemory(unsigned int cpuIdent, ANSL_CPU_CLEARMEM_MODES memType);


int ioGetch();
unsigned int ConnectPV(char *name, unsigned int parentIdent, unsigned int evFlags, unsigned int refresh, void *pUserData, unsigned int *pIdent);
void DoWork();
void StartNoThread();

#ifdef _WORKERTHREAD
int CreateWorkerThread(int* pRunningState, T_PTHREAD_MUTEX* pMutex, T_PTHREAD* pThreadId, void* fpThread);
#ifdef WIN32
static unsigned long __stdcall WorkerThread (void* pParam);
void pthread_join(T_PTHREAD threadId, void* pAttribs) { return; };
#else
static void* WorkerThread(void* pParam);
#endif
#endif


int ErrorCheck(unsigned int status, const char* error);
void DoExit();
int UninitializeObjects();

void PrintVInfo() 
{
	unsigned int uifVersion = 0;
	unsigned int oifVersion = 0;
	int retVal = 0;
    int verMaj = 0;
    int verMin = 0;
    int verBuild = 0;

	retVal = ANSLI_GetVersionInfos(&uifVersion, &oifVersion);
    verMaj = uifVersion / 10000;
    verMin = (uifVersion - verMaj * 10000) / 100;
    verBuild = (uifVersion - verMaj * 10000) - verMin * 100;
	printf("ANSLI VersionInfo:  ANSLUIF: V%d.%d.%d   ANSLONLINE: V%d\n", verMaj, verMin, verBuild, oifVersion);
}

/***************************************************************************************************/
/** MAIN PROGRAM ***********************************************************************************/

int main(int argc,char *argv[]) 
{
	int logLevel = 0;
	unsigned long t1, t2, tH, tM, tS, tMS;
	t1 = GetSystemTickCount();
	printf("SystemTickCount=%lu\n", t1);
	
	_bBreakThread = 0;
	_RunningState = 0;
	_CPUConnClosed = 0;
	_bConnectionClosed = 0;
	memset(strIP, 0, 16); 		
	memset(strConPara, 0, 256);

	/** begin: init and connect **/
	
	/* init ANSL */
	sprintf(strIP, "%s", TARGET_IP);
	sprintf(strConPara, "%s", CONNECTION_PARAMS);
	sprintf(_strTask, "%s", TASK_NAME);

	printf("+ANSL Sample   IPAddress=\"%s\" %s \n\n", strIP, strConPara);
	printf("Press r to refresh all the variables!\n");
	printf("Press s to set the specific variable!\n");
	printf("Press l to list the variable!\n");
	printf("Press w for WarmRestart!\n");
	printf("Press c for ColdRestart!\n");
	printf("Press e to stop ANSL communication!\n\n");
	status = ANSLI_Initialize();
	if (ErrorCheck(status, "ERROR ANSLI_Initialize") != 0) return -1;


	/* connect to CPU object */
	status = ANSLI_CPU_Connect(strIP, strConPara, (unsigned int*)&identCpu380, (ANSL_CALL_BACK)ANSLCB_all, (void*)&userDataCpu);
	if (ErrorCheck(status, "ERROR ANSLI_CPU_Connect") != 0) return -1;
	mygpvParentIdent = identCpu380;
	
	/* connect to task object */

	status = ANSLI_TASK_Connect(identCpu380, _strTask, (unsigned int*)&identTask, (ANSL_CALL_BACK)ANSLCB_all, (void*)&userDataTask);
	if (ErrorCheck(status, "ERROR ANSLI_TASK_Connect TASK_NAME") != 0) return -1;
	mypvParentIdent = identTask;
	t1 = GetSystemTickCount();
	/*connect to variable object*/
	
	for(i = 0; i<VAR_NO; i++)
	{
	  printf("No.(%d)  ",i);
	  userDataRead = i;
	  connectVariable(mypvParentIdent, _strPV[i], (unsigned int*)&myIdentVarRead[i], ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV1, (ANSL_CALL_BACK)ANSLCB_all, (void*)&userDataRead);
	  delay(60);
	 
	  
	}
	for(i = VAR_NO; i<VAR_TOTAL; i++)
	{
	  printf("No.(%d)  ",i);
	  userDataRead = i;
	  connectVariable(mygpvParentIdent, _strPV[i], (unsigned int*)&myIdentVarRead[i], ANSL_VAR_EV_FLAG_FORMAT | ANSL_VAR_EV_FLAG_DATA, rfPV1, (ANSL_CALL_BACK)ANSLCB_all, (void*)&userDataRead);
	  delay(60);	  
	}

	printf("\nPress 's' to set specific variables!\n");
	
//#endif // !#ifdef _NO_AUTO_CONNECT

//#ifdef _WORKERTHREAD
	// either start a new thread
	//CreateWorkerThread(&_RunningState, &_MutexObject, &_ThreadId, WorkerThread);
//#else
	// or do the work in a loop until 0 pressed
	StartNoThread();
//#endif

//#ifdef _WORKERTHREAD
//	pthread_join(_ThreadId, NULL);
//#endif
/** begin: deinit and disconnect **/
	
	

	t2 = GetSystemTickCount();
	tS = ((t2-t1)/1000);
	tMS = ((t2-t1)%1000);
	tM = tS/60;
	tS = tS%60;
	tH = tM/60;
	tM = tM%60;
	printf("-ANSLSample    Execution time: %luh %lum %lus %lums\n", tH, tM, tS, tMS);

	/** begin: deinit and disconnect **/
	return EXIT_SUCCESS;
}

/* connect to CPU object */
int connectCPU(char* strIP, char* strConPara, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData)
{
	int status = 0;
	status = ANSLI_CPU_Connect(strIP, strConPara, (unsigned int*)&identCpu380, (ANSL_CALL_BACK)ANSLCB_all, (void*)&userDataCpu);
	if (ErrorCheck(status, "ERROR ANSLI_CPU_Connect") != 0) return -1;
	return status;
}

/* connect to task object */
int connectTask(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData)
{
	int status = 0;
	if (0 != name[0])
	{
		status = ANSLI_TASK_Connect(parentID, name, ident, (ANSL_CALL_BACK)fpCB, userData);
		ErrorCheck(status, "ERROR ANSLI_TASK_Connect TASK_NAME");
	}
	return status;
}

/* connect to task object */
int connectLogger(unsigned int parentID, char* name, unsigned int* ident, ANSL_CALL_BACK fpCB, void* userData)
{
	int status = 0;
	if (0 != name[0])
	{
		status = ANSLI_LOGGER_Connect(parentID, name, ident, ANSL_LOGM_EV_FLAG_NONE, (ANSL_CALL_BACK)fpCB, userData);
		ErrorCheck(status, "ERROR ANSLI_LOG_Connect LOGGER_NAME");
	}
	return status;
}

/* connect to variable object */
int connectVariable(unsigned int parentID, char* name, unsigned int* ident, int flags, int refresh, ANSL_CALL_BACK fpCB, void* userData)
{
	int status = 0;
	ConnectPV(name, parentID, flags, refresh, userData, ident);
	status = ANSLI_VAR_ReadDataType(*ident);
	ErrorCheck(status, "ERROR ANSLI_VAR_ReadDataType VAR_NAME_READ");
	return status;
}


unsigned long GetSystemTickCount() // Zeit seit dem Booten in Sekunden
{
#ifdef WIN32
	return GetTickCount();
#else
	struct timeval tv;
	gettimeofday(&tv,NULL);
	return (unsigned long)(tv.tv_sec*1000 + (tv.tv_usec / 1000));
#endif
}

unsigned long GetSystemTickCountA() // Zeit seit dem Booten in Sekunden
{
#ifdef WIN32
	return GetTickCount();
#else
	struct timespec abstime;
	clock_gettime(CLOCK_REALTIME, &abstime);
	return (unsigned long)(abstime.tv_sec*1000 + (abstime.tv_nsec / 1000));
#endif
}

unsigned long GetSystemTickCountB() // Zeit seit dem Booten in Sekunden
{
#ifdef WIN32
	return GetTickCount();
#else
	struct timeb tp;
	ftime (&tp);
	return (unsigned long)(tp.time*1000 + tp.millitm);
#endif
}

#ifdef _WORKERTHREAD
int CreateWorkerThread(int* pRunningState, T_PTHREAD_MUTEX* pMutex, T_PTHREAD* pThreadId, void* fpThread)
{
	int retVal = 0; 
	*pRunningState = 0;
#ifdef WIN32
	*pThreadId = CreateThread(NULL, 0, fpThread, NULL, 0, NULL);

	WaitForSingleObject(*pThreadId, 0xFFFFFFFF);
	
	CloseHandle(*pThreadId);
#else

	pthread_mutexattr_t mutex_attr;
	pthread_attr_t	threadAttr;

	/* create event objekt: */
	pthread_mutexattr_init (&mutex_attr);
	retVal = pthread_mutex_init(pMutex, &mutex_attr);
	if(retVal < 0)
	{
		printf( "Mutex NOT created.\n" );
		return RC_OBJECT_NOT_CREATED;
	}

	retVal = pthread_mutex_lock (pMutex);
	if (-1 == retVal)
	{
	   printf( "Mutex LOCK ERROR.\n" );
	   return RC_LOCK_ERROR;
	}

	retVal = pthread_attr_init(&threadAttr);
	if(retVal < 0)
	{
		printf( "pthread_attr_init ERROR.\n" );
		return RC_THREAD_ATTR_ERROR;
	}

	retVal = pthread_attr_setstacksize(&threadAttr, 120*1024);
	if(retVal < 0)
	{
		printf( "pthread_attr_setstacksize ERROR.\n" );
		return RC_STACKSIZE_ERROR;
	}

	retVal = pthread_mutex_unlock (pMutex);
	if (-1 == retVal)
	{
	   printf( "Mutex UNLOCK ERROR.\n" );
	   return RC_UNLOCK_ERROR;
	}

	//printf("pthread_create...\n");
	retVal = pthread_create(pThreadId, NULL/*&threadAttr*/, (void*)fpThread, pMutex);

	if(retVal != 0)
	{
		pthread_mutex_destroy(pMutex);
		printf( "Thread NOT created!\n" );
		return RC_THREAD_NOT_CREATED;
	}
#endif
	return status;
}
#endif

#ifdef _WORKERTHREAD
#ifdef WIN32
static unsigned long __stdcall WorkerThread (void* pParam)
#else
static void* WorkerThread(void* pParam)
#endif
{

	T_PTHREAD_MUTEX* pMutex = (T_PTHREAD_MUTEX*)pParam;

	int status;

	// Nun kann losgeloest von der restlichen Kommunikation
	// mit der Datenaufbereitung begonnen werden
	while (0 == _bBreakThread)
	{
		status = pthread_mutex_lock(pMutex);
		if (0 != status)
		{
			printf("pthread_mutex_lock   RC_LOCK_ERROR!\n");
		}
		DoWork();
		pthread_mutex_unlock(pMutex);
	}
	UninitializeObjects();	
	while (0 == _bConnectionClosed)
	{
		status = pthread_mutex_lock(pMutex);
		if (0 != status)
		{
			printf("pthread_mutex_lock   RC_LOCK_ERROR!\n");
		}
		DoExit();
		pthread_mutex_unlock(pMutex);
	}

	return (0);
}
#endif

void DoWork()
{
	unsigned int status = 0;
	int charRead = 0;
	switch (_RunningState)
	{
		case 0:
		{
			_RunningState = 1;
		} break;
		case 1:
		{
			_RunningState = 2;
		} break;
		case 2:
		{
			charRead = ioGetch();
			switch (charRead)
			{
				case 114://read
				{
					
					printf("\n");
					for(i = 0; i<VAR_NO; i++)
					{
					      printf("No.(%d)  ",i);
					      userDataRead = i;
					      ReadVariable(myIdentVarRead[i]);
					      delay(15);
					} 
					for(i = VAR_NO; i< VAR_TOTAL; i++)
					{
					      printf("No.(%d)  ",i);
					      userDataRead = i;
					      ReadVariable(myIdentVarRead[i]);
					      delay(15);
					} 
					printf("\nPress 's' to set specific parameter!\n"); 
					
				}break;
				
				case 115://set parameters
				{
					printf("Enter No. of the parameter: ");
					scanf(" %d",&varNo);
					userDataRead = varNo;
					ReadVariable(myIdentVarRead[varNo]);
					delay(100);
					//printf(_myStrFormat[varNo]);
					
					if(strcmp(_myStrFormat[varNo], "Type=f32 Len=4")==0)
					{
					  printf("Enter REAL value: ");
					  scanf(" %f",&var5);
					  WriteVariableReal(myIdentVarRead[varNo], var5);
					 
					  
					}else if(strcmp(_myStrFormat[varNo], "Type=bool Len=1")==0)
					{
					  printf("Enter BOOL value(0/1): ");
					  scanf(" %d",&var2);
					  if((var2 >= 0) & (var2 <= 1) )
					  {
					  WriteVariableInt(myIdentVarRead[varNo], var2);
					  
					  }else
					  {
					    printf("Value out of range. Press 's' to try again!\n");
					  }
					    
					}else if(strcmp(_myStrFormat[varNo], "Type=i32 Len=4")==0)  
					{
					  printf("Enter DINT value(-2147483648 ~ 2147483647): ");
					  scanf(" %d",&var3);
					  
					  if((var3 >= -2147483648) ||( (var3>=0) & (var3 <= 2147483647)) )
					  {
					  WriteVariableInt(myIdentVarRead[varNo], var3);
					 
					  }else
					  {
					    printf("Value out of range. Press 's' to try again!\n");
					  }
					  	
					}else if(strcmp(_myStrFormat[varNo], "Type=i16 Len=2")==0)  
					{
					  printf("Enter INT value(-32768 ~ 32767): ");
					  scanf(" %d",&var3);
					  
					  if((var3 >= -32768) & (var3 <=32767) )
					  {
					  WriteVariableInt(myIdentVarRead[varNo], var3); 
					  
					  }else
					  {
					    printf("Value out of range. Press 's' to try again!\n");
					  }
					   	
					}else if(strcmp(_myStrFormat[varNo], "Type=i8 Len=1")==0)  
					{
					  printf("Enter SINT value(-128 ~ 127): ");
					  scanf(" %d",&var3);
					  
					  if((var3 >= -128) & (var3 <=127) )
					  {
					  WriteVariableInt(myIdentVarRead[varNo], var3); 
					  
					  }else
					  {
					    printf("Value out of range. Press 's' to try again!\n");
					  }
					  	
					}else if(strcmp(_myStrFormat[varNo], "Type=ui32 Len=4")==0)  
					{
					  printf("Enter UDINT value: ");
					  scanf(" %u",&var4);
					  
					  if((var4 >= 0) & (var4 <=4294967295) )
					  {
					  WriteVariableInt(myIdentVarRead[varNo], var4); 
					  
					  }else
					  {
					    printf("Value out of range. Press 's' to try again!\n");
					  }
					  	
					}else if(strcmp(_myStrFormat[varNo], "Type=ui16 Len=2")==0)  
					{
					  printf("Enter UINT value(0 ~ 65535): ");
					  scanf(" %u",&var4);
			
					  if((var4 >= 0) & (var4 <=65535) )
					  {
					  WriteVariableInt(myIdentVarRead[varNo], var4); 
					  
					  }else
					  {
					    printf("Value out of range. Press 's' to try again!\n");
					  }
					  
					}else if(strcmp(_myStrFormat[varNo], "Type=ui8 Len=1")==0)  
					{
					  printf("Enter USINT value(0 ~ 255): ");
					  scanf(" %u",&var4);
					  
					  if((var4 >= 0) & (var4 <= 255) )
					  {
					  WriteVariableInt(myIdentVarRead[varNo], var4); 
					  
					  }else
					  {
					    printf("Value out of range. Press 's' to try again!\n");
					  }
					  
					}else
					{
					  printf("Enter value: ");
					  scanf(" %d",&var3);
					  WriteVariableInt(myIdentVarRead[varNo], var3);
					  
					}
					 
					
				}break;
				
				case 108://list
				{
					ANSLI_TASK_ListVariables(mypvParentIdent,0xFFFF);
					
				} break;
				
				case 116://test list struct
				{
					
					
				} break;
				
				case 101://e
				{
					printf("Exiting...\n");
					_RunningState = 3;
				} break;
				case 119://w
				{
					printf("WarmRestart\n");
					status = ANSLI_CPU_Restart(identCpu380, ANSL_CPU_WARM_RESTART);
					ErrorCheck(status, "ANSLI_CPU_Restart");
				} break;
				case 99://c
				{
					printf("ColdRestart\n");
					status = ANSLI_CPU_Restart(identCpu380, ANSL_CPU_COLD_RESTART);
					ErrorCheck(status, "ANSLI_CPU_Restart");
				} break;
			  
			}
		} break;
		case 3:
		{
			_bBreakThread = 1;
		} break;
	}
}

void StartNoThread()
{
	// read date and time of the cpu
	status = ANSLI_CPU_DTRead(identCpu380);
	ErrorCheck(status, "ANSLI_CPU_DTRead");
	delay(60);

	while (0 == _bBreakThread)
	{
		DoWork();
		/*printf("\n");
		for(i = 0; i<VAR_NO; i++)
		{
		    printf("No.(%d)  ",i);
		    userDataRead = i;
		    ReadVariable(myIdentVarRead[i]);
		    delay(1);
		} 
					
		for(i = VAR_NO; i< VAR_TOTAL; i++)
		{
		    printf("No.(%d)  ",i);
		    userDataRead = i;
		    ReadVariable(myIdentVarRead[i]);
		    delay(1);
		} */
					
	}
	UninitializeObjects();	
	while (0 == _bConnectionClosed)
	{
		DoExit();
	}
	/** end: working **/
}

int UninitializeObjects()
{
	int status = 0;
	_bConnectionClosed = 1;
	/* cancel the PV connection */
	
	for(i = 0; i<VAR_NO; i++)
	{
	  userDataRead = i;
	  ANSLI_VAR_Disconnect(myIdentVarRead[i]);
	}

	/* cancel the task connection */
	if (0 != identTask)
	{
		status = ANSLI_TASK_Disconnect(identTask);
		if (ErrorCheck(status, "ERROR ANSLI_TASK_Disconnect") != 0) return -1;
	}

	/* cancel the CPU connection */
	if (0 != identCpu380)
	{
		status = ANSLI_CPU_Disconnect(identCpu380);
		if (ErrorCheck(status, "ERROR ANSLI_CPU_Disconnect") != 0) return -1;
		_bConnectionClosed = 0;
	}

	/* deinit ANSL */
	// Wait for deinit in CPU callback!!

	return 0;
}

void DoExit()
{
	if (1 == _CPUConnClosed)
	{
		ANSLI_DeInitialize();
		_bConnectionClosed = 1;
	}
}

unsigned int ReadCpuDateTime(unsigned int cpuIdent)
{
	unsigned int status = 0;

	status = ANSLI_CPU_DTRead(cpuIdent);
	if (status != 0)
	{
		printf("ERROR READ DT : %i\n",status);
	}
	return status;
}

unsigned int ConnectPV(char *name, unsigned int parentIdent, unsigned int evFlags, unsigned int refresh, void *pUserData, unsigned int *pIdent)
{
	unsigned int status = 0;
	struct ANSL_VARIABLE_EVENT_PARAMETER 	evParam;
	evParam.m_EventFlags = evFlags;
	evParam.m_RefreshTime = refresh;
	evParam.m_IntegerHyst = 0;
	evParam.m_FloatHyst = 0;

	status = ANSLI_VAR_Connect(parentIdent, name, pIdent, (ANSL_CALL_BACK)ANSLCB_all, pUserData, (ANSL_VARIABLE_EVENT_PARAMETER*)&evParam);
	
	if (0 != status)
	{
		printf("ANSLI_VAR_Connect  \"%s\"   E#%i\n", name, status);
	}
	return status;
}

unsigned int ReadVariable(unsigned int varIdent)
{
	unsigned int status = 0;

	//printf("Reading PV ident=%u \n", varIdent);
	status = ANSLI_VAR_ReadValue(varIdent, NULL, 0);
	if (status !=0)
	{
		printf("ERROR %u READ PV ident=%u \n", status, varIdent);
	}
	return status;
}

unsigned int WriteVariable(unsigned int varIdent, long varValue)
{
	unsigned int status = 0;

	printf("Write PV ident=%u to %d\n", varIdent, (int)varValue);
	status = ANSLI_VAR_SetValue(varIdent, &varValue, sizeof(varValue));
	if (status !=0)
	{
		printf("ERROR %u SET PV ident=%u \n", status, varIdent);
	}
	return status;
}

unsigned int WriteVariableInt(unsigned int varIdent, long varValue)
{
	unsigned int status = 0;

	//printf("Write PV ident=%u to %d\n", varIdent, varValue);
	status = ANSLI_VAR_SetValue(varIdent, &varValue, sizeof(varValue));
	if (status !=0)
	{
		printf("ERROR %u SET PV ident=%u \n", status, varIdent);
	}
	return status;
}

unsigned int WriteVariableReal(unsigned int varIdent, float varValue)
{
	unsigned int status = 0;

	printf("Write PV ident=%u to %.2f\n", varIdent, varValue);
	status = ANSLI_VAR_SetValue(varIdent, &varValue, sizeof(varValue));
	if (status !=0)
	{
		printf("ERROR %u SET PV ident=%u \n", status, varIdent);
	}
	return status;
}



int ioGetch ()
{
	 int ch;
#ifdef WIN32
	 ch = _getch();
#else
	 struct termios old_t, tmp_t;

	 // alte Attribute holen
	 if (tcgetattr (STDIN_FILENO, &old_t))
	 {
		  // moeglicherweise kein Terminal
		  // moeglicherweise anderer Fehler
		  return -1;
	 }
	 // struktur kopieren
	 memcpy (&tmp_t, &old_t, sizeof (old_t));
	 // Zeilenmodus sowie Spezialzeichen deaktivieren, Bildschirmausgabe von eingegebenen Zeichen deaktivieren
	 tmp_t.c_lflag &= ~ICANON & ~ECHO;
	 // Neue Flags setzen
	 if (tcsetattr (STDIN_FILENO, TCSANOW, (const struct termios *)&tmp_t))
	 {
		  // moeglicherweise irgend ein Fehler
		  return -1;
	 }
	 // zeichen einlesen (darauf verlassen, STDIN_FILENO blockierend ist)
	 ch = getchar ();
	 // Alte Flags zuruecksetzen
	 // Rueckgabewert ignorieren, da das Terminal bei Fehler sowieso
	 // nicht korrekt reagieren wird, wir dagegen aber nichts tun koennen
	 tcsetattr (STDIN_FILENO, TCSANOW, (const struct termios *)&old_t);
#endif
	 // Zeichen zurueckgeben
	 return ch;
}

/***************************************************************************************************/
/** ERROR HANDLING *********************************************************************************/

int ErrorCheck(unsigned int status, const char* error)
{
	if (status != 0)
	{
		printf("%s: %i\n", error, status);
		return 1;
	}
	return 0;
}

/***************************************************************************************************/
/** CALLBACK FUNCTION ******************************************************************************/

/*my*/

unsigned int ANSLCB_all(unsigned int errorCode, 
						unsigned int ident, 
						unsigned int objectType,
						unsigned int cbType, 
						void* pData, 
						unsigned int dataLen, 
						void* pUserData)
{	
	char strOut[6000];
	char strName[128];
	char strFmt[FMT_STR_SIZE];
	char strTyp[INFO_STR_SIZE];
	char strEvent[INFO_STR_SIZE];
	memset(strOut, 0, 6000);
	memset(strName, 0, 128);
	memset(strFmt, 0, FMT_STR_SIZE);
	memset(strTyp, 0, INFO_STR_SIZE);
	memset(strEvent, 0, INFO_STR_SIZE);

	GetObjTypeString(objectType, strTyp);
	
	/*my*/
	ANSL_VARIABLE_DATA_TYPE* pDataTypePV[100];
	
	switch (cbType)
	{
		case ANSL_RESP_CONNECT:
		{
			sprintf(strOut, "CONNECT  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			//printf(strOut);
		} break;
		case ANSL_RESP_DISCONNECT:
		{
			sprintf(strOut, "DISCONNECT  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			//printf(strOut);
			_CPUConnClosed = 1;
		} break;
		case ANSL_EVENT_ERROR:
		{
			sprintf(strOut, "ERROR  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			printf(strOut);
		} break;
		case ANSL_EVENT_STATUS:
		{
			sprintf(strOut, "STATUS  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			printf(strOut);
		} break;
		case ANSL_EVENT_FINAL:
		{
			sprintf(strOut, "FINAL  %s error=%u ident=%u   serviceNumber=%u\n", 
					strTyp, errorCode, ident, dataLen);	
			//printf(strOut);
		} break;
		/* read date and time */
		case ANSL_RESP_GET_DATETIME:
		{
			CPU_DATETIME_TYPE cpuDataTime;

			if	(
					(errorCode == 0) &&
					(sizeof(CPU_DATETIME_TYPE) == dataLen)
				)
			{
				memcpy(&cpuDataTime, pData, dataLen);
				//                        year starts at 1900         month starts at 0    
				printf("Date: %i/%i/%i ", cpuDataTime.tm_Year + 1900, cpuDataTime.tm_Month + 1, cpuDataTime.tm_DayOfMonth);
				printf("Time: %i:%i:%i\n", cpuDataTime.tm_Hour, cpuDataTime.tm_Minute, cpuDataTime.tm_Second);
			}
			else
			{
				printf("ERROR GET_DATETIME: %i\n", errorCode);
			}
		} break;

		/* read READ_VAR */
		case ANSL_RESP_GET_VALUE:
		/* read READ_VAR */
		case ANSL_EVENT_VALUE:
		{
			sprintf(strName, "%s", _strPV[*((unsigned int*)pUserData)]);
			pDataTypePV[*((unsigned int*)pUserData)] = &_DataTypePV[*((unsigned int*)pUserData)]; 
			GetCBTypeString(cbType, strEvent);
			
			sprintf(strOut, "%s \"%s\": %s  error=%u ident=%u   dataLen=%u\n", 
					strEvent,
					strName,
					 
					((0 == errorCode) ? DataToString(pData, dataLen, pDataTypePV[*((unsigned int*)pUserData)]) : "ERROR"),
					errorCode, 
					ident, 
					dataLen);	
			printf(strOut);
		} break;

		/* read format SET_VAR */
		case ANSL_EVENT_FORMAT:
		case ANSL_RESP_GET_FORMAT:
		{
			sprintf(strName, "%s", _strPV[*((unsigned int*)pUserData)]);
			memcpy(&_DataTypePV[*((unsigned int*)pUserData)], pData, sizeof(ANSL_VARIABLE_DATA_TYPE));
			memset (strFmt, 0, FMT_STR_SIZE);
			
			if (0 == errorCode)
			{
				GetFormatString((ANSL_VARIABLE_DATA_TYPE*)pData, strFmt);
				GetFormatString((ANSL_VARIABLE_DATA_TYPE*)pData, _myStrFormat[*((unsigned int*)pUserData)]);
				
			} else {
				sprintf(strFmt, "???");
			}
			GetCBTypeString(cbType, strEvent);
			sprintf(strOut, "%s  %s \"%s\": %s  error=%u ident=%u   dataLen=%u\n", 
					strTyp, 
					strEvent,
					strName,
					strFmt, 
					errorCode, 
					ident, 
					dataLen);
			
			printf(strOut);

		} break;
		case ANSL_RESP_SET_ATTRIBUTES:
		{
			sprintf(strName, "%s", _strPV[*((unsigned int*)pUserData)]);
			GetCBTypeString(cbType, strEvent);
			sprintf(strOut, "%s \"%s\": error=%u ident=%u\n", 
					strEvent,
					strName,					 
					errorCode, 
					ident);	
			//printf(strOut);
		} break;
		/* set SET_VAR */
		case ANSL_RESP_SET_VALUE:
		{	
			sprintf(strName, "%s", _strPV[*((unsigned int*)pUserData)]);
			
			GetCBTypeString(cbType, strEvent);
			sprintf(strOut, "%s : error=%u ident=%u\n", 
					strEvent,
										 
					errorCode, 
					ident);	
			printf(strOut);
		} break;
		case ANSL_RESP_GET_HARDWARE:
		{
			sprintf(strOut, "%s\n", (char*)pData);	
			printf(strOut);
		} break;
		case ANSL_RESP_GET_INFO:
		{
			sprintf(strOut, "%s\n", (char*)pData);	
			printf(strOut);
		} break;
		case ANSL_RESP_GET_STORAGE:
		{
			sprintf(strOut, "%s\n", (char*)pData);	
			printf(strOut);
		} break;
		
		/*my*/
		case ANSL_RESP_LIST_VARIABLES:
		{
		  
		  ANSL_VARIABLE_DESC* pPVDesc;
		  pPVDesc = (ANSL_VARIABLE_DESC*)pData;
		  
		  while(NULL != pPVDesc)
		  {	
		   if(NULL!= pPVDesc->m_Name)
		   {	
		    //DEBUG_WriteLine("PV name = %s Scope = %u", pPVDesc->m_Name, pPVDesc->m_Scope);
			sprintf(strOut, "PV name = %s  \n", pPVDesc->m_Name);	
			printf(strOut);
		   }
		    pPVDesc = pPVDesc->m_pNext;
		  }
		  
		}break;
		
		
		default:
		{	
			GetCBTypeString(cbType, strEvent);
			sprintf(strOut, "%s  %s error=%u ident=%u   dataLen=%u\n", 
					strTyp, 
					strEvent, 
					errorCode, 
					ident, 
					dataLen);	
			//printf(strOut);
		} break;
	}
	return 0;
}

/*my*/


int GetCBTypeString(unsigned int eventCode, char strValue[INFO_STR_SIZE])
{
	memset(strValue, 0, INFO_STR_SIZE);
	switch (eventCode)
	{
		case ANSL_EVENT_FINAL: { strcpy(strValue, "EVENT_FINAL"); } break;
		case ANSL_EVENT_ERROR: { strcpy(strValue, "EVENT_ERROR"); } break;
		case ANSL_EVENT_STATUS: { strcpy(strValue, "EVENT_STATUS"); } break;
		case ANSL_REQ_CONNECT: { strcpy(strValue, "REQ_CONNECT"); } break;
		case ANSL_RESP_CONNECT: { strcpy(strValue, "RESP_CONNECT"); } break;
		case ANSL_CANCEL_CONNECT: { strcpy(strValue, "CANCEL_CONNECT"); } break;
		case ANSL_REQ_DISCONNECT: { strcpy(strValue, "REQ_DISCONNECT"); } break;
		case ANSL_RESP_DISCONNECT: { strcpy(strValue, "RESP_DISCONNECT"); } break;
		case ANSL_CANCEL_DISCONNECT: { strcpy(strValue, "CANCEL_DISCONNECT"); } break;
		case ANSL_REQ_GET_DATETIME: { strcpy(strValue, "REQ_GET_DATETIME"); } break;
		case ANSL_RESP_GET_DATETIME: { strcpy(strValue, "RESP_GET_DATETIME"); } break;
		case ANSL_CANCEL_GET_DATETIME: { strcpy(strValue, "CANCEL_GET_DATETIME"); } break;
		case ANSL_REQ_SET_DATETIME: { strcpy(strValue, "REQ_SET_DATETIME"); } break;
		case ANSL_RESP_SET_DATETIME: { strcpy(strValue, "RESP_SET_DATETIME"); } break;
		case ANSL_CANCEL_SET_DATETIME: { strcpy(strValue, "CANCEL_SET_DATETIME"); } break;
		case ANSL_REQ_WARM_RESTART: { strcpy(strValue, "REQ_WARM_RESTART"); } break;
		case ANSL_RESP_WARM_RESTART: { strcpy(strValue, "RESP_WARM_RESTART"); } break;
		case ANSL_CANCEL_WARM_RESTART: { strcpy(strValue, "CANCEL_WARM_RESTART"); } break;
		case ANSL_REQ_COLD_RESTART: { strcpy(strValue, "REQ_COLD_RESTART"); } break;
		case ANSL_RESP_COLD_RESTART: { strcpy(strValue, "RESP_COLD_RESTART"); } break;
		case ANSL_CANCEL_COLD_RESTART: { strcpy(strValue, "CANCEL_COLD_RESTART"); } break;
		case ANSL_REQ_GET_INFO: { strcpy(strValue, "REQ_GET_INFO"); } break;
		case ANSL_RESP_GET_INFO: { strcpy(strValue, "RESP_GET_INFO"); } break;
		case ANSL_CANCEL_GET_INFO: { strcpy(strValue, "CANCEL_GET_INFO"); } break;
		case ANSL_REQ_GET_STORAGE: { strcpy(strValue, "REQ_GET_STORAGE"); } break;
		case ANSL_RESP_GET_STORAGE: { strcpy(strValue, "RESP_GET_STORAGE"); } break;
		case ANSL_CANCEL_GET_STORAGE: { strcpy(strValue, "CANCEL_GET_STORAGE"); } break;
		case ANSL_REQ_GET_TASKCLASSES: { strcpy(strValue, "REQ_GET_TASKCLASSES"); } break;
		case ANSL_RESP_GET_TASKCLASSES: { strcpy(strValue, "RESP_GET_TASKCLASSES"); } break;
		case ANSL_CANCEL_GET_TASKCLASSES: { strcpy(strValue, "CANCEL_GET_TASKCLASSES"); } break;
		case ANSL_REQ_GET_HARDWARE: { strcpy(strValue, "REQ_GET_HARDWARE"); } break;
		case ANSL_RESP_GET_HARDWARE: { strcpy(strValue, "RESP_GET_HARDWARE"); } break;
		case ANSL_CANCEL_GET_HARDWARE: { strcpy(strValue, "CANCEL_GET_HARDWARE"); } break;
		case ANSL_REQ_UPLOAD: { strcpy(strValue, "REQ_UPLOAD"); } break;
		case ANSL_RESP_UPLOAD: { strcpy(strValue, "RESP_UPLOAD"); } break;
		case ANSL_CANCEL_UPLOAD: { strcpy(strValue, "CANCEL_UPLOAD"); } break;
		case ANSL_REQ_DOWNLOAD: { strcpy(strValue, "REQ_DOWNLOAD"); } break;
		case ANSL_RESP_DOWNLOAD: { strcpy(strValue, "RESP_DOWNLOAD"); } break;
		case ANSL_CANCEL_DOWNLOAD: { strcpy(strValue, "CANCEL_DOWNLOAD"); } break;
		case ANSL_REQU_CREATE_MODULE: { strcpy(strValue, "REQU_CREATE_MODULE"); } break;
		case ANSL_RESP_CREATE_MODULE: { strcpy(strValue, "RESP_CREATE_MODULE"); } break;
		case ANSL_CANCEL_CREATE_MODULE: { strcpy(strValue, "CANCEL_CREATE_MODULE"); } break;
		case ANSL_REQU_DELETE_MODULE: { strcpy(strValue, "REQU_DELETE_MODULE"); } break;
		case ANSL_RESP_DELETE_MODULE: { strcpy(strValue, "RESP_DELETE_MODULE"); } break;
		case ANSL_CANCEL_DELETE_MODULE: { strcpy(strValue, "CANCEL_DELETE_MODULE"); } break;
		case ANSL_REQU_LIST_MODULES: { strcpy(strValue, "REQU_LIST_MODULES"); } break;
		case ANSL_RESP_LIST_MODULES: { strcpy(strValue, "RESP_LIST_MODULES"); } break;
		case ANSL_CANCEL_LIST_MODULES: { strcpy(strValue, "CANCEL_LIST_MODULES"); } break;
		case ANSL_EVENT_LIST_MODULES: { strcpy(strValue, "EVENT_LIST_MODULES"); } break;
		case ANSL_EVENT_FORMAT: { strcpy(strValue, "EVENT_FORMAT"); } break;
		case ANSL_REQ_GET_FORMAT: { strcpy(strValue, "REQ_GET_FORMAT"); } break;
		case ANSL_RESP_GET_FORMAT: { strcpy(strValue, "RESP_GET_FORMAT"); } break;
		case ANSL_CANCEL_GET_FORMAT: { strcpy(strValue, "CANCEL_GET_FORMAT"); } break;
		case ANSL_EVENT_VALUE: { strcpy(strValue, "EVENT_VALUE"); } break;
		case ANSL_CANCEL_EVENT_VALUE: { strcpy(strValue, "CANCEL_EVENT_VALUE"); } break;
		case ANSL_REQ_GET_VALUE: { strcpy(strValue, "REQ_GET_VALUE"); } break;
		case ANSL_RESP_GET_VALUE: { strcpy(strValue, "RESP_GET_VALUE"); } break;
		case ANSL_CANCEL_GET_VALUE: { strcpy(strValue, "CANCEL_GET_VALUE"); } break;
		case ANSL_REQ_SET_ATTRIBUTES: { strcpy(strValue, "REQ_SET_ATTRIBUTES"); } break;
		case ANSL_RESP_SET_ATTRIBUTES: { strcpy(strValue, "RESP_SET_ATTRIBUTES"); } break;
		case ANSL_CANCEL_SET_ATTRIBUTES: { strcpy(strValue, "CANCEL_SET_ATTRIBUTES"); } break;
		case ANSL_REQ_SET_VALUE: { strcpy(strValue, "REQ_SET_VALUE"); } break;
		case ANSL_RESP_SET_VALUE: { strcpy(strValue, "RESP_SET_VALUE"); } break;
		case ANSL_CANCEL_SET_VALUE: { strcpy(strValue, "CANCEL_SET_VALUE"); } break;
		case ANSL_REQ_GET_CONTENT: { strcpy(strValue, "REQ_GET_CONTENT"); } break;
		case ANSL_RESP_GET_CONTENT: { strcpy(strValue, "RESP_GET_CONTENT"); } break;
		case ANSL_CANCEL_GET_CONTENT: { strcpy(strValue, "CANCEL_GET_CONTENT"); } break;
		case ANSL_EVENT_CONTENT: { strcpy(strValue, "EVENT_CONTENT"); } break;
		case ANSL_CANCEL_EVENT_CONTENT: { strcpy(strValue, "CANCEL_EVENT_CONTENT"); } break;
		case ANSL_REQ_READ_ENTRIES: { strcpy(strValue, "REQ_READ_ENTRIES"); } break;
		case ANSL_RESP_READ_ENTRIES: { strcpy(strValue, "RESP_READ_ENTRIES"); } break;
		case ANSL_CANCEL_READ_ENTRIES: { strcpy(strValue, "CANCEL_READ_ENTRIES"); } break;
		case ANSL_REQ_NEW_LENTRY: { strcpy(strValue, "REQ_NEW_LENTRY"); } break;
		case ANSL_RESP_NEW_LENTRY: { strcpy(strValue, "RESP_NEW_LENTRY"); } break;
		case ANSL_CANCEL_NEW_LENTRY: { strcpy(strValue, "CANCEL_NEW_LENTRY"); } break;
		case ANSL_REQ_GLKN_LINKLIST: { strcpy(strValue, "REQ_GLKN_LINKLIST"); } break;
		case ANSL_RESP_GLKN_LINKLIST: { strcpy(strValue, "RESP_GLKN_LINKLIST"); } break;
		case ANSL_CANCEL_GLKN_LINKLIST: { strcpy(strValue, "CANCEL_GLKN_LINKLIST"); } break;
		case ANSL_REQ_GLKN_FORCEOFF: { strcpy(strValue, "REQ_GLKN_FORCEOFF"); } break;
		case ANSL_RESP_GLKN_FORCEOFF: { strcpy(strValue, "RESP_GLKN_FORCEOFF"); } break;
		case ANSL_CANCEL_GLKN_FORCEOFF: { strcpy(strValue, "CANCEL_GLKN_FORCEOFF"); } break;
		case ANSL_REQ_LKN_FORMAT: { strcpy(strValue, "REQ_LKN_FORMAT"); } break;
		case ANSL_RESP_LKN_FORMAT: { strcpy(strValue, "RESP_LKN_FORMAT"); } break;
		case ANSL_CANCEL_LKN_FORMAT: { strcpy(strValue, "CANCEL_LKN_FORMAT"); } break;
		case ANSL_REQ_LKN_DATA: { strcpy(strValue, "REQ_LKN_DATA"); } break;
		case ANSL_RESP_LKN_DATA: { strcpy(strValue, "RESP_LKN_DATA"); } break;
		case ANSL_CANCEL_LKN_DATA: { strcpy(strValue, "CANCEL_LKN_DATA"); } break;
		case ANSL_REQ_LKN_FORCEDATA: { strcpy(strValue, "REQ_LKN_FORCEDATA"); } break;
		case ANSL_RESP_LKN_FORCEDATA: { strcpy(strValue, "RESP_LKN_FORCEDATA"); } break;
		case ANSL_CANCEL_LKN_FORCEDATA: { strcpy(strValue, "CANCEL_LKN_FORCEDATA"); } break;
		case ANSL_REQ_LKN_FORCEON: { strcpy(strValue, "REQ_LKN_FORCEON"); } break;
		case ANSL_RESP_LKN_FORCEON: { strcpy(strValue, "RESP_LKN_FORCEON"); } break;
		case ANSL_CANCEL_LKN_FORCEON: { strcpy(strValue, "CANCEL_LKN_FORCEON"); } break;
		case ANSL_REQ_LKN_FORCEOFF: { strcpy(strValue, "REQ_LKN_FORCEOFF"); } break;
		case ANSL_RESP_LKN_FORCEOFF: { strcpy(strValue, "RESP_LKN_FORCEOFF"); } break;
		case ANSL_CANCEL_LKN_FORCEOFF: { strcpy(strValue, "CANCEL_LKN_FORCEOFF"); } break;
		case ANSL_REQ_LOGGER_INFO: { strcpy(strValue, "REQ_LOGGER_INFO"); } break;
		case ANSL_RESP_LOGGER_INFO: { strcpy(strValue, "RESP_LOGGER_INFO"); } break;
		case ANSL_CANCEL_LOGGER_INFO: { strcpy(strValue, "CANCEL_LOGGER_INFO"); } break;
		case ANSL_REQ_MODULE_INFO: { strcpy(strValue, "REQ_MODULE_INFO"); } break;
		case ANSL_RESP_MODULE_INFO: { strcpy(strValue, "RESP_MODULE_INFO"); } break;
		case ANSL_CANCEL_MODULE_INFO: { strcpy(strValue, "CANCEL_MODULE_INFO"); } break;
		case ANSL_REQ_TASK_INFO: { strcpy(strValue, "REQ_TASK_INFO"); } break;
		case ANSL_RESP_TASK_INFO: { strcpy(strValue, "RESP_TASK_INFO"); } break;
		case ANSL_CANCEL_TASK_INFO: { strcpy(strValue, "CANCEL_TASK_INFO"); } break;
		case ANSL_REQU_TASK_SECTION: { strcpy(strValue, "REQU_TASK_SECTION"); } break;
		case ANSL_RESP_TASK_SECTION: { strcpy(strValue, "RESP_TASK_SECTION"); } break;
		case ANSL_CANCEL_TASK_SECTION: { strcpy(strValue, "CANCEL_TASK_SECTION"); } break;
		case ANSL_REQU_LIST_VARIABLES: { strcpy(strValue, "REQU_LIST_VARIABLES"); } break;
		case ANSL_RESP_LIST_VARIABLES: { strcpy(strValue, "RESP_LIST_VARIABLES"); } break;
		case ANSL_CANCEL_LIST_VARIABLES: { strcpy(strValue, "CANCEL_LIST_VARIABLES"); } break;
		case ANSL_SERVER_CLIENT_CONNECT: { strcpy(strValue, "SERVER_CLIENT_CONNECT"); } break;
		case ANSL_SERVER_CLIENT_DISCONNECT: { strcpy(strValue, "SERVER_CLIENT_DISCONNECT"); } break;
		case ANSL_DATA_CHANGED: { strcpy(strValue, "DATA_CHANGED"); } break;
		case ANSL_CANCEL_DATA_CHANGED: { strcpy(strValue, "CANCEL_DATA_CHANGED"); } break;
		case ANSL_EVENT_DATA: { strcpy(strValue, "EVENT_DATA"); } break;
		case ANSL_CANCEL_EVENT_DATA: { strcpy(strValue, "CANCEL_EVENT_DATA"); } break;
		case ANSL_REQ_READ_DATA: { strcpy(strValue, "REQ_READ_DATA"); } break;
		case ANSL_RESP_READ_DATA: { strcpy(strValue, "RESP_READ_DATA"); } break;
		case ANSL_CANCEL_READ_DATA: { strcpy(strValue, "CANCEL_READ_DATA"); } break;
		case ANSL_REQ_WRITE_DATA: { strcpy(strValue, "REQ_WRITE_DATA"); } break;
		case ANSL_RESP_WRITE_DATA: { strcpy(strValue, "RESP_WRITE_DATA"); } break;
		case ANSL_CANCEL_WRITE_DATA: { strcpy(strValue, "CANCEL_WRITE_DATA"); } break;
		case ANSL_REQU_LIST_DATA_OBJECTS: { strcpy(strValue, "REQU_LIST_DATA_OBJECTS"); } break;
		case ANSL_RESP_LIST_DATA_OBJECTS: { strcpy(strValue, "RESP_LIST_DATA_OBJECTS"); } break;
		case ANSL_CANCEL_LIST_DATA_OBJECTS: { strcpy(strValue, "CANCEL_LIST_DATA_OBJECTS"); } break;
		case ANSL_CLIENT_DATA_OBJ_CONNECT: { strcpy(strValue, "CLIENT_DATA_OBJ_CONNECT"); } break;
		case ANSL_CLIENT_DATA_OBJ_DISCONNECT: { strcpy(strValue, "CLIENT_DATA_OBJ_DISCONNECT"); } break;
		case ANSL_REQ_RESET: { strcpy(strValue, "REQ_RESET"); } break;
		case ANSL_RESP_RESET: { strcpy(strValue, "RESP_RESET"); } break;
		case ANSL_CANCEL_RESET: { strcpy(strValue, "CANCEL_RESET"); } break;
		case ANSL_REQ_CLEAR_MEMORY: { strcpy(strValue, "REQ_CLEAR_MEMORY"); } break;
		case ANSL_RESP_CLEAR_MEMORY: { strcpy(strValue, "RESP_CLEAR_MEMORY"); } break;
		case ANSL_CANCEL_CLEAR_MEMORY: { strcpy(strValue, "CANCEL_CLEAR_MEMORY"); } break;
		case ANSL_REQ_RESUME: { strcpy(strValue, "REQ_RESUME"); } break;
		case ANSL_RESP_RESUME: { strcpy(strValue, "RESP_RESUME"); } break;
		case ANSL_CANCEL_RESUME: { strcpy(strValue, "CANCEL_RESUME"); } break;
		case ANSL_REQ_SUSPEND: { strcpy(strValue, "REQ_SUSPEND"); } break;
		case ANSL_RESP_SUSPEND: { strcpy(strValue, "RESP_SUSPEND"); } break;
		case ANSL_CANCEL_SUSPEND: { strcpy(strValue, "CANCEL_SUSPEND"); } break;
		case ANSL_REQ_CLEAR_LOGBUFFER: { strcpy(strValue, "REQ_CLEAR_LOGBUFFER"); } break;
		case ANSL_RESP_CLEAR_LOGBUFFER: { strcpy(strValue, "RESP_CLEAR_LOGBUFFER"); } break;
		case ANSL_CANCEL_CLEAR_LOGBUFFER: { strcpy(strValue, "CANCEL_CLEAR_LOGBUFFER"); } break;
		case ANSL_REQ_BITMASK: { strcpy(strValue, "REQ_BITMASK"); } break;
		case ANSL_RESP_BITMASK: { strcpy(strValue, "RESP_BITMASK"); } break;
		case ANSL_CANCEL_BITMASK: { strcpy(strValue, "CANCEL_BITMASK"); } break;
		case ANSL_PROCEEDING_LIST_VARIABLES: { strcpy(strValue, "PROCEEDING_LIST_VARIABLES"); } break;
		case ANSL_EVENT_MODULE_CHANGED: { strcpy(strValue, "EVENT_MODULE_CHANGED"); } break;
		case ANSL_CANCEL_MODULE_CHANGED: { strcpy(strValue, "CANCEL_MODULE_CHANGED"); } break;
		case ANSL_SRV_REGISTER_OBJECT: { strcpy(strValue, "SRV_REGISTER_OBJECT"); } break;
		case ANSL_SRV_OBJECT_REGISTERED: { strcpy(strValue, "SRV_OBJECT_REGISTERED"); } break;
		case ANSL_SRV_REG_OBJECT_FINAL: { strcpy(strValue, "SRV_REG_OBJECT_FINAL"); } break;
		case ANSL_SRV_DEREGISTER_OBJECT: { strcpy(strValue, "SRV_DEREGISTER_OBJECT"); } break;
		case ANSL_SRV_OBJECT_DEREGISTER: { strcpy(strValue, "SRV_OBJECT_DEREGISTER"); } break;
		case ANSL_SRV_TELEGRAM_SEND: { strcpy(strValue, "SRV_TELEGRAM_SEND"); } break;
		case ANSL_SRV_TELEGRAM_SENT: { strcpy(strValue, "SRV_TELEGRAM_SENT"); } break;
		case ANSL_SRV_STREAM_SEND: { strcpy(strValue, "SRV_STREAM_SEND"); } break;
		case ANSL_SRV_STREAM_SENT: { strcpy(strValue, "SRV_STREAM_SENT"); } break;
		case ANSL_SRV_TELEGRAM_RECEIVE: { strcpy(strValue, "SRV_TELEGRAM_RECEIVE"); } break;
		case ANSL_SRV_TELEGRAM_RECEIVED: { strcpy(strValue, "SRV_TELEGRAM_RECEIVED"); } break;
		case ANSL_SRV_STREAM_RECEIVE: { strcpy(strValue, "SRV_STREAM_RECEIVE"); } break;
		case ANSL_SRV_STREAM_RECEIVED: { strcpy(strValue, "SRV_STREAM_RECEIVED"); } break;
		case ANSL_REQ_CONTROL: { strcpy(strValue, "REQ_CONTROL"); } break;
		case ANSL_RESP_CONTROL: { strcpy(strValue, "RESP_CONTROL"); } break;
		case ANSL_CANCEL_CONTROL: { strcpy(strValue, "CANCEL_CONTROL"); } break;
		case ANSL_CANCEL_EVENT_FORMAT: { strcpy(strValue, "CANCEL_EVENT_FORMAT"); } break;
		case ANSL_PROCEEDING_EVENT: { strcpy(strValue, "PROCEEDING_EVENT"); } break;
		case ANSL_REQ_ASSIGN_VALUE: { strcpy(strValue, "REQ_ASSIGN_VALUE"); } break;
		case ANSL_EVENT_TASKCLASSES_CHANGED: { strcpy(strValue, "EVENT_TASKCLASSES_CHANGED"); } break;
		case ANSL_CANCEL_TASKCLASSES_CHANGED: { strcpy(strValue, "CANCEL_TASKCLASSES_CHANGED"); } break;
		case ANSL_CB_TYPES_END: { strcpy(strValue, "CB_TYPES_END"); } break;
		case ANSL_PROCEEDING_DOWNLOAD: { strcpy(strValue, "PROCEEDING_DOWNLOAD"); } break;
		case ANSL_PROCEEDING_UPLOAD: { strcpy(strValue, "PROCEEDING_UPLOAD"); } break;
		case ANSL_INITIALIZE: { strcpy(strValue, "INITIALIZE"); } break;
		case ANSL_DEINITIALIZE: { strcpy(strValue, "DEINITIALIZE"); } break;
		case ANSL_CONNECTION_STATES: { strcpy(strValue, "CONNECTION_STATES"); } break;
		case ANSL_CONNECTION_ROUTER_ERROR: { strcpy(strValue, "CONNECTION_ROUTER_ERROR"); } break;
		case ANSL_CONNECTION_LISTEN_ERROR: { strcpy(strValue, "CONNECTION_LISTEN_ERROR"); } break;
		case ANSL_CONNECTION_LISTEN_CLOSE: { strcpy(strValue, "CONNECTION_LISTEN_CLOSE"); } break;
		case ANSL_CONNECTION_READY: { strcpy(strValue, "CONNECTION_READY"); } break;
		case ANSL_CONNECTION_ERROR: { strcpy(strValue, "CONNECTION_ERROR"); } break;
		case ANSL_CONNECTION_CLOSE: { strcpy(strValue, "CONNECTION_CLOSE"); } break;
		case ANSL_SERVER_START: { strcpy(strValue, "SERVER_START"); } break;
		case ANSL_SERVER_STARTED: { strcpy(strValue, "SERVER_STARTED"); } break;
		case ANSL_SERVER_STOP: { strcpy(strValue, "SERVER_STOP"); } break;
		case ANSL_SERVER_STOPPED: { strcpy(strValue, "SERVER_STOPPED"); } break;
		case ANSL_SERVER_ERROR: { strcpy(strValue, "SERVER_ERROR"); } break;
		case ANSL_ERROR_LISTEN: { strcpy(strValue, "ERROR_LISTEN"); } break;
		case ANSL_LISTEN_CLOSE: { strcpy(strValue, "LISTEN_CLOSE"); } break;
		case ANSL_ERROR_CONN: { strcpy(strValue, "ERROR_CONN"); } break;
		case ANSL_CONN_CLOSE: { strcpy(strValue, "CONN_CLOSE"); } break;
		case ANSL_CONN_READY: { strcpy(strValue, "CONN_READY"); } break;
		case ANSL_SERVER_STATUS: { strcpy(strValue, "SERVER_STATUS"); } break;
		default:
		{
			sprintf_s(strValue, INFO_STR_SIZE-1, "%u", eventCode);
		} break;
	}
	return strlen(strValue);
}

int GetObjTypeString(unsigned int objType, char strValue[INFO_STR_SIZE])
{
	switch (objType)
	{
		case ANSL_OBJTYPE_ROOT: { strcpy(strValue, "ROOT"); } break;
		case ANSL_OBJTYPE_SERVER: { strcpy(strValue, "SERVER"); } break;
		case ANSL_OBJTYPE_CONNECTION: { strcpy(strValue, "CONNECTION"); } break;
		case ANSL_OBJTYPE_CPU: { strcpy(strValue, "CPU"); } break;
		case ANSL_OBJTYPE_CPUPVS: { strcpy(strValue, "CPUPVS"); } break;
		case ANSL_OBJTYPE_CPUPV: { strcpy(strValue, "CPUPV"); } break;
		case ANSL_OBJTYPE_MODULE: { strcpy(strValue, "MODULE"); } break;
		case ANSL_OBJTYPE_LOGGER: { strcpy(strValue, "LOGGER"); } break;
		case ANSL_OBJTYPE_LIBRARY: { strcpy(strValue, "LIBRARY"); } break;
		case ANSL_OBJTYPE_TASK: { strcpy(strValue, "TASK"); } break;
		case ANSL_OBJTYPE_TASKPV: { strcpy(strValue, "TASKPV"); } break;
		case ANSL_OBJTYPE_GLOBAL_LINK_NODES: { strcpy(strValue, "GLOBAL_LINK_NODES"); } break;
		case ANSL_OBJTYPE_LINK_NODE: { strcpy(strValue, "LINK_NODE"); } break;
		case ANSL_OBJTYPE_REGISTERD: { strcpy(strValue, "REGISTERD"); } break;
		case ANSL_OBJTYPE_REGED_STREAM: { strcpy(strValue, "REGED_STREAM"); } break;
		case ANSL_OBJTYPE_REGED_TELEGRAM: { strcpy(strValue, "REGED_TELEGRAM"); } break;
		case ANSL_OBJTYPE_STREAM: { strcpy(strValue, "STREAM"); } break;
		case ANSL_OBJTYPE_TELEGRAM: { strcpy(strValue, "TELEGRAM"); } break;
		case ANSL_OBJTYPE_VARIABLE: { strcpy(strValue, "VARIABLE"); } break;
		default:
		{
			sprintf_s(strValue, INFO_STR_SIZE-1, "%u", objType);
		} break;
	}
	return strlen(strValue);
}

int GetFormatString(ANSL_VARIABLE_DATA_TYPE* pTypeInfo, char strValue[FMT_STR_SIZE])
{
	int dims = 0;
	char strDim[40];
	
	memset(strValue, 0, FMT_STR_SIZE);
	switch (pTypeInfo->m_DataType)
	{
		case ANSL_VAR_TYPE_UINT8: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=ui8 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=ui8 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_UINT16: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=ui16 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=ui16 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_UINT32: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=ui32 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=ui32 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_UINT64: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=ui64 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=ui64 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_INT8: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=i8 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=i8 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_INT16: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=i16 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=i16 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_INT32: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=i32 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=i32 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_INT64: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=i64 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=i64 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_DT: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=DateNTime Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=DateNTime Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_BOOL: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=bool Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=bool Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_FLOAT32: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=f32 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=f32 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_FLOAT64: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=f64 Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=f64 Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_CHAR_STRING: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=string Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=string Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_WCHAR_STRING: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=wstring Len=%d ArDim=%d Size=%d ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=wstring Len=%d", pTypeInfo->m_ElementLength);
			}
		} break;
		case ANSL_VAR_TYPE_STRUCT: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=struct Len=%d ArDim=%d Size=%d Name=%s ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size, pTypeInfo->m_Name);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=struct Len=%d Name=%s", pTypeInfo->m_ElementLength, pTypeInfo->m_Name);
			}
		} break;
		default: 
		{ 
			if (0 < pTypeInfo->m_ArrayDimensions)
			{
				sprintf(strValue, "Type=UNKNOWN Len=%d ArDim=%d Size=%d Name=%s ", pTypeInfo->m_ElementLength, pTypeInfo->m_ArrayDimensions, pTypeInfo->m_Size, pTypeInfo->m_Name);
				for (dims = 0; dims < pTypeInfo->m_ArrayDimensions; dims++)
				{
					memset(strDim, 0, 40);
					sprintf(strDim, " [%d]=%d", dims, pTypeInfo->m_DimensionElements[dims]);
					sprintf(strValue, "%s%s", strValue, strDim);
				}
			} else {
				sprintf(strValue, "Type=UNKNOWN Len=%d Name=%s", pTypeInfo->m_ElementLength, pTypeInfo->m_Name);
			}
		} break;
	}
	return strlen(strValue);
}

char* DataToString(char *pData, unsigned int dataLen, ANSL_VARIABLE_DATA_TYPE *pDataType)
{
	char strTmp[256];
	
	int i32V;
	unsigned int ui32V;
	short i16V;
	unsigned short ui16V;
	char i8V;
	unsigned char ui8V;
	double f64V = 0.0;
	float f32V;
	unsigned int xCnt, yCnt, xOffset;


	if ((NULL != pData) && (0 < dataLen))
	{
		switch (pDataType->m_DataType)
		{
			case ANSL_VAR_TYPE_BOOL:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&i8V, pData, 1);
					sprintf(_strValue, "%s", ((0 != i8V) ? "true" : "false"));
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&i8V, pData+xOffset, 1);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%s;",  ((0 != i8V) ? "true" : "false"));
							sprintf(_strValue, "%s%s", _strValue, strTmp);
							xOffset++;
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_INT8:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&i8V, pData, 1);
					sprintf(_strValue, "%d",  i8V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&i8V, pData+xOffset, 1);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%d;",  i8V);
							xOffset+=1;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_INT16:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&i16V, pData, 2);
					sprintf(_strValue, "%d",  i16V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&i16V, pData+xOffset, 2);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%d;",  i16V);
							xOffset+=2;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_INT32:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&i32V, pData, 4);
					sprintf(_strValue, "%d",  i32V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&i32V, pData+xOffset, 4);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%d;",  i32V);
							xOffset+=4;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_UINT8:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&ui8V, pData, 1);
					sprintf(_strValue, "%u",  ui8V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&ui8V, pData+xOffset, 1);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%u;",  ui8V);
							xOffset+=1;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_UINT16:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&ui16V, pData, 2);
					sprintf(_strValue, "%u",  ui16V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&ui16V, pData+xOffset, 2);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%u;",  ui16V);
							xOffset+=2;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_UINT32:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&ui32V, pData, 4);
					sprintf(_strValue, "%u",  ui32V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&ui32V, pData+xOffset, 4);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%u;",  ui32V);
							xOffset+=4;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_FLOAT32:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					memcpy(&f32V, pData, 4);
					sprintf(_strValue, "%.2f",  f32V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&f32V, pData+xOffset, 4);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%f;",  f32V);
							xOffset+=4;
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_FLOAT64:
			{
				//if (0 == pDataType->m_ArrayDimensions)
				//{
					memcpy(&f64V, pData, 8);
					sprintf(_strValue, "%g",  f64V);
				/*} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memcpy(&f64V, pData+xOffset, 8);
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%g;",  f64V);
							xOffset+=8;
							sprintf(_strValue, "%s%s", _strValue, strTmp;
						}
					}
				}
				*/
			} break;
			case ANSL_VAR_TYPE_CHAR_STRING:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					sprintf(_strValue, "%s", pData);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%s;", (char*)(pData+xOffset));
							sprintf(_strValue, "%s%s", _strValue, strTmp);
							xOffset+=pDataType->m_ElementLength;
						}
					}
				}
			} break;
			case ANSL_VAR_TYPE_STRUCT:
			{
				return "";
			} break;
			case ANSL_VAR_TYPE_DT:
			{
				CPU_DATETIME_Type* pTmpDT;
				if (0 == pDataType->m_ArrayDimensions)
				{
					pTmpDT = (CPU_DATETIME_Type*)pData;
					sprintf(_strValue, 
							"%04d/%02d/%02d %02d:%02d:%02d",
							(pTmpDT->Year + 1900),
							pTmpDT->Mon,
							pTmpDT->MDay,
							pTmpDT->Hour,
							pTmpDT->Min,
							pTmpDT->Sec);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							pTmpDT = (CPU_DATETIME_Type*)(pData+xOffset);
							memset(strTmp, 0, 256);
							sprintf(strTmp, 
									"%04d/%02d/%02d %02d:%02d:%02d",
									pTmpDT->Year + 1900,
									pTmpDT->Mon,
									pTmpDT->MDay,
									pTmpDT->Hour,
									pTmpDT->Min,
									pTmpDT->Sec);
							sprintf(_strValue, "%s%s", _strValue, strTmp);
							xOffset+=pDataType->m_ElementLength;
						}
					}
				}
			} break;
			default:
			{
				if (0 == pDataType->m_ArrayDimensions)
				{
					switch (pDataType->m_ElementLength)
					{
						case 1:
						{
							memcpy(&i8V, pData, 1);
							i32V = i8V;
						} break;
						case 2:
						{
							memcpy(&i16V, pData, 2);
							i32V = i16V;
						} break;
						case 4:
						{
							memcpy(&i32V, pData, 4);
						} break;
					}
					sprintf(_strValue, "%d",  i32V);
				} else {
					xOffset = 0;
					for (xCnt = 0; xCnt < pDataType->m_ArrayDimensions; xCnt++)
					{
						for (yCnt = 0; yCnt < pDataType->m_DimensionElements[xCnt]; yCnt++)
						{
							switch (pDataType->m_ElementLength)
							{
								case 1:
								{
									memcpy(&i8V, pData+xOffset, 1);
									i32V = i8V;
									xOffset+=1;
								} break;
								case 2:
								{
									memcpy(&i16V, pData+xOffset, 2);
									i32V = i16V;
									xOffset+=2;
								} break;
								case 4:
								{
									memcpy(&i32V, pData+xOffset, 4);
									xOffset+=4;
								} break;
							}
							memset(strTmp, 0, 256);
							sprintf(strTmp, "%d;", i32V);
							sprintf(_strValue, "%s%s", _strValue, strTmp);
						}
					}
				}
			} break;
		}
	} else {
		sprintf(_strValue, "INVALID DATA");
	}
	return _strValue;
}

void AnslGetVersionInfo()
{
	unsigned int status = 0;
	unsigned int uifVersion = 0;
	unsigned int oifVersion = 0;

	status = ANSLI_GetVersionInfos(&uifVersion, &oifVersion);
	if (status != 0)
	{
		printf("ERROR ANSLI_GetVersionInfos : %i\n",status);
	} else {
		printf("ANSLI_GetVersionInfos: uifV=%d.%d  oifV=%d\n", (uifVersion/100), (uifVersion%100), oifVersion);
	}
}

void TaskClassResume(unsigned int cpuIdent, unsigned short tcNumber)
{
	unsigned int status = 0;

	status = ANSLI_CPU_ResumeTaskClass(cpuIdent, tcNumber);
	if (status != 0)
	{
		printf("ERROR RESUME TC : %i\n",status);
	}
}

void TaskClassSuspend(unsigned int cpuIdent, unsigned short tcNumber)
{
	unsigned int status = 0;

	status = ANSLI_CPU_SuspendTaskClass(cpuIdent, tcNumber);
	if (status != 0)
	{
		printf("ERROR SUSPEND TC : %i\n",status);
	}
}

void TaskResume(unsigned int taskIdent, unsigned short numOfCycles)
{
	unsigned int status = 0;

	status = ANSLI_TASK_Resume(taskIdent, numOfCycles);
	if (status != 0)
	{
		printf("ERROR RESUME TASK : %i\n",status);
	}
}

void TaskSuspend(unsigned int taskIdent, unsigned short numOfCycles)
{
	unsigned int status = 0;

	status = ANSLI_TASK_Suspend(taskIdent, numOfCycles);
	if (status != 0)
	{
		printf("ERROR SUSPEND TASK : %i\n",status);
	}
}

void LoggerClearBuffer(unsigned int loggerIdent)
{
	unsigned int status = 0;

	status = ANSLI_LOGGER_ClearBuffer(loggerIdent);
	if (status != 0)
	{
		printf("ERROR CLEAR BUFFER  LOGGER: %i\n",status);
	}
}

void CpuClearMemory(unsigned int cpuIdent, ANSL_CPU_CLEARMEM_MODES memType)
{
	unsigned int status = 0;

	status = ANSLI_CPU_ClearMemory(cpuIdent, memType);
	if (status != 0)
	{
		printf("ERROR CLEAR MEM  CPU : %i\n",status);
	}
}


void delay(unsigned int ms)
{
	 unsigned int i, m, k;
	 for (i=ms;i>0;i--)
	 {
	   for(m=428000;m>0;m--){k++;};
	 }


}
